#include "BoltzmannME.h"
using namespace std;

BoltzmannME::BoltzmannME( MASSB &P ) {
	P.T.seconds = time(NULL);
	P.lfile << "Beginning Boltzmann ME: " << ctime(&P.T.seconds) << endl;
	cout << "Beginning Boltzmann ME:\n";

	P.set_visual();

	Inputs BSIN( P.searchfile );
	if ( !strcmp(P.searchmethod, "sa") ) {
		import_sa( BSIN );
	} else if ( !strcmp(P.searchmethod, "tabu") ) {
		import_tabu( BSIN );
	} else if ( !strcmp(P.searchmethod, "newton") ) {
		import_newton( BSIN );
	} else {
		cerr << "SEARCHMETHOD " << P.searchmethod << " not recognized" << endl;
	}
	BSIN.report_inputs( &P.lfile );
}

int BoltzmannME::lagrange_search( MASSB &P ) {
	if ( !strcmp(P.searchmethod, "sa") ) {
		// run simulated annealing search method;
		P.lfile << "Beginning Simulated Annealing ..." << endl;
		cout << "Beginning Simulated Annealing ..." << endl;
		if (  ! Lagrange_siman( P ) ) cout << "SEARCH OK!" << endl;
	} else if ( !strcmp(P.searchmethod, "tabu") ) {
		// run tabu search method;
		P.lfile << "Beginning Tabu stepping ..." << endl;
		cout << "Beginning Tabu stepping ..." << endl;
		tabu_restarts = -1; 
		int pass = 1;
		while (tabu_restarts <= max_tabu_restarts && pass == 1) {
			tabu_restarts++;
			pass = Lagrange_tabu( P );  
		}
		cout << "SEARCH OK!" << endl;
	} else if ( !strcmp(P.searchmethod, "newton") ) {
		// run newton search method;
		P.lfile << "Beginning Newtonian descent ..." << endl;
		cout << "Beginning Newtonian descent ..." << endl;
		P.copy_to_public();
		if (  ! Lagrange_newton( P ) ) cout << "SEARCH OK!" << endl;
		P.copy_from_public();
	} 
	
	P.T.seconds = time(NULL);
    P.lfile << "Ended Boltzmann ME: " << ctime(&P.T.seconds) << endl;

	return(0);
}

BoltzmannME::~BoltzmannME() {
}

int BoltzmannME::import_sa( Inputs &IN ) {
	// EDIT THIS BLOCK FOR THE INPUT TAGS AND NUMBER OF TAGS
	// REQUIRED BY THE APPLICATION
	int num_of_strings = 0;
	int num_of_ints = 1;
	int num_of_doubles = 5;

	char (*chararr)[20];
	char (*intarr)[20];
	char (*doublearr)[20];

	if (num_of_strings > 0 ) chararr = new char[num_of_strings][20];
	else chararr = 0;

	if (num_of_ints > 0) intarr = new char[num_of_ints][20];
	else intarr = 0;
	strcpy(intarr[0],"ITERS_AT_TEMP");	

	if (num_of_doubles > 0) doublearr = new char[num_of_doubles][20];
	else doublearr = 0;
	strcpy(doublearr[0],"MAX_STEP_SIZE");	strcpy(doublearr[1],"BOLTZMANN");
	strcpy(doublearr[2],"INIT_TEMP");		strcpy(doublearr[3],"ATTEN_TEMP");
	strcpy(doublearr[4],"MIN_TEMP");

	// Inputs class reads input file and finds input details
	IN.alloc_arrays( num_of_strings, chararr, num_of_ints, 
		intarr, num_of_doubles, doublearr );

	// copy values out of the Inputs class into this class for ease of use later

	// STRCPY CHARACTER ARRAYS

	// INTEGERS
	iterT = IN.intvals[0];

	// DOUBLES
	maxstep = IN.doublevals[0];			boltzmann = IN.doublevals[1];
	inittemp = IN.doublevals[2];		attentemp = IN.doublevals[3];
	mintemp = IN.doublevals[4];

	return(0);
}

int BoltzmannME::import_tabu ( Inputs &IN ) {
	// EDIT THIS BLOCK FOR THE INPUT TAGS AND NUMBER OF TAGS
	// REQUIRED BY THE APPLICATION
	int num_of_strings = 0;
	int num_of_ints = 5;
	int num_of_doubles = 1;

	char (*chararr)[20];
	char (*intarr)[20];
	char (*doublearr)[20];

	if (num_of_strings > 0 ) chararr = new char[num_of_strings][20];
	else chararr = 0;

	if (num_of_ints > 0) intarr = new char[num_of_ints][20];
	else intarr = 0;
	strcpy(intarr[0],"TENURE");		strcpy(intarr[1],"NEIGHBOURHOODSIZE");	
	strcpy(intarr[2],"TABUSTEPS");	strcpy(intarr[3],"MAX_SB");	
	strcpy(intarr[4],"MAX_RESTARTS");

	if (num_of_doubles > 0) doublearr = new char[num_of_doubles][20];
	else doublearr = 0;
	strcpy(doublearr[0],"STEPSIZE");	

	// Inputs class reads input file and finds input details
	IN.alloc_arrays( num_of_strings, chararr, num_of_ints, 
		intarr, num_of_doubles, doublearr );

	// copy values out of the Inputs class into this class for ease of use later

	// STRCPY CHARACTER ARRAYS

	// INTEGERS
	tenure = IN.intvals[0];		neighbourhoodsize = IN.intvals[1];
	tabusteps = IN.intvals[2];	max_sb = IN.intvals[3];
	max_tabu_restarts = IN.intvals[4];	

	// DOUBLES
	stepsize = IN.doublevals[0];	


	return(0);
}

int BoltzmannME::import_newton( Inputs &IN ) {
	// EDIT THIS BLOCK FOR THE INPUT TAGS AND NUMBER OF TAGS
	// REQUIRED BY THE APPLICATION
	int num_of_strings = 0;
	int num_of_ints = 3;
	int num_of_doubles = 3;

	char (*chararr)[20];
	char (*intarr)[20];
	char (*doublearr)[20];

	chararr = new char[num_of_strings][20];

	intarr = new char[num_of_ints][20];
	strcpy(intarr[0],"JSTEPS");		strcpy(intarr[1],"MAXITER");
	strcpy(intarr[2],"MINITER");

	doublearr = new char[num_of_doubles][20];
	strcpy(doublearr[0],"JANNEAL");	strcpy(doublearr[1],"FITPRECISION");
	strcpy(doublearr[2],"JSCALE");

	// Inputs class reads input file and finds input details
	IN.alloc_arrays( num_of_strings, chararr, num_of_ints, 
		intarr, num_of_doubles, doublearr );

	// copy values out of the Inputs class into this class for ease of use later

	// STRCPY CHARACTER ARRAYS

	// INTEGERS
	Jsteps = IN.intvals[0];	maxiter = IN.intvals[1];
	miniter = IN.intvals[2];

	// DOUBLES
	Janneal = IN.doublevals[0];	fitprecision = IN.doublevals[1];
	Jscale = IN.doublevals[2];

	return(0);
}

int BoltzmannME::Lagrange_siman( MASSB &P ) {
	char str[80];
	P.T.seconds = time(NULL);
	P.lfile << "Beginning Lagrange_siman: " << ctime(&P.T.seconds) << endl;
	cout << "Beginning Lagrange_siman:\n";

//	Gnuplot gsb = Gnuplot("impulses");

	const gsl_rng_type * Tr;
	gsl_rng * r;
     
	gsl_rng_env_setup();
     
//	Tr = gsl_rng_default;
	Tr = gsl_rng_taus2;
	r = gsl_rng_alloc(Tr);

//	get a reasonably random seed -- the current process ID
	pid_t pid;
	if ((pid = getpid()) < 0) {
		perror("unable to get pid\n");
	} else {
		sprintf(str,"The process id (i.e. random number seed) is %d\n", pid);
		P.lfile << str;
		cout << str;
	}

	int num = int(pid);
	int rev = 0;
	int digits = log10(num)+1;
	for (int i=0; i< digits; i++) {
		rev *= 10;
		rev += (num % 10);
		num /= 10;
	}
	sprintf(str,"reversed id (i.e. more random seed) is %d\n", rev);
        P.lfile << str;
        cout << str;

	gsl_rng_set(r, rev );

	// THIS TO CHANGE AFTER STATE FUNCTIONALITY INCLUDED IN MASSB     
//	gsl_vector_set_all( P.lagrange,0.1);
	char app[3];
	sprintf(app,".%d",P.etat);

	ofstream saout;	
	char saname[179];
	strcpy(saname,P.outpath);
	strcat(saname,P.outbase);
	strcat(saname,"/");
	strcat(saname,P.outbase);
	strcat(saname,".sa");	
	strcat(saname,app);	
	saout.open(saname, ios::out);

	char format[79];
	strcpy(format,"%6d %9ld %12g %10g %8.5f %4.1f %4.1f %4.1f ");
	sprintf(str,"#%5s %9s %12s %10s %8s %4s %4s %4s\n",
		"iter","evals","temp","chi2","H ","#lss","#acc","$rej");
	saout << str;

	MASSB x(P), new_x(P), best_x(P);  

	double E, new_E, best_E;
	double H, new_H, best_H;
	int i, done;
	double T = inittemp;
	double u;
	int n_evals = 0, n_iter = 0, n_accepts, n_rejects, n_eless;

	E = P.calc_state();
	H = P.entropy;
	best_E = E;
	best_H = H;
	

	sprintf(str,"Initial Chi2 Energy: %12.6g Inital Entropy: %12.6g\n",E,H); 
	P.lfile << str << flush;

	sprintf(str,"%5d B %8g S %8g Temp: %8g-%8g by %6.4f\n",
		iterT,boltzmann,maxstep,T,mintemp,attentemp);
	P.lfile << str;
	cout << str;
	
	done = 0;

	while (!done) {
		n_accepts = 0;
		n_rejects = 0;
		n_eless = 0;
    	for (i = 0; i < iterT; ++i) {
			new_x.copy_min(x);
	
			// take a step
			for (int lm=0; lm<P.nsb; lm++) {
				u = gsl_rng_uniform(r);
				gsl_vector_set( new_x.lagrange, lm, u * 2.0 * maxstep - 
					maxstep + gvg( x.lagrange, lm) );
			}
			new_E = new_x.calc_state();
			new_H = new_x.entropy;
/*
			sprintf(str,"E %12.6g L: ",new_E); cout << str;
			for (int lm=0; lm<P.nsb; lm++) {
				sprintf(str,"%12.6g ",gvg(new_x.lagrange,lm));
				cout << str;
			}
			cout << endl;
*/

			// if we fell over the edge, report it
			if ( !isfinite(new_E) ) {
				for (int bs = 0; bs<62; bs++ ) cout << "\b";
				cout << "BANG. Chi2 is not finite at temp " << T ;
				++n_rejects;
				continue;
			}
		
			// if energy of configuration following step is the lowest yet, 
			// replace best known configuration with it
			if(new_E <= best_E){
				best_x.copy_min(new_x);
	        	best_E=new_E;
				best_H=new_H;
				P.copy_min(new_x);
			}
			// keep track of finite Ef() evaluations 
			++n_evals;

			// now take the crucial step: see if the new point is accepted
			// or not, as determined by the boltzman probability. If step is
			// is taken, write it to the .la file together with the 
			// Lagrange multiplier configuration.
			// if the energy of the step is lower than the current config
			// take the step. Also take a higher energy step, with a random
			// probability.

			if (new_E < E) {
	    	   	// take a step 
				x.copy_min(new_x);
				E = new_E;
				H = new_H;
				++n_eless;
			} else if (gsl_rng_uniform(r) < exp (-(new_E-E)/(boltzmann*T)) ) {
				// also take a step 
				x.copy_min(new_x);
				E = new_E;
				H = new_H;
				++n_accepts;
			} else {
				++n_rejects;
			}
    	}
	
		//print report
		sprintf(str,format,n_iter,n_evals,T,x.chi2,x.entropy, 
			100.0*double(n_eless)/double(iterT), 
			100.0*double(n_accepts)/double(iterT), 
			100.0*double(n_rejects)/double(iterT) );
		saout << str;
		for (int i=0; i<P.nsb; i++) {
			sprintf(str,"%10g ", gvg(x.lagrange,i) );
			saout << str;
		}
		saout << endl;
	
		// Some updates to screen
		if ( double(n_iter)/5.0 - double(n_iter/5) < 1E-6 ) {
			for (int bs = 0; bs<96; bs++ ) cout << "\b";
			printf("Temp %10.8f Current/Best E %14.6g/%14.6g Current/Best H %8.5f/%8.5f",T,E,best_E,H,best_H);
			P.write_out();
			P.update_visual();
		}
	
	    // apply the cooling schedule to the temperature 
		T /= attentemp;
		++n_iter;
		if (T < mintemp) done = 1;
	}

	//at the end, copy the result onto the initial point
	P.copy_min(best_x);

	P.calc_state();

	cout << endl;
	P.lfile << "CHI2   " <<	P.chi2 << endl;
	P.lfile << "AVGDEV " <<  P.avgdev << endl;
	P.lfile << "ENTROPY " << P.entropy << endl;

	cout << "CHI2   " << 	P.chi2 << endl;
	cout << "AVGDEV " <<  P.avgdev << endl;
	cout << "ENTROPY " << P.entropy << endl;

	saout << flush;
	saout.close();
	P.write_out();

	P.T.seconds = time(NULL);
	P.lfile << "Ending  Lagrange_siman: " << ctime(&P.T.seconds) << endl;
	cout << "Ending  Lagrange_siman: " << ctime(&P.T.seconds) << endl;
	
	gsl_rng_free(r);

	P.janneal = 1.0;
	P.jscale = 1.0;
	P.log_jacobian();
	double mx = gsl_matrix_max(P.J);
	size_t row,col;
	gsl_matrix_max_index(P.J,&row,&col);
	sprintf(str,"Jacobian max = %9g at row %d col %d\n",mx,row,col);
	P.lfile << str;

	P.write_state();
	return(0);
}

int BoltzmannME::Lagrange_tabu( MASSB &P ) {
	char str[80];
	P.T.seconds = time(NULL);
	P.lfile << "Beginning Lagrange_tabu: " << ctime(&P.T.seconds) << endl;
	cout << "Beginning Lagrange_tabu:\n";

	// CHECK THAT NEIGHBOURHOOD SIZE IS <= NSB

//	Gnuplot gsb = Gnuplot("impulses");

	// Random number generator and seed
	const gsl_rng_type * Tr;
	gsl_rng * r;
     
	gsl_rng_env_setup();
     
//	Tr = gsl_rng_default;
	Tr = gsl_rng_taus2;
	r = gsl_rng_alloc(Tr);

//	get a reasonably random seed -- the current process ID
	pid_t pid;
	if ((pid = getpid()) < 0) {
		perror("unable to get pid\n");
	} else {
		sprintf(str,"The process id (i.e. random number seed) is %d\n", pid);
		P.lfile << str;
		cout << str;
	}

	int num = int(pid);
	int rev = 0;
	int digits = log10(num)+1;
	for (int i=0; i< digits; i++) {
		rev *= 10;
		rev += (num % 10);
		num /= 10;
	}
	sprintf(str,"reversed id (i.e. more random seed) is %d\n", rev);
        P.lfile << str;
        cout << str;

	gsl_rng_set(r, rev );

	char app[3];
	sprintf(app,".%d",P.etat);

	ofstream tbout;	
	char tbname[179];
	strcpy(tbname,P.outpath);
	strcat(tbname,P.outbase);
	strcat(tbname,"/");
	strcat(tbname,P.outbase);
	strcat(tbname,".tb");	
	strcat(tbname,app);	
	tbout.open(tbname, ios::out);

	// ADAPT THIS
//	char format[79];
//	strcpy(format,"%6d %9ld %12g %10g %8.5f %4.1f %4.1f %4.1f ");
//	sprintf(str,"#%5s %9s %12s %10s %8s %4s %4s %4s\n",
//		"iter","evals","temp","chi2","H ","#lss","#acc","$rej");
//	saout << str;

	MASSB x(P), mem_x(P), test_x(P), best_x(P);  
	double E, test_E, best_E, initial_E;
	double H, test_H, best_H;

	E = P.calc_state();
	H = P.entropy;
	best_E = E;
	initial_E = E;
	best_H = H;

	sprintf(str,"Initial Chi2 Energy: %12.6g Inital Entropy: %12.6g\n",E,H); 
	P.lfile << str << flush;

//	sprintf(str,"%5d B %8g S %8g Temp: %8g-%8g by %6.4f\n",
//		iterT,boltzmann,maxstep,T,mintemp,attentemp);
//	P.lfile << str;
//	cout << str;

	int * tabuleft;
	tabuleft = new int[P.nsb];
	for (int s=0; s<P.nsb; s++) tabuleft[s] = 0;

/* TABU search: user specified parameters: 
	stepsize 	tenure 	neighbourhoodsize 	tabusteps
*/

	int *randomindices = new int[neighbourhoodsize];
	double optEthisiter;
	int best_iter = -1;
	int best_this_iter;
	
	// for each of the user-specified number of steps
    for (int i=0; i < tabusteps; i++) {
		// reset all randomindices to clear previous set. 
		for (int n=0; n < neighbourhoodsize; n++) randomindices[n] = -1;

		// vector of random lagrange multiplier indices to change this iteration
		for (int n=0; n < neighbourhoodsize; n++) {
			int u = int( gsl_rng_uniform(r) * double(P.nsb) );

			// Check that sideband for corresponding Lagrange multiplier random
			// index is not greater than max_sb
			if ( abs(gvg(P.sideband,u)) > max_sb ) {
				n--;
				continue;
			}

			// check that proposed index isn't already slated for change.
			// do over if it is.
			bool OK = 1;
			for (int m=0; m<n; m++) {
				if ( randomindices[m] == u ) {
					n--;
					OK = 0;
					break;
				}
			}
			if ( OK == 1 ) randomindices[n] = u;
		}

		sprintf(str,"randomindices: ");
		P.lfile << str;
		for (int rr=0; rr<neighbourhoodsize; rr++) {
			sprintf(str,"%d ", randomindices[rr]);
			P.lfile << str;
		}
		P.lfile << endl;
		

		// log tabu count, selected indices
		sprintf(str,"Iter %02d/%03d. index(tabuleft)+yes/-not: ",i,tabusteps);
		P.lfile << str;
		for ( int lm=0; lm<P.nsb; lm++ ) {
			bool selected = 0;
			for ( int n=0; n < neighbourhoodsize; n++) {
				if (randomindices[n] == lm) selected = 1;
			}
			if (selected == 0) sprintf(str,"%02d(%02d)- ",lm,tabuleft[lm] );
			else sprintf(str,"%02d(%02d)+ ",lm,tabuleft[lm] );
			P.lfile << str;
		}
		P.lfile << endl;
		

		// for each lagrange multiplier randomly selected for change (above).
		// if index is not tabu, add or subtract step size and test. mem_x.lagrange
		// holds the best configuration from one iter to the next. This config
		// is copied into copy x at the beginning of the loop, and the changes
		// are made to copy test_x. The best of the neighbourhoodsize
		// test configurations are kept, and copied into mem_x.lagrange for the
		// next iteration. Of course if a configuration is identified which
		// is the best yet, remember it for ultimate output from this search 
		// algorithm

		gsl_vector_memcpy(x.lagrange,mem_x.lagrange);
		optEthisiter = -1;
		for (int n=0; n < neighbourhoodsize; n++) {
	
			// start at the same point for each try in this neighbourhood	
			gsl_vector_memcpy(test_x.lagrange,x.lagrange);

			// if index is tabu, skip. Decrement will be done at end
			if (tabuleft[randomindices[n]] > 0 ) continue;

			//  select whether to add or subtract stepsize			
			double rndm = gsl_rng_uniform(r);
			double step = stepsize;
			if ( rndm < 0.5 ) step *= -1.0;

			// set new test configuration	
			gsl_vector_set( test_x.lagrange, randomindices[n], 
				step + gvg( x.lagrange, randomindices[n]) );

			// calc and assess test configuration
			test_E = test_x.calc_state();
			test_H = test_x.entropy;
			if (optEthisiter < 0 || test_E < optEthisiter ) {
				gsl_vector_memcpy(mem_x.lagrange,test_x.lagrange);
				optEthisiter = test_E;
				best_this_iter = n;
			}

			// log configuration
			sprintf(str,"%2d E %12.6g H %12.6g    L: ",n,test_E,test_H);
			P.lfile << str;
			for ( int lm=0; lm<P.nsb; lm++ ) {
				sprintf(str,"%7.2f ",gvg(test_x.lagrange,lm));
				P.lfile << str;
			}
			P.lfile << endl;

			if ( test_E < best_E ) {
				gsl_vector_memcpy(best_x.lagrange,test_x.lagrange);
				best_E = test_E;
				best_iter = i;
			}
		}
	
		// log
		sprintf(str,"best this iter: #%d. index %d tabu for %d iters\n",
			best_this_iter,randomindices[best_this_iter],tenure);
		P.lfile << str;

		// decrement the tabu count for any positions which are > 0	
		for ( int lm=0; lm<P.nsb; lm++ ) if ( tabuleft[lm] > 0 ) tabuleft[lm]--;

		// set the position which was optimum to tabu for the next tenure rounds
		tabuleft[randomindices[best_this_iter]] = tenure;

		P.lfile << endl;

	}

	gsl_rng_free(r);
	delete randomindices;
	delete tabuleft;
	// if the scheme hasn't worked, try again up to the specified number of restarts
	if ( abs(initial_E - best_E) < 1E-12 ) {
		cerr << "Tabu did not find a better configuration than the initial one"
			<< endl;	
		P.lfile 
			<< "Tabu did not find a better configuration than the initial one"
			<< endl;
		return(1);
	}

	// Carry best configuration found through the above forward
	P.copy_min(best_x);
	P.calc_state();

	cout << endl;
	P.lfile << "best configuration found at " << best_iter << " / restart " << tabu_restarts << endl;
	P.lfile << "CHI2   " <<	P.chi2 << endl;
	P.lfile << "AVGDEV " <<  P.avgdev << endl;
	P.lfile << "ENTROPY " << P.entropy << endl;

	cout << "best configuration found at " << best_iter << " / restart " << tabu_restarts <<  endl;
	cout << "CHI2   " << 	P.chi2 << endl;
	cout << "AVGDEV " <<  P.avgdev << endl;
	cout << "ENTROPY " << P.entropy << endl;

	tbout << flush;
	tbout.close();
	P.write_out();

	P.T.seconds = time(NULL);
	P.lfile << "Ending  Lagrange_tabu: " << ctime(&P.T.seconds) << endl;
	cout << "Ending  Lagrange_tabu: " << ctime(&P.T.seconds) << endl;
	
	P.janneal = 1.0;
	P.jscale = 1.0;
	P.log_jacobian();
	double mx = gsl_matrix_max(P.J);
	size_t row,col;
	gsl_matrix_max_index(P.J,&row,&col);
	sprintf(str,"Jacobian max = %9g at row %d col %d\n",mx,row,col);
	P.lfile << str;

	P.write_state();
	P.update_visual();

	return(0);
}


int BoltzmannME::Lagrange_newton( MASSB &P ) {
	P.check = 0 ;
	
	P.janneal = Janneal;
	P.jscale = Jscale;
	
	char app[3];
	sprintf(app,".%d",P.etat);

	ofstream nwtnout;	
	char nwtnname[179];
	strcpy(nwtnname,P.outpath);
	strcat(nwtnname,P.outbase);
	strcat(nwtnname,"/");
	strcat(nwtnname,P.outbase);
	strcat(nwtnname,".nwtn");	
	strcat(nwtnname,app);	
	nwtnout.open(nwtnname, ios::out);
	
	char str[80];
	char format[79];
	strcpy(format,"%6d %10.4g %10.4g %10.4g %8.5f ");
	sprintf(str,"#%5s %10s %10s %10s %8s %24s",
		"iter","janneal","Z","chi2","H ","lagrange multipliers ...");
	nwtnout << str << endl;

	const gsl_multifit_fdfsolver_type *T;
	T = gsl_multifit_fdfsolver_lmsder;
	gsl_multifit_fdfsolver *s;
	s = gsl_multifit_fdfsolver_alloc(T,P.nsb,P.nsb);

	int status;
	size_t i, iter=0;

	gsl_matrix *covar = gsl_matrix_alloc(P.nsb,P.nsb);
	gsl_multifit_function_fdf f;
	f.f = _nonlinfit_f;
	f.df = _nonlinfit_df;
	f.fdf = _nonlinfit_fdf;
	f.n = P.nsb;
	f.p = P.nsb;
	f.params = &P;

	gsl_vector * x = gsl_vector_calloc(P.nsb);
	gsl_vector_memcpy(x,P.lagrange);
	gsl_multifit_fdfsolver_set(s,&f,x);

	P.calc_state();	
	sprintf(str,format,-1,P.janneal,P.Z,P.chi2,P.entropy);
	nwtnout << str;
	P.lfile << str << endl;
	for (int lambda=0; lambda<P.nsb; lambda++) {
		sprintf(str,"%12.5g ", gvg(P.lagrange,lambda) );
		nwtnout << str;
	}
	nwtnout << endl;

	if ( abs(P.janneal) < 1E-12 ) {
		P.janneal = P.Z;
		Janneal = P.Z;
	}
		

	status = gsl_multifit_fdfsolver_iterate(s);

	gsl_vector_memcpy(P.lagrange,s->x);
	P.calc_state();	
	P.write_out();
//	P.update_visual();
		
	sprintf(str,format,iter,P.janneal,P.Z,P.chi2,P.entropy);
	nwtnout << str;
	P.lfile << str;
	for (int lambda=0; lambda<P.nsb; lambda++) {
		sprintf(str,"%12.5g ", gvg(P.lagrange,lambda) );
		nwtnout << str;
	}
	nwtnout << endl;
	
	double atten = pow(P.janneal,1.0/double(Jsteps) );
	P.lfile << "Janneal atten = " << atten << endl;
	do {
		iter++;
		
		status = gsl_multifit_fdfsolver_iterate(s);

		gsl_vector_memcpy(P.lagrange,s->x);
		P.calc_state();	
		
//		for (int bs = 0; bs<57; bs++ ) cout << "\b";
		printf("%5d Z: %10.6g Janneal: %10.6g chi2: %12.6g H: %8.5f\n",iter,P.janneal,P.Z,P.chi2,P.entropy);
		sprintf(str,format,iter,P.janneal,P.Z,P.chi2,P.entropy);
		nwtnout << str;
		P.lfile << str;
		for (int lambda=0; lambda<P.nsb; lambda++) {
			sprintf(str,"%12.5g ", gvg(P.lagrange,lambda) );
			nwtnout << str;
		}
		nwtnout << endl;
		P.write_out();
//		P.update_visual();

		if (iter <= miniter) P.janneal /= atten;

//		if (status) {
			sprintf (str,"iter  status: %s\n", gsl_strerror (status));
			P.lfile << str << endl;
			cout << str;
//		}
//		if (status && iter > miniter) break;

		double lx = gsl_blas_dnrm2(s->f)/double(P.nsb);
		double ldx = gsl_blas_dnrm2(s->dx);

		status = gsl_multifit_test_delta(s->dx,s->x,fitprecision,fitprecision);
		sprintf (str,"(|dx|=%9.3g) <? epsabs + epsrel * (|f|/nsb = %9.3g) = %9.3g\n", 
			ldx, lx, fitprecision + fitprecision * lx);
		P.lfile << str;
		cout << str;

	} while ( (status==GSL_CONTINUE && iter < maxiter ) || iter <= miniter );

	gsl_vector_memcpy(P.lagrange,s->x);
	P.calc_state();	
	P.write_out();

	cout << endl;
	P.lfile << "CHI2   " <<	P.chi2 << endl;
	P.lfile << "AVGDEV " <<  P.avgdev << endl;
	P.lfile << "ENTROPY " << P.entropy << endl;

	cout << "CHI2   " << 	P.chi2 << endl;
	cout << "AVGDEV " <<  P.avgdev << endl;
	cout << "ENTROPY " << P.entropy << endl;

	gsl_multifit_fdfsolver_free(s);

	nwtnout.close();
	
	P.janneal = 1.0;
	P.jscale = 1.0;
	P.log_jacobian();
	double mx = gsl_matrix_max(P.J);
	size_t row,col;
	gsl_matrix_max_index(P.J,&row,&col);
	sprintf(str,"Jacobian max = %9g at row %d col %d\n",mx,row,col);
	P.lfile << str;
	
	P.write_state();
	
	P.update_visual();

	return(0);
}

/*
int BoltzmannME::MC_error ( MASSB &goodMASSB  ) {
	goodMASSB.T.seconds = time(NULL);
	goodMASSB.lfile << "Beginning Boltzmann MC: " << ctime(&goodMASSB.T.seconds) << endl;
	cout << "Beginning Boltzmann MC:\n";
	
	gsl_vector * distsum = gsl_vector_calloc( goodMASSB.np );
	gsl_vector * distsumsq = gsl_vector_calloc( goodMASSB.np );

	gsl_vector * blockavg = gsl_vector_calloc( goodMASSB.np );
	gsl_vector * blockstdev = gsl_vector_calloc( goodMASSB.np );

	gsl_vector * bl_avgsum = gsl_vector_calloc( goodMASSB.np );
	gsl_vector * bl_avgsumsq = gsl_vector_calloc( goodMASSB.np );
	gsl_vector * bl_stdevsum = gsl_vector_calloc( goodMASSB.np );
	gsl_vector * bl_stdevsumsq = gsl_vector_calloc( goodMASSB.np );

	gsl_vector * avgofavg = gsl_vector_calloc( goodMASSB.np );
	gsl_vector * stdevofavg = gsl_vector_calloc( goodMASSB.np );
	gsl_vector * avgofstdev = gsl_vector_calloc( goodMASSB.np );
	gsl_vector * stdevofstdev = gsl_vector_calloc( goodMASSB.np );

	gsl_vector * fracstdevofstdev = gsl_vector_calloc( goodMASSB.np );

	gsl_vector * workspace = gsl_vector_calloc( goodMASSB.np );

	const gsl_rng_type * T;
	gsl_rng * r;
     
	gsl_rng_env_setup();
     
	T = gsl_rng_default;
	r = gsl_rng_alloc(T);
    
	MASSB E(goodMASSB);

	int numblocks = 0;
	int numrejected = 0;
	double dataSSo, max_sofs, grand;
	while ( numblocks < E.maxmcblocks ) {
		++numblocks;

		gsl_vector_set_zero(distsum);
		gsl_vector_set_zero(distsumsq);
	
		for (int bl=0; bl<E.permcblock; bl++) {
			// Deep-copy Paar structure for error analysis
			E.copy_min(goodMASSB);

			// add grandom error value to each data point with appropriate
			// sigma
			for ( int d=0; d<E.nsb; d++ ) {
				dataSSo = gvg( E.sidebandamp, d );	
				grand = _grandom_sigma( r, gvg(E.sidebanderr,d));
				gsl_vector_set( E.sidebandamp, d, dataSSo + grand);
//cout << "sigma: " << gvg(E.error,d) << " old: " << dataSSo << 
//	" grand: " << grand << " new: " << gvg(E.redor,d) << endl;
			}

			// run siman solver on noise-adjusted data, no printing
			BoltzmannME Fit(E);
			Fit.Lagrange_siman_MC(E);
//			E.calc_state();

//FIXME: rejection by rmsd selects for lower noise configurations. Should
//	rerun siman with slightly adjusted parameters so that distributions
//	are not biased
			goodMASSB.lfile << numblocks << "." << bl << "Err CHI2 " 
				<< E.chi2 << " AVGDEV " <<  E.avgdev;
			cout << numblocks << "." << bl << "Err CHI2 " <<	
				E.chi2 << " AVGDEV " <<  E.avgdev;
		
			if ( E.chi2 > E.maxmcrmsd ) {
				numrejected += 1;
				goodMASSB.lfile << " REJECTED" << endl;
				cout << " REJECTED" << endl;
				bl--;
				continue;
			} 
			cout << endl;
			goodMASSB.lfile << endl;
		
			// update running sums and sum of squares for this block	
			gsl_vector_add(distsum,E.population);
			_gv_pow(E.population,2.0)	;
			gsl_vector_add(distsumsq,E.population);	
		}

		// average density value for the block above
		gsl_vector_memcpy( blockavg, distsum );
		gsl_vector_scale(blockavg,1.0/E.permcblock);
		//stdev of density values for the block above
		_gv_stdev(blockstdev,distsum,distsumsq,E.permcblock,workspace);

		// update running sums and sum of squares over all blocks thus far
		gsl_vector_add(bl_avgsum,blockavg);
		gsl_vector_add(bl_stdevsum,blockstdev);
		_gv_pow(blockavg,2.0);
		_gv_pow(blockstdev,2.0);
		gsl_vector_add(bl_avgsumsq,blockavg);
		gsl_vector_add(bl_stdevsumsq,blockstdev);

		if ( numblocks < 5 ) continue;

		// calculate stdev of stdev ...
		_gv_stdev(stdevofstdev,bl_stdevsum,bl_stdevsumsq,numblocks,workspace);

		// ... and test whether or not we're done
	gsl_vector_memcpy(	fracstdevofstdev, stdevofstdev);
	gsl_vector_div(fracstdevofstdev, avgofstdev);
		max_sofs = gsl_vector_max( stdevofstdev );	
		goodMASSB.lfile << "BLOCK " << 
			numblocks << " max sigma err " << max_sofs << endl; 
		goodMASSB.lfile << "rejection rate: " << 
			double(numrejected)/
				(double(numblocks)*double(E.permcblock)+double(numrejected)) 
				<< endl;
		cout << "BLOCK " << numblocks << " max sigma err " << max_sofs << endl; 
		cout << "rejection rate: " << 
			double(numrejected)/
				(double(numblocks)*double(E.permcblock)+double(numrejected)) 
				<< endl;
	// average at each density point (should be same as data)
	gsl_vector_memcpy(avgofavg,bl_avgsum);
	gsl_vector_scale(avgofavg,1.0/numblocks);

	// error in average
	_gv_stdev(stdevofavg,bl_avgsum,bl_avgsumsq,numblocks,workspace);
	
	// average at each density point (should be same as data)
	gsl_vector_memcpy(avgofstdev,bl_stdevsum);
	gsl_vector_scale(avgofstdev,1.0/numblocks);

	// should be same as above, but just in case... error in stdev
	_gv_stdev(stdevofstdev,bl_stdevsum,bl_stdevsumsq,numblocks,workspace);

	gsl_vector_memcpy( goodMASSB.mcavg, avgofavg);
	gsl_vector_memcpy( goodMASSB.mcavgerr, stdevofavg);
	gsl_vector_memcpy( goodMASSB.mcstdev, avgofstdev);
	gsl_vector_memcpy( goodMASSB.mcstdeverr, stdevofstdev);
		goodMASSB.write_out();
//		goodMASSB.gnuplot_out();
	
		if ( max_sofs <= E.minstdev2 ) break;	
	}

	if (numblocks < goodMASSB.maxmcblocks) {
		goodMASSB.lfile 
			<< "Distributions at each discrete distance converged in " <<
			numblocks << " x " << goodMASSB.permcblock << " mc points " << endl;
		cout << "Distributions at each discrete distance converged in " <<
			numblocks << " x " << goodMASSB.permcblock << " mc points " << endl;
	} else {
		goodMASSB.lfile 
			<< "Maximum monte carlo steps reached. Largest stdev is "<<
			max_sofs << endl;
		cout << "Maximum monte carlo steps reached. Largest stdev is "<<
			max_sofs << endl;
	}

	goodMASSB.lfile << "number rejected: " << numrejected << " of total " 
		<< numblocks*goodMASSB.permcblock << endl;
	cout << "number rejected: " << numrejected << " of total " 
		<< numblocks*goodMASSB.permcblock << endl;

	// calculate stats for distribution and set in original Paar structure

	// average at each density point (should be same as data)
	gsl_vector_memcpy(avgofavg,bl_avgsum);
	gsl_vector_scale(avgofavg,1.0/numblocks);

	// error in average
	_gv_stdev(stdevofavg,bl_avgsum,bl_avgsumsq,numblocks,workspace);
	
	// average at each density point (should be same as data)
	gsl_vector_memcpy(avgofstdev,bl_stdevsum);
	gsl_vector_scale(avgofstdev,1.0/numblocks);

	// should be same as above, but just in case... error in stdev
	_gv_stdev(stdevofstdev,bl_stdevsum,bl_stdevsumsq,numblocks,workspace);

	gsl_vector_memcpy( goodMASSB.mcavg, avgofavg);
	gsl_vector_memcpy( goodMASSB.mcavgerr, stdevofavg);
	gsl_vector_memcpy( goodMASSB.mcstdev, avgofstdev);
	gsl_vector_memcpy( goodMASSB.mcstdeverr, stdevofstdev);

	gsl_vector_free(distsum);
	gsl_vector_free(distsumsq);
	gsl_vector_free(blockavg);
	gsl_vector_free(blockstdev);
	gsl_vector_free(bl_avgsum);
	gsl_vector_free(bl_avgsumsq);
	gsl_vector_free(bl_stdevsum);
	gsl_vector_free(bl_stdevsumsq);
	gsl_vector_free(avgofavg);
	gsl_vector_free(stdevofavg);
	gsl_vector_free(avgofstdev);
	gsl_vector_free(stdevofstdev);
	gsl_vector_free(workspace);
	return 0;

}
*/

int BoltzmannME::Lagrange_siman_MC( MASSB &P ) {
//FIXME: adapt this for use in MC: use MC simulated annealing parameters,
// and minimize report output.
	const gsl_rng_type * Tr;
	gsl_rng * r;
     
	gsl_rng_env_setup();
     
//	Tr = gsl_rng_default;
	Tr = gsl_rng_taus2;
	r = gsl_rng_alloc(Tr);
     
	MASSB x(P), new_x(P), best_x(P);  
	double E, new_E, best_E;
	int i, done;
	double T, u;
	int n_evals = 1, n_iter = 0, n_accepts, n_rejects, n_eless;

	E = P.calc_state();
	best_E = E;

	T = errinittemp;
	done = 0;

	while (!done) {
		n_accepts = 0;
		n_rejects = 0;
		n_eless = 0;
    	for (i = 0; i < erriterT; ++i) {
			new_x.copy_min(x);
			// take a step
			for (int lm=0; lm<P.nsb; lm++) {
				u = gsl_rng_uniform(r);
				gsl_vector_set( new_x.lagrange, lm, u * 2.0 * errmaxstep - 
					errmaxstep + gvg( new_x.lagrange, lm) );
			}
			new_E = new_x.calc_state();
			if ( !isfinite(new_E) ) {
				for (int bs = 0; bs<62; bs++ ) cout << "\b";
				cout << "BANG. MC Chi2 is not finite at temp " << T 
					<< " avgdev: " << new_x.avgdev;
				for (int lg=0; lg<new_x.nsb; lg++) 
					cout << gvg( new_x.lagrange, lg ) << endl;
				exit(0);
			}
			if(new_E <= best_E){
				best_x.copy_min(new_x);
	        	best_E=new_E;
			}
			// keep track of Ef() evaluations 
			++n_evals;
			//now take the crucial step: see if the new point is accepted
			//or not, as determined by the boltzman probability 
			if (new_E < E) {
	        	// take a step 
				x.copy_min(new_x);
				E = new_E;
				++n_eless;
			} else if (gsl_rng_uniform(r) < 
				exp (-(new_E-E)/(errboltzmann*T)) ) {
				// also take a step */
				x.copy_min(new_x);
				E = new_E;
				++n_accepts;
			} else {
				++n_rejects;
			}
    	}

    	// apply the cooling schedule to the temperature 
		T /= errattentemp;
		++n_iter;
		if (T < errmintemp) done = 1;
	}

	//at the end, copy the result onto the initial point
	P.copy_min(best_x);
	P.calc_state();

	return 0;
}

/*
double MaximumEntropy_siman (void *);
int search_scale_me_siman( void *);
double wrapper_scale_me_siman(double , void *);
void Dstruct ( void * );
void *Cstruct ( void * );
void *Cstruct3 ( void * );
void Cpystruct (void *, void *);
void P1(void *xp);
void S1(const gsl_rng *, void *, double);
double M1(void *, void *);

double error_siman ( void *  );


double error_siman ( void * p ) {

	int np = ((struct Paar *) p)->np;
	int nc = ((struct Paar *) p)->nc;

	gsl_vector * distsum = gsl_vector_calloc( np );
	gsl_vector * distsumsq = gsl_vector_calloc( np );

	gsl_vector * blockavg = gsl_vector_calloc( np );
	gsl_vector * blockstdev = gsl_vector_calloc( np );

	gsl_vector * bl_avgsum = gsl_vector_calloc( np );
	gsl_vector * bl_avgsumsq = gsl_vector_calloc( np );
	gsl_vector * bl_stdevsum = gsl_vector_calloc( np );
	gsl_vector * bl_stdevsumsq = gsl_vector_calloc( np );

	gsl_vector * avgofavg = gsl_vector_calloc( np );
	gsl_vector * stdevofavg = gsl_vector_calloc( np );
	gsl_vector * avgofstdev = gsl_vector_calloc( np );
	gsl_vector * stdevofstdev = gsl_vector_calloc( np );

	gsl_vector * workspace = gsl_vector_calloc( np );

	const gsl_rng_type * T;
	gsl_rng * r;
     
	gsl_rng_env_setup();
     
	T = gsl_rng_default;
	r = gsl_rng_alloc(T);
    
	// User decides what's appropriate here, as we're probably pretty
	// close already. 
	gsl_siman_params_t params = {ntries, erriterT, errmaxstep, errboltzmann, 
		errinittemp, errattentemp, errmintemp}; 

	struct Paar E;
	E.cycles = gsl_vector_alloc(nc);
	E.Snot = gsl_vector_alloc(nc);
	E.S = gsl_vector_alloc(nc);
	E.redor = gsl_vector_alloc(nc);
	E.SoverSnot = gsl_vector_alloc(nc);
	E.avg = gsl_vector_alloc(nc);
	E.error = gsl_vector_alloc(nc);
	E.sqrerr = gsl_vector_alloc(nc);
	E.stdev = gsl_vector_alloc(nc);
	E.residualy = gsl_vector_alloc(nc);
	E.density = gsl_vector_alloc(np);
	E.dephase_by_cycles = gsl_matrix_alloc(nc,np);
	E.lagrange = gsl_vector_alloc(nc);
	E.mcavg = gsl_vector_calloc(np);
	E.mcavgerr = gsl_vector_calloc(np);
	E.mcstdev = gsl_vector_calloc(np);
	E.mcstdeverr = gsl_vector_calloc(np);
	E.adjust = gsl_vector_calloc(nc);

	int numblocks = 0;
	int numrejected = 0;
	double dataSSo, max_sofs;
	while ( numblocks < maxmcblocks ) {
		++numblocks;

		gsl_vector_set_zero(distsum);
		gsl_vector_set_zero(distsumsq);
	
		for (int bl=0; bl<permcblock; bl++) {
			// Deep-copy Paar structure for error analysis
			deep_copy_Paar(p, &E);

			// add grandom error value to each data point with appropriate
			// sigma
			for ( int d=0; d<E.nc; d++ ) {
				dataSSo = gvg( E.redor, d );	
				gsl_vector_set( E.redor, d, 
					dataSSo + grandom_sigma( r, gvg(E.error,d)));
	// average at each density point (should be same as data)
	gsl_vector_memcpy(avgofavg,bl_avgsum);
	gsl_vector_scale(avgofavg,1.0/numblocks);

	// error in average
	_gv_stdev(stdevofavg,bl_avgsum,bl_avgsumsq,numblocks,workspace);
	
	// average at each density point (should be same as data)
	gsl_vector_memcpy(avgofstdev,bl_stdevsum);
	gsl_vector_scale(avgofstdev,1.0/numblocks);

	// should be same as above, but just in case... error in stdev
	_gv_stdev(stdevofstdev,bl_stdevsum,bl_stdevsumsq,numblocks,workspace);

	gsl_vector_memcpy( goodPaar.mcavg, avgofavg);
	gsl_vector_memcpy( goodPaar.mcavgerr, stdevofavg);
	gsl_vector_memcpy( goodPaar.mcstdev, avgofstdev);
	gsl_vector_memcpy( goodPaar.mcstdeverr, stdevofstdev);
			}

			// run siman solver on noise-adjusted data, no printing
			gsl_siman_solve(r, p, &calc_state, S1, M1, NULL, 
				Cpystruct, Cstruct, Dstruct, 0, params);
			
			calc_state(&E);

//FIXME: rejection by rmsd selects for lower noise configurations. Should
//	rerun siman with slightly adjusted parameters so that distributions
//	are not biased
			lfile << "Err CHI2 " <<	E.rmsd << " Old RMSD " << E.oldrmsd 
				<< " AVGDEV " <<  E.avgdev;
			cout << "Err CHI2 " <<	E.rmsd << " Old RMSD " << E.oldrmsd 
				<< " AVGDEV " <<  E.avgdev;
		
			if ( E.rmsd > maxmcrmsd ) {
				numrejected += 1;
				lfile << " REJECTED" << endl;
				cout << " REJECTED" << endl;
				bl--;
				continue;
			} else {
				cout << endl;
				lfile << endl;
			}
		
			// update running sums and sum of squares for this block	
			gsl_vector_add(distsum,E.density);
			gv_pow(E.density,2.0)	;
//			gsl_vector_mul(E.density,E.density);	
			gsl_vector_add(distsumsq,E.density);	
		}

		// average density value for the block above
		gsl_vector_memcpy( blockavg, distsum );
		gsl_vector_scale(blockavg,1.0/permcblock);
		//stdev of density values for the block above
		gv_stdev(blockstdev,distsum,distsumsq,permcblock,workspace);

		// update running sums and sum of squares over all blocks thus far
		gsl_vector_add(bl_avgsum,blockavg);
		gsl_vector_add(bl_stdevsum,blockstdev);
		gv_pow(blockavg,2.0);
		gv_pow(blockstdev,2.0);
		gsl_vector_add(bl_avgsumsq,blockavg);
		gsl_vector_add(bl_stdevsumsq,blockstdev);

		if ( numblocks < 2 ) continue;

		// calculate stdev of stdev ...
		gv_stdev(stdevofstdev,bl_stdevsum,bl_stdevsumsq,numblocks,workspace);

		// ... and test whether or not we're done
		max_sofs = gsl_vector_max( stdevofstdev );	
		lfile << "BLOCK " << 
			numblocks << " max sigma err " << max_sofs << endl; 
		lfile << "rejection rate: " << 
			double(numrejected)/
				(double(numblocks)*double(permcblock)+double(numrejected)) 
				<< endl;
		cout << "BLOCK " << numblocks << " max sigma err " << max_sofs << endl; 
		cout << "rejection rate: " << 
			double(numrejected)/
				(double(numblocks)*double(permcblock)+double(numrejected)) 
				<< endl;
	
		if ( max_sofs <= minstdev2 ) break;	
	}

	if (numblocks < maxmcblocks) {
		lfile << "Distributions at each discrete distance converged in " <<
			numblocks << " x " << permcblock << " mc points " << endl;
		cout << "Distributions at each discrete distance converged in " <<
			numblocks << " x " << permcblock << " mc points " << endl;
	} else {
		lfile << "Maximum monte carlo steps reached. Largest stdev is "<<
			max_sofs << endl;
		cout << "Maximum monte carlo steps reached. Largest stdev is "<<
			max_sofs << endl;
	}

	lfile << "number rejected: " << numrejected << " of total " 
		<< numblocks*permcblock << endl;
	cout << "number rejected: " << numrejected << " of total " 
		<< numblocks*permcblock << endl;

	// calculate stats for distribution and set in original Paar structure

	// average at each density point (should be same as data)
	gsl_vector_memcpy(avgofavg,bl_avgsum);
	gsl_vector_scale(avgofavg,1.0/numblocks);

	// error in average
	gv_stdev(stdevofavg,bl_avgsum,bl_avgsumsq,numblocks,workspace);
	
	// average at each density point (should be same as data)
	gsl_vector_memcpy(avgofstdev,bl_stdevsum);
	gsl_vector_scale(avgofstdev,1.0/numblocks);

	// should be same as above, but just in case... error in stdev
	gv_stdev(stdevofstdev,bl_stdevsum,bl_stdevsumsq,numblocks,workspace);

	gsl_vector_memcpy( ((struct Paar *) p)->mcavg, avgofavg);
	gsl_vector_memcpy( ((struct Paar *) p)->mcavgerr, stdevofavg);
	gsl_vector_memcpy( ((struct Paar *) p)->mcstdev, avgofstdev);
	gsl_vector_memcpy( ((struct Paar *) p)->mcstdeverr, stdevofstdev);

	gsl_vector_free(distsum);
	gsl_vector_free(distsumsq);
	gsl_vector_free(blockavg);
	gsl_vector_free(blockstdev);
	gsl_vector_free(bl_avgsum);
	gsl_vector_free(bl_avgsumsq);
	gsl_vector_free(bl_stdevsum);
	gsl_vector_free(bl_stdevsumsq);
	gsl_vector_free(avgofavg);
	gsl_vector_free(stdevofavg);
	gsl_vector_free(avgofstdev);
	gsl_vector_free(stdevofstdev);
	gsl_vector_free(workspace);
	return 0;
}
*/
int	_gv_stdev( gsl_vector *stdev, const gsl_vector *sum, 
	const gsl_vector *sumsq, int n, gsl_vector *work) {

	gsl_vector_memcpy( work, sum);
	_gv_pow(work,2.0);
	gsl_vector_scale(work,-1.0);
	gsl_vector_memcpy( stdev, sumsq);
	gsl_vector_scale( stdev, double(n) );
	gsl_vector_add( stdev,work);
	gsl_vector_scale( stdev, 1.0/(double(n)*(double(n)-1.0) ) );
	_gv_pow( stdev, 0.5);

	return(0);
}	

int _gv_pow (gsl_vector * a, const double x) {
	const size_t N = a->size;
	double v;
	for (size_t i = 0; i < N; i++) {
		v = pow( gvg(a,i), x);	
		if ( !isfinite(v) &&  -gvg(a,i) < 1E-12 ) 
			gsl_vector_set( a,i,0 );
		else gsl_vector_set( a, i, v );
		if (!isfinite(gvg(a,i) ) ) {
			cout << "NAN" << endl; 
			exit(0);
		}
	}
	return GSL_SUCCESS;
}

double _grandom_sigma ( const gsl_rng * r, double sigma) {
	double sum = 0.0;
	for (int i=0; i<12; i++) sum += gsl_rng_uniform(r);
	return (sigma*(sum-6.0));
}

/*
double wrapper_scale_me_siman(double s, void *P) {
	((struct Paar *)P)->scale = s;
	MaximumEntropy_siman(P);
//	return( ((struct Paar *)P)->rmsd + abs( ((struct Paar *)P)->avgdev ) );
	return( ((struct Paar *)P)->rmsd );
//	return( abs( ((struct Paar *)P)->avgdev ) );
}

int search_scale_me_siman(void *P) {
	int max_iter = 100, status, iter = 0;

	gsl_function F;
	F.function = &wrapper_scale_me_siman;
	F.params = P;

	double fa,fb,fm;	
	double a = scaling1; 
	double b = scaling2;
	double m = (scaling1 + scaling2)/2.0;
//cout << "DEBUG 1\n";

//	const gsl_min_fminimizer_type *T;
//	T = gsl_min_fminimizer_brent;
	const gsl_min_fminimizer_type * T = gsl_min_fminimizer_brent;
//cout << "DEBUG 1a\n";
	gsl_min_fminimizer * s = gsl_min_fminimizer_alloc(T);
//	gsl_min_fminimizer *s;
//	s = gsl_min_fminimizer_alloc(T);
//cout << "DEBUG 1b\n";
	gsl_min_fminimizer_set(s, &F, m, a, b);
//cout << "DEBUG 1c\n";
	
	lfile << "minimizing (chi2 + avgdev) with " << gsl_min_fminimizer_name(s)
		<< endl;
	cout << "minimizing (chi2 + avgdev) with " << gsl_min_fminimizer_name(s)
		<< endl;

//cout << "DEBUG 2\n";
	sprintf(str,"%5s [%9s, %9s] %9s %9s\n", "iter", "lower", "upper",
		"min", "err(est)");
	lfile << str;
	cout << str;

	sprintf(str,"%5d [%.7f, %.7f] %.7f %.7f\n",iter,a,b,m,b-a);
	lfile << str;
	cout << str;
	
//cout << "DEBUG 3\n";
	do {
		iter++;
		status = gsl_min_fminimizer_iterate(s);

		m = gsl_min_fminimizer_x_minimum(s);
		a = gsl_min_fminimizer_x_lower(s);
		b = gsl_min_fminimizer_x_upper(s);
		fm = gsl_min_fminimizer_f_minimum(s);
		fa = gsl_min_fminimizer_f_lower(s);
		fb = gsl_min_fminimizer_f_upper(s);
//cout << "DEBUG 4\n";
		
		status = gsl_min_test_interval(a,b, 0.0002, 0);
	
		if (status == GSL_SUCCESS) {
			cout << "Converged:\n";
			lfile << "Converged:\n";
		}

		sprintf(str,"%5d [%.7f, %.7f] %.7f %.7f\n",iter,a,b,m,b-a);
		lfile << str;
		cout << str;
		sprintf(str,"%5d [%.7f, %.7f] %.7f\n\n",iter,fa,fb,fm);
		lfile << str;
		cout << str;

	} while (status == GSL_CONTINUE && iter < max_iter);	

	((struct Paar *)P)->scale = m;
	calc_state(P);
	lfile << "CHI2   " <<	((struct Paar *)P)->rmsd << endl;
	lfile << "OLD RMSD   " <<	((struct Paar *)P)->oldrmsd << endl;
	lfile << "AVGDEV " <<  ((struct Paar *)P)->avgdev << endl;
	lfile << "SCALING " << ((struct Paar *)P)->scale << endl;
	cout << "CHI2   " << 	((struct Paar *)P)->rmsd << endl;
	cout << "OLD RMSD   " << 	((struct Paar *)P)->oldrmsd << endl;
	cout << "AVGDEV " <<  ((struct Paar *)P)->avgdev << endl;
	cout << "SCALING " << ((struct Paar *)P)->scale << endl;

	return status;
	
}
*/

int _nonlinfit_f (const gsl_vector * x, void *p, gsl_vector * f) {
	// this function to serve as the gsl nonlinear fitting _f function;
	// the input x shall be the trial lagrange configuration, and the
	// output f is the error-scaled vector of residuals. p is vestigal in a
	// object oriented class structure of programming. Function is essentially
	// identical to calc_state, except trial lagrange config is used instead
	// of object's.

	MASSB * P = (class MASSB *)p;

	P->check = 1;
	gsl_vector_memcpy(P->plagrange,x);
	P->copy_from_public();
	P->calc_state();
	P->copy_to_public();
	gsl_vector_memcpy(f,P->psidebandres);
	
	return GSL_SUCCESS;
	
}

int _nonlinfit_df (const gsl_vector * x, void *p, gsl_matrix * J) {
	// this function to serve as the gsl nonlinear fitting _df function;
	// the input x shall be the trial lagrange configuration, and the
	// output J is the Jacobian. p is vestigal in a
	// object oriented class structure of programming. 

	MASSB * P = (class MASSB *)p;

	gsl_vector_memcpy(P->plagrange,x);
	P->copy_from_public();
	P->calc_jacobian();
	P->copy_to_public();
	gsl_matrix_memcpy(J,P->pJ);
	
	return GSL_SUCCESS;
}

int _nonlinfit_fdf (const gsl_vector * x, void *p, 
	gsl_vector * f, gsl_matrix * J) {
	
	_nonlinfit_f(x,p,f);	
	_nonlinfit_df(x,p,J);	

	return GSL_SUCCESS;
}
