#ifndef BSME_
#define BSME_

#include <stdio.h>
#include <cstring>
#include <fstream>
#include <iomanip>
#include <cstdlib>
#include <unistd.h>
#include <cmath>
#include <ctime>
#include <iostream>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_multiroots.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_deriv.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_errno.h>

#include <gsl/gsl_math.h>
#include <gsl/gsl_diff.h>
#include <gsl/gsl_eigen.h>

#include <gsl/gsl_linalg.h>
#include <unistd.h>
//#include "gnuplot_i.hpp"

#include <gsl/gsl_siman.h>
#include "inputs.h"
#include "sidebands.h"
using namespace std;

class BoltzmannME {
public:
	double chi2;
	BoltzmannME( class MASSB& );
	~BoltzmannME();
	int Lagrange_siman( class MASSB& );
	int lagrange_search( class MASSB& );
	int MC_error ( MASSB& );
	int Lagrange_siman_MC( MASSB& );
	int Lagrange_newton( MASSB& );
	int Lagrange_tabu( MASSB& );
private:
	int import_sa( class Inputs& );
	int import_tabu( class Inputs& );
	int import_newton( class Inputs& );

	//
	// simulated annealing parameters
	double mintemp;		// temperature below which simulated annealing stops
	double inittemp;	// initial high temperature
	double maxstep;		// maximum step size for any one lagrange multiplier
	double boltzmann;	// the boltzmann constant used in determination of
						// step selection
	double attentemp;	// the factor used to attenuate the temp between steps
	int iterT;			// the number of random steps taken at each temperature

	//
	// monte carlo/simulated annealing parameters and MC statistical structures
	int erriterT, permcblock, maxmcblocks;
	double errmaxstep, errboltzmann, errinittemp, errattentemp; 
	double errmintemp, minstdev2, maxmcrmsd;

	// MC error statistics
    gsl_vector *mcavg;
    gsl_vector *mcavgerr;
    gsl_vector *mcstdev;
    gsl_vector *mcstdeverr;

	// Newtonian search parameters
	double Janneal;
	double Jscale;
	int Jsteps;
	double fitprecision;
	int maxiter;
	int miniter;

	// Tabu search parameters
	int tenure;				// number of iterations a position is tabu
	int neighbourhoodsize;	// number of changes to make per iteration
	int tabusteps;			// total number of iterations to try
	double stepsize;		// step size to make (+/- is randomized)
	int max_sb;				// maximum sideband index value (+ or -) to change
							// i.e. concentrate on the center ones
	int max_tabu_restarts; 	// max restarts of tabu if search set off in the wrong direction
	int tabu_restarts;		// actual number of restarts
						
};

	
int _gv_stdev( gsl_vector *, const gsl_vector *, const gsl_vector *,
		int, gsl_vector *);
double _grandom_sigma ( const gsl_rng *, double);
int _gv_pow (gsl_vector *, const double);

int _nonlinfit_f (const gsl_vector *, void *, gsl_vector *);
int _nonlinfit_df (const gsl_vector * x, void *p, gsl_matrix * J);
int _nonlinfit_fdf (const gsl_vector *, void *, gsl_vector *, gsl_matrix *);

#endif 
