#include "cputhreadtime.h"
#include "utils.h"
#include "gp_params.h"
#include "massb.h"

int main ( int argc, char *argv[]  ) {

	MASSB P;
	P.append_logfile(); 
	GPPARAMS G;
	G.lfile = &P.lfile;

	int check = 15;
	for (int i=1; i<argc; i+=2) {
		if ( !strcmp(argv[i],"DISTFILEROOT" ) ) check -= 8;
		else if ( !strcmp(argv[i],"THREADS" ) ) check -= 4;
		else if ( !strcmp( argv[i],"RFILE") ) check -= 2;
		else if ( !strcmp( argv[i],"STATE") ) check -= 1;
	}
	if (check != 0) {
		cerr << "[R_eps_2D_dist] Need to specify: " << endl;
		cerr << "DISTFILEROOT basename of distance files" << endl;
		cerr << "THREADS string of integers for threads to be plotted" << endl;
		cerr << "RFILE basename of output" << endl;
		cerr << "STATE: state integer" << endl;
		exit(0);
	}
	G.parseinputs( argc, argv );

	char Rname[256];
	char epsname[256];
	sprintf(Rname,"%s.R",G.Rfilename);
	sprintf(epsname,"%s.eps",G.Rfilename);

	ofstream Rfile;;
	Rfile.open(Rname, ios::out);

	if ( !Rfile.is_open() ) {
		cerr << "Could not open R script file" << Rname << endl;
		exit(0);
	}

//	char str[312];
	char dfile[256];
	char line[256];
//	char tmp[10];
//	int distnum;

//	ifstream Dfile;
//	string buf;
//	sprintf(dfile,"%s.%02d.ndist.%d",G.distfileroot,G.threadstoplot[0],G.state);
//	Dfile.open(dfile, ios::in);
//	getline(data,buf);
//	sscanf(buf.c_str()+ptr, "DISTNUM %d", &distnum);
//	Dfile.close();

	Rfile << "#!/usr/bin/Rscript" << endl;
	Rfile << "require(grDevices)" <<endl;
	Rfile << "require(graphics)" <<endl << endl;

	// assume x,y coordinates same in all files
//	sprintf(line,"sp <- array(0, %d);",G.numthreadstoplot);
//	Rfile << line << endl;
//	sprintf(line,"z <- array(0, %d);",G.numthreadstoplot);
//	Rfile << line << endl;
//	sprintf(line,"z2 <- array(0, %d);",G.numthreadstoplot);
//	Rfile << line << endl << endl;

	sprintf(dfile,"%s.t%02d.ndist.%d",
		G.distfileroot,G.threadstoplot[0],G.state);
	sprintf(line,"tmp <- scan(\"%s\",list(0,0,0,0,0,0),skip=2);",dfile);
	Rfile << line << endl;
	Rfile << " delta <- tmp[[1]];" << endl;
	Rfile << "   eta <- tmp[[2]];" << endl;
	Rfile << "z0 <- tmp[[3]] + tmp[[4]]" << endl;
	Rfile << "udelta <- unique(delta)" << endl;
	Rfile << "  ueta <- unique(eta)" << endl;
	Rfile << "ldelta <- length(udelta)" << endl;
	Rfile << "  leta <- length(ueta)" << endl << endl;

	for (int s=1; s<G.numthreadstoplot; s++) {
		sprintf(dfile,"%s.t%02d.ndist.%d",
			G.distfileroot,G.threadstoplot[s],G.state);
		sprintf(line,"tmp <- scan(\"%s\",list(0,0,0,0,0,0),skip=2);",dfile);
		Rfile << line << endl;
		sprintf(line,"z%d <- tmp[[3]] + tmp[[4]];",s);
		Rfile << line << endl;
	}
	Rfile << endl;
	for (int s=0; s<G.numthreadstoplot; s++) {
		sprintf(line,"z2_%d <- matrix(z%d,nrow=ldelta,ncol=leta,byrow=TRUE);",s,s);
      Rfile << line << endl;
   }
	Rfile << endl;
	for (int s=0; s<G.numthreadstoplot; s++) {
		sprintf(line,"scale(z2_%d, center=FALSE);",s);
		Rfile << line << endl;
	}
	Rfile << endl;
	Rfile << "heat <- colorRampPalette(c(\"black\",\"purple\",\"blue\",\"red\","
		<< "\"orange\",\"yellow\"),space=\"rgb\");" << endl;
	Rfile << "setEPS();" << endl;
	Rfile << "postscript(\"" << G.Rfilename 
		<< ".R.eps\", width=7.5, height=9.5, onefile=TRUE)" << endl;
	Rfile << "par(omi=c(0.1,0.1,0.1,0.1));" << endl;
	Rfile << "par(mai=c(0.5,0.5,0.1,0.1));" << endl;
	Rfile << "layout( matrix(c(";
	for (int s=0; s<G.numthreadstoplot; s++) {
		Rfile << s+1;
		if (s<G.numthreadstoplot-1) {
			Rfile << ",";
		}
	}
	sprintf(line,"),%d,2,byrow=TRUE),",int(G.numthreadstoplot/2.0+0.5001));
	Rfile << line << endl;
	Rfile << "      widths=c(3.75,3.75)," << endl;
	Rfile << "      heights=c(";
	for (int s=0; s<int(G.numthreadstoplot/2.0+0.5001); s++) {
		Rfile << "2.375";
		if (s<int(G.numthreadstoplot/2.0+0.5001)-1) {
			Rfile << ",";
		}
	}
	Rfile << ")," << endl;
	Rfile << "      respect=TRUE" << endl;
	Rfile << ");" << endl << endl;

	for (int s=0; s<G.numthreadstoplot; s++) {
		sprintf(line,"contour(x=udelta,y=ueta,z=z2_%d,",s);
      Rfile << line << endl;
		sprintf(line,"   levels=pretty(c(range(z2_%d)),10),",s);
		Rfile << line << endl;
		Rfile << "   xlim=c(rev(range(udelta)))," << endl;
		Rfile << "   ylim=c(range(ueta))," << endl;
		Rfile << "   xlab=\"delta\"," << endl;
		Rfile << "   ylab=\"eta\"," << endl;
		Rfile << "   axes=TRUE," << endl;
		Rfile << "   lwd=1," << endl;
		Rfile << "   labels=\"\"," << endl;
		Rfile << "   col=heat(10) );" << endl << endl;
   }
	Rfile << "dev.off()" << endl;

	Rfile.close();
	
	char com[80];
	sprintf(com,"chmod u+x ./%s.R &>R.log",G.Rfilename);
	system(com);
	sprintf(com,"./%s.R &>R.log",G.Rfilename);
	system(com);

}
