#include "massb.h"
#include "cputhreadtime.h"
#include "utils.h"

MASSB::MASSB( ) {
//	T.init(20);
    // open log file
//    char logfile[256];
	outpath = getenv("OUTPATH");
	outbase = getenv("OUTBASE");
	safety = getenv("OWSAFETY");
	// default
	etat=0;

	cpuid = 0;

	char dir[149];
	strcpy(dir,outpath);
    strcat(dir,"/" );
	strcat(dir,outbase);
	struct stat st;
//	cerr << " stat(dir,&st) " <<  stat(dir,&st) << endl;

	// if nosafe active, don't check
	if ( strcmp(safety,"nosafe") ) {
		if ( stat(dir,&st) != 0 ) {	
    		char command[179];
    		strcpy(command,"mkdir ");
    		strcat(command,dir);
    		system(command);
    		
			strcpy(command,"touch ");
    		strcat(command,dir);
    		strcat(command,"/");
    		strcat(command,outbase);
    		strcat(command,".log");
    		system(command);
		} else {
			cerr << "* * * * * * * * * * *  * * * * * * *" << endl;
			cerr << "* " << outpath << " already exists." << endl;
			cerr << "* Potentially over-write files (y|N)? ";
			char yN[2];
			cin >> yN;
			if ( strcmp(yN,"y") ) exit(0);
//			exit(0);
		}
	} // even if nosafe is active, create if it doesn't exist
	else if ( stat(dir,&st) != 0 ) {
		char command[179];
		strcpy(command,"mkdir ");
		strcat(command,dir);
		system(command);
	
		sprintf(command, "touch %s/%s.t%i.log.%i",dir,outbase,cpuid,etat);
		system(command);
	}

}

int MASSB::append_logfile() {
	const char * ext = ".log";
	open_append_file(ext, &lfile, logfilename);
	return(0);
}

MASSB::MASSB( const MASSB& srcMASSB ) {
	// copy everything by hand. 

	chi2 = srcMASSB.chi2;
	avgdev = srcMASSB.avgdev;
	entropy = srcMASSB.entropy;
	datanum =srcMASSB.datanum;
	datadim = srcMASSB.datadim;
	distdim = srcMASSB.distdim;
	distnum =srcMASSB.distnum;
	num_norms = srcMASSB.num_norms;
	pp2pi = srcMASSB.pp2pi;
	Z = srcMASSB.Z;
	presetadjdensity = srcMASSB.presetadjdensity;

	//J = gsl_matrix_calloc(datanum,datanum);
	data_params = gsl_matrix_calloc( datanum, datadim);
	data_norm = gsl_matrix_calloc( datanum, 5);
	datanamp = gsl_matrix_column(data_norm,0);
	calcfix = gsl_matrix_column(data_norm,2);
	calcfit = gsl_matrix_column(data_norm,3);
	residual = gsl_matrix_column(data_norm,4);
	
	dist_params = gsl_matrix_calloc(distnum,distdim);
	distribution = gsl_matrix_calloc(distnum,4);
	population = gsl_matrix_column(distribution,1);
	fixedpopulation = gsl_matrix_column(distribution,0);
	
	lagrange = gsl_vector_alloc(datanum);	
	dist2data = gsl_matrix_alloc(datanum,distnum);		

	gsl_matrix_memcpy( dist2data, srcMASSB.dist2data );
	gsl_matrix_memcpy( dist_params, srcMASSB.dist_params );	
	gsl_matrix_memcpy( data_norm, srcMASSB.data_norm );	

	gsl_vector_memcpy( lagrange, srcMASSB.lagrange );	
	gsl_matrix_memcpy( distribution, srcMASSB.distribution );
	gsl_matrix_memcpy( dist2data, srcMASSB.dist2data );
	//gsl_matrix_memcpy( J, srcMASSB.J );
}

int MASSB::copy_min( const MASSB &srcMASSB ) {
	// Here it's important to deep-copy only critical structures.
	// everything is already allocated
	// It's OK that everything else refers to the exact same thing

//	distnum = srcMASSB.distnum;
//	datanum = srcMASSB.datanum;
	chi2 = srcMASSB.chi2;
	avgdev = srcMASSB.avgdev;
	entropy = srcMASSB.entropy;
	presetadjdensity = srcMASSB.presetadjdensity;

	gsl_vector_memcpy( lagrange, srcMASSB.lagrange );	
	gsl_matrix_memcpy( distribution, srcMASSB.distribution );	
	gsl_matrix_memcpy( data_norm, srcMASSB.data_norm );	
	//gsl_matrix_memcpy( J, srcMASSB.J );
	
	return (0);
}
	
MASSB::~MASSB () {
	if ( lfile.is_open() ) {
	//	T.reportall( &lfile );
		lfile << flush;
		lfile.close();
	}
/*
	gsl_vector_free( lagrange );
	gsl_vector_free( sideband );
	gsl_matrix_free( dist2data );
	if ( strcmp(mode,"sim") ) {	
		gsl_vector_free( mas_speeds );
		gsl_vector_free( sidebandcalc );
		gsl_vector_free( population );
		gsl_vector_free( sidebandnerr );
		gsl_vector_free( sidebandrelnerr );
		gsl_vector_free( sidebandnamp );
		gsl_vector_free( sidebandres );
		gsl_vector_free( presetpop );
		gsl_vector_free( presetnamp );
		gsl_matrix_free( J );
	}
*/

	data_norm = gsl_matrix_alloc(1,1);
	gsl_matrix_free( data_norm);
	data_notnorm = gsl_matrix_alloc(1,1);
	gsl_matrix_free( data_notnorm);
	dist_params = gsl_matrix_alloc(1,1);
	gsl_matrix_free( dist_params);
	distribution = gsl_matrix_alloc(1,1);
    gsl_matrix_free( distribution );
	
}

int MASSB::read_data() {
	char str[80];

	ifstream data;
	data.open( datafile, ios::in );
	if ( !data.is_open() )	{
		cout << "Can't open datafile " << datafile << endl;
		exit(0);
	}
	
	datanum = -1;
	datadim = -2;
	string buf;
	double d;
	int ptr = 0; int ptr_i = 0;
	do {
		getline(data,buf);
		if ( iscomment( buf ) ) continue;	
		datanum++;
		if (datanum == 0) {
			sscanf(buf.c_str()+ptr, "%le%n", &d,&ptr_i);
			ptr += ptr_i;
			do {
				ptr_i=0;
				sscanf(buf.c_str()+ptr, "%le%n", &d,&ptr_i);
				ptr += ptr_i;
				datadim++;
			} while (ptr_i != 0);
		}
	} while ( !data.eof() );

	sprintf(str,"INFERRED: object.datanum = %i\n",datanum);
	lfile << str << flush;
	sprintf(str,"INFERRED: object.datadim = %i\n",datadim);
	lfile << str << flush;

	// allocate data structures of appropriate dimensions
	// data is datadim+2 cols (characteristic params + amp + error)
	// by datanum rows
	data_params = gsl_matrix_calloc( datanum, datadim);
	data_notnorm = gsl_matrix_calloc( datanum, 2);
	
	// reset file-read pointers to the beginning of the data file
	data.clear();
	data.seekg(0);

	for (int r=0; r<datanum; r++) {
		ptr = ptr_i = 0;
		getline(data,buf);
		if ( iscomment( buf ) ) {
			r--;
			continue;
		}
		for (int p=0; p<datadim; p++) {
			ptr_i=0;
			sscanf(buf.c_str()+ptr, "%le%n", &d,&ptr_i);
			ptr += ptr_i;
			gms(data_params,r,p,d);
		}
		for (int p=0; p<2; p++) {
			ptr_i=0;
			sscanf(buf.c_str()+ptr, "%le%n", &d,&ptr_i);
			ptr += ptr_i;
			gms(data_notnorm,r,p,d);
		}
	}
	data.close();
	return(0);
}

int MASSB::norm_data () {

	data_norm = gsl_matrix_calloc( datanum, 5);
	lagrange = gsl_vector_calloc( datanum );

	num_norms = 0;	// the number of different normalizations that will be needed
	// won't need more normalizations than the number of data points!
	gsl_vector * tmp_normdim = gsl_vector_alloc(datanum);	

	for (int r=0; r<datanum; r++) {
		// keep track of how many different unique vals have been used in dim datanormdim	
		int seen = 0;
		for (int n=0; n<num_norms; n++) {
			if ( abs( gmg(data_params,r,datanormdim) 
				- gvg(tmp_normdim,n)) < 1E-6 ) seen = 1;
		}
		if ( seen != 1 || num_norms == 0) {
			gvs(tmp_normdim,num_norms,gmg(data_params,r,datanormdim));
			num_norms++;
		}
	}
	lfile << "# normalized data subsets (object.num_norms): " << num_norms << endl;

	norm_params = gsl_vector_alloc(num_norms);
	for (int m=0; m<num_norms; m++) gvs( norm_params, m, gvg(tmp_normdim,m) );

	// output data and normalized data while calculating normalizations

	// Need separate normalizations for each spinning speed
	gsl_vector * totalintensity = gsl_vector_calloc(num_norms);
	gsl_vector * totalsqerror = gsl_vector_calloc(num_norms);
	for (int r=0; r<datanum; r++) {
		double nv = gmg(data_params,r,datanormdim);
		double dv = gmg(data_notnorm,r,0);
		double ev = gmg(data_notnorm,r,1);
		for (int n=0; n<num_norms; n++) {
			if ( abs( nv - gvg(tmp_normdim,n) ) < 1E-6 ) {
				gvs(totalintensity,n, gvg(totalintensity,n)+ dv);
				gvs(totalsqerror,n, gvg(totalsqerror,n) + pow(ev,2.0));
			}
		}
	}

	
	lfile << "Data normalized ... " << endl;
	double nerr;
	double nerrsum = 0;
	int * num_at_norm = new int[num_norms];

	for (int r=0; r<datanum; r++) {
		for (int n=0; n<num_norms; n++) {
			double nv = gmg(data_params,r,datanormdim);
			double dv = gmg(data_notnorm,r,0);
			double ev = gmg(data_notnorm,r,1);
			if ( abs( nv - gvg(tmp_normdim,n) ) < 1E-6 ) {
				nerr = pow(dv,2.0) * (gvg(totalsqerror,n)-pow(ev,2.0));
				nerr += pow(ev,2.0)*pow( gvg(totalintensity,n)-dv,2.0);
				nerr = pow( nerr, 0.5);
				nerr /= pow( gvg(totalintensity,n), 2.0);

                // set normalized intensity and error
				gms( data_norm,r,0,dv/gvg(totalintensity,n) );
				gms( data_norm,r,1, nerr);
				nerrsum += nerr;
				num_at_norm[n] += 1;
			}
		}
	}

	lfile << "DATANUM " << datanum << endl;
	lfile << "DATADIM " << datadim << endl;
	lfile << "NUMNORM " << num_norms << endl;

	char num[15];
	for (int r=0; r<datanum; r++) {
		for (int n=0; n<datadim; n++) {
			sprintf(num,"%14.8f ",gmg(data_params,r,n) );
			lfile << num;
		}
		for (int n=0; n<2; n++) {
			sprintf(num,"%14.8f ",gmg(data_norm,r,n) );
			lfile << num;
		}
		for (int n=0; n<2; n++) {
			sprintf(num,"%14.8f ",gmg(data_notnorm,r,n) );
			lfile << num;
		}
		lfile << endl;
	}
	write_ndata();
	return(0);
}

int	MASSB::write_ndata_header( ofstream *outfile ) {
	*outfile << "DATANUM " << datanum << endl;
	*outfile << "DATADIM " << datadim << endl;
	lfile << "DATANUM " << datanum << endl;
	lfile << "DATADIM " << datadim << endl;
	*outfile << "NUMNORM " << num_norms << endl;
	lfile << "NUMNORM " << num_norms << endl;
	return(0);
}

int	MASSB::write_ndata_records( ofstream *outfile ) {

	char num[15];
	for (int r=0; r<datanum; r++) {
		for (int n=0; n<datadim; n++) {
			sprintf(num,"%14.8f ",gmg(data_params,r,n) );
			*outfile << num;
			lfile << num;
		}
		for (int n=0; n<5; n++) {
			sprintf(num,"%14.8f ",gmg(data_norm,r,n) );
			*outfile << num;
			lfile << num;
		}
		sprintf(num,"%14.8f ",gvg(lagrange,r) );
		*outfile << num;
		lfile << num;

		*outfile << endl;
		lfile << endl;
		
	}

	return(0);
}

int	MASSB::write_ndata() {
	char outapp[20];
	sprintf(outapp,".t%02i.ndata.%i",cpuid,etat);
	
	ofstream outfile;
	open_write_file(outapp,&outfile);

	write_ndata_header( &outfile );	
	write_ndata_records( &outfile );

	outfile.close();	
	return(0);
}

int MASSB::pre2dat() {
//	gsl_vector_memcpy(&datanamp.vector,&calcfix.vector);
//	gsl_vector_set_zero(&calcfix.vector);
	gsl_matrix_swap_columns(data_norm,0,2);
	presetadjdensity = 0;
	gsl_vector_view err = gsl_matrix_column(data_norm,1);
	gsl_vector_set_all(&err.vector,sigma);
	return(0);
}

int MASSB::read_ndata_records( ifstream *data ) {
	string buf;
	// need to make this more general is we ever find the case that
	// the ndist file needs to report more than DATANUM and DATADIM
	for (int j=0; j<3; j++) {
		getline(*data,buf);
		parseinputline(buf);
	}

	// allocate data structures of appropriate dimensions
	// data is datadim cols (characteristic params) by datanum rows
	data_params = gsl_matrix_calloc( datanum, datadim);
	data_norm = gsl_matrix_calloc( datanum, 5);
	lagrange = gsl_vector_calloc( datanum );

	datanamp = gsl_matrix_column(data_norm,0);
	calcfix = gsl_matrix_column(data_norm,2);
	calcfit = gsl_matrix_column(data_norm,3);
	residual = gsl_matrix_column(data_norm,4);

	double d;	
	for (int r=0; r<datanum; r++) {
		int ptr = 0;
		int ptr_i = 0;
		getline(*data,buf);
		if ( iscomment( buf ) ) continue;	
		for (int p=0; p<datadim; p++) {
			ptr_i=0;
			sscanf(buf.c_str()+ptr, "%le%n", &d,&ptr_i);
			ptr += ptr_i;
			gms(data_params,r,p,d);
		}
		for (int p=0; p<5; p++) {
			ptr_i=0;
			sscanf(buf.c_str()+ptr, "%le%n", &d,&ptr_i);
			ptr += ptr_i;
			gms(data_norm,r,p,d);
		}
		sscanf(buf.c_str()+ptr, "%le%n", &d,&ptr_i);
		ptr += ptr_i;
		gvs(lagrange,r,d);
	}
	return(0);
}

int MASSB::read_ndata() {
	// this file a wrapper for opening single ndata file and reading

	ifstream data;
	data.open( datafile, ios::in );
	if ( !data.is_open() )	{
		cout << "Can't open datafile " << datafile << endl;
		exit(0);
	}
	read_ndata_records( &data );
	data.close();
	return(0);
}

int	MASSB::write_full_bsetat() {
	
	char app[7];
	sprintf(app,".t%02i.bs.%i",cpuid,etat);

	ofstream outfile;
	open_write_file(app,&outfile);

	write_ndata_header( &outfile );
	write_ndata_records( &outfile );
	write_dist_header( &outfile );
	write_dist_records( &outfile );
	write_dist2data_matrix_records( &outfile );


	outfile.close();	
	return(0);
}

int	MASSB::read_full_bsetat() {
	
	ifstream bsetat;
	string buf;

	bsetat.open( bsetatfile, ios::in );
	if ( !bsetat.is_open() )	{
		cout << "Can't open bsetat file " << bsetatfile << endl;
		exit(0);
	}

	read_ndata_records( &bsetat );
	read_dist_records( &bsetat );
	getline(bsetat,buf);
	read_dist2data_matrix_records( &bsetat );

	bsetat.close();	
	return(0);
}
	
int MASSB::read_dist_records( ifstream *dist ) {
	string buf;
	// need to make this more general is we ever find the case that
	// the ndist file needs to report more than DATANUM and DATADIM
	for (int j=0; j<2; j++) {
		getline(*dist,buf);
		parseinputline(buf);
	}

	// allocate data structures of appropriate dimensions
	dist_params = gsl_matrix_calloc(distnum,distdim);
	distribution = gsl_matrix_calloc(distnum,4);
	population = gsl_matrix_column(distribution,1);
	fixedpopulation = gsl_matrix_column(distribution,0);

	int ptr, ptr_i;
	double d;
	presetadjdensity = 0;
	for (int r=0; r<distnum; r++) {
		ptr = ptr_i = 0;
		getline(*dist,buf);
		for (int p=0; p<distdim; p++) {
			ptr_i=0;
			sscanf(buf.c_str()+ptr, "%le%n", &d,&ptr_i);
			ptr += ptr_i;
			gms(dist_params,r,p,d);
		}
		for (int p=0; p<4; p++) {
			ptr_i=0;
			sscanf(buf.c_str()+ptr, "%le%n", &d,&ptr_i);
			ptr += ptr_i;
			gms(distribution,r,p,d);
		}
		presetadjdensity += gvg(&fixedpopulation.vector,r);
	}
	if (cpuid == 0) {
		lfile << "total pre-set density: " << presetadjdensity << endl;
	}
	return(0);
}

int MASSB::read_dist() {

	ifstream dist;
	dist.open( distfile, ios::in );
	if ( !dist.is_open() )	{
		cout << "Can't open datafile " << distfile << endl;
		exit(0);
	}
	read_dist_records( &dist );
	dist.close();
	return(0);
}

int MASSB::set_dist() {
	char str[128];

	// init the total number of distribution parameter permutations
	distnum = 1;

	// alloc the array to hold the number of vals for each dist parameter
	dist_dim_num = new int[distdim];

	// loop through each min max res triplet given in DISTPARAMS input		
	for (int i=0; i<distdim; i+=3 ) {	
		// 1+ (max - min)/res for each parameter triplet. If min = max,
		// avoid potential /(res=0) problem with if/else
		// if max-min is not evenly divisble by res, 
		// this should be caught, user informed, and aborted.
		if ( abs(rawdistparamlimits[i+1] -rawdistparamlimits[i]) < 1E-9 ) {
			dist_dim_num[i/3] = 1;
			sprintf(str,"distribution parameter %i: 1 value = %f\n",
				i,rawdistparamlimits[i]);
			lfile << str;
		} else { dist_dim_num[i/3] = (1+ int((rawdistparamlimits[i+1]
			-rawdistparamlimits[i]) /rawdistparamlimits[i+2] + 0.5 )); 
			sprintf(str,
			   "distribution parameter %i: %i values, min %f, max %f, res %f\n",
				i/3, dist_dim_num[i/3], rawdistparamlimits[i],
				rawdistparamlimits[i+1], rawdistparamlimits[i+2]);
			lfile << str;
		}
		// multiply above result into increasing product which represents 
		// total number of permutations of distribution parameters.
		distnum *= dist_dim_num[i/3];
	}
//	delete rawdistparamlimits;
	distdim /= 3;
	
	lfile << "Total number of distribution parameter permutations: " 
		<< distnum << endl;

	// Allocate distribution data structures.

	// dist[ribution]_param[eter]s is matrix, all permutations of all 
	// distribution parameters by row, incremental values of each parameter from
	// min to max in res steps down each column.
	dist_params = gsl_matrix_calloc(distnum,distdim);

	// distribution is columns: 
	// (1) fixed_population (set elsewhere), (2) fit_pop, (3) cumulative_pop, 
	//	(4) error_pop (e.g. by MC)
	// population is a vector_view of the second column from distribution 
	//  matrix for use in cblas linear algebra routines later.
	distribution = gsl_matrix_calloc(distnum,4);
	population = gsl_matrix_column(distribution,1);
	fixedpopulation = gsl_matrix_column(distribution,0);

	// In order to allow for any number of distribution parameters, we want
	// to be complete general and therefore cannot explicitly nest loops to
	// fill the dist_params matrix. Therefore use a recursive loop trick.
	double * mem; mem = new double [distdim];
	int r=0;
	setdistparamrecurloop(0,&r,mem);
	delete mem;
	delete dist_dim_num;

	// write output ndist file with 0s mostly.
	// output normalized data to file
	write_dist();
	return(0);
}

int MASSB::set_data() {
	char str[128];

	// init the total number of distribution parameter permutations
	datanum = 1;

	// alloc the array to hold the number of vals for each dist parameter
	data_dim_num = new int[datadim];

	// loop through each min max res triplet given in DISTPARAMS input		
	for (int i=0; i<datadim; i+=3 ) {	
		// 1+ (max - min)/res for each parameter triplet. If min = max,
		// avoid potential /(res=0) problem with if/else
		// if max-min is not evenly divisble by res, 
		// this should be caught, user informed, and aborted.
		if ( abs(rawdataparamlimits[i+1] -rawdataparamlimits[i]) < 1E-9 ) {
			data_dim_num[i/3] = 1;
			sprintf(str,"data parameter %i: 1 value = %f\n",
				i,rawdataparamlimits[i]);
			lfile << str;
		} else { data_dim_num[i/3] = (1+ int((rawdataparamlimits[i+1]
			-rawdataparamlimits[i]) /rawdataparamlimits[i+2] + 0.5 )); 
			sprintf(str,
			   "data parameter %i: %i values, min %f, max %f, res %f\n",
				i/3, data_dim_num[i/3], rawdataparamlimits[i],
				rawdataparamlimits[i+1], rawdataparamlimits[i+2]);
			lfile << str;
		}
		// multiply above result into increasing product which represents 
		// total number of permutations of data parameters.
		datanum *= data_dim_num[i/3];
	}
//	delete rawdataparamlimits;
	datadim /= 3;
	
	lfile << "Total number of data parameter permutations: " 
		<< datanum << endl;

	// Allocate data data structures.

	// data_param[eter]s is matrix, all permutations of all 
	// data parameters by row, incremental values of each parameter from
	// min to max in res steps down each column.
	data_params = gsl_matrix_calloc(datanum,datadim);
	data_norm = gsl_matrix_calloc( datanum, 5);
	lagrange = gsl_vector_calloc( datanum );

	// In order to allow for any number of distribution parameters, we want
	// to be complete general and therefore cannot explicitly nest loops to
	// fill the dist_params matrix. Therefore use a recursive loop trick.
	double * mem; mem = new double [datadim];
	int r=0;
	setdataparamrecurloop(0,&r,mem);
	delete mem;
	delete data_dim_num;

	// write output ndata file with 0s mostly.
	// output normalized data to file
	write_ndata();
	return(0);
}

int MASSB::set_data_explicit() {

	// init the total number of distribution parameter permutations
	datadim = 2;
	datanum = datadim1*datadim2;

	lfile << "Total number of data parameter permutations: " 
		<< datanum << endl;

	// Allocate data data structures.

	// data_param[eter]s is matrix, all permutations of all 
	// data parameters by row, incremental values of each parameter from
	// min to max in res steps down each column.
	data_params = gsl_matrix_calloc(datanum,datadim);
	data_norm = gsl_matrix_calloc( datanum, 5);
	lagrange = gsl_vector_calloc( datanum );

	// fill the parameters into data_params explicitly. 
	// future applications that want to use this with more parameters
	// will just have to add more nested loops
	for (int d1=0; d1<datadim1; d1++) {
		for (int d2=0; d2<datadim2; d2++ ) {
			int parindex = d1*datadim2+d2;
			gms(data_params,parindex,0,params1[d1] );
			gms(data_params,parindex,1,params2[d2] );
		}
	}
			
	// write output ndata file with 0s mostly.
	// output normalized data to file
	write_ndata();
	return(0);
}

int MASSB::write_dist_header( ofstream *outfile ) {

	*outfile << "DISTNUM " << distnum << endl;
	lfile << "DISTNUM " << distnum << endl;
	*outfile << "DISTDIM " << distdim << endl;
	lfile << "DISTDIM " << distdim << endl;
	return(0);
}

int MASSB::write_dist_records( ofstream *outfile ) {

	char num[15];
	for (int r=0; r<distnum; r++) {
		for (int n=0; n<distdim; n++) {
			sprintf(num,"%14.7f ",gmg(dist_params,r,n) );
			*outfile << num;
		}
		for (int n=0; n<4; n++) {
			sprintf(num,"%14.8f ",gmg(distribution,r,n) );
			*outfile << num;
		}
		*outfile << endl;
	}
	return(0);

}

int MASSB::write_dist() {
	char outapp[20];
	sprintf(outapp,".t%02i.ndist.%i",cpuid,etat);
	
	ofstream outfile;
	open_write_file(outapp,&outfile);

	write_dist_header( &outfile );
	write_dist_records( &outfile );

	outfile.close();	
	return(0);

}

void MASSB::setdataparamrecurloop 
						( int level, int *r, double * parampermutation ) {
	if (level == datadim) {
		for (int d=0; d<datadim; d++) {
			gms(data_params,*r,d,parampermutation[d]);
		}
		(*r)++;
	} else {
		for (int i=0; i<data_dim_num[level]; i++) {
			parampermutation[level] = double(i)*rawdataparamlimits[level*3+2] 
				+ rawdataparamlimits[level*3];
			setdataparamrecurloop(level+1,r,parampermutation);
		}
	}
}

void MASSB::setdistparamrecurloop 
						( int level, int *r, double * parampermutation ) {
	if (level == distdim) {
		for (int d=0; d<distdim; d++) {
			gms(dist_params,*r,d,parampermutation[d]);
		}
		(*r)++;
	} else {
		for (int i=0; i<dist_dim_num[level]; i++) {
			parampermutation[level] = double(i)*rawdistparamlimits[level*3+2] 
				+ rawdistparamlimits[level*3];
			setdistparamrecurloop(level+1,r,parampermutation);
		}
	}
}

int MASSB::calc_dist2data_crystfile_abg() {
// If all three angles and weights are supplied by file
	string buf;
	char str[80];

	double alpha,beta,gamma,weight;
	double sinalpha,sinbeta,singamma,cosalpha,cosbeta,cosgamma;
	double sin2alpha,sin2beta,sin2gamma,cos2alpha,cos2beta,cos2gamma;
	double PhiA,PhiB,Phitmp,Phitmp2,Phi_Nprime,Iab,N,csum,ssum,wsum;
	double delta, eta, mas;

	gsl_matrix * Csum = gsl_matrix_calloc(datanum, distnum); 
	gsl_matrix * Ssum = gsl_matrix_calloc(datanum, distnum); 
	dist2data = gsl_matrix_calloc( datanum, distnum ); 
		
	// Read alpha, beta, gamma and weights from crystal file
	ifstream crystfile;
	crystfile.open( crystname, ios::in );
	if ( !crystfile.is_open() )	{
		cout << "Can't open crystal file " << crystname << endl;
		exit(1);
	}
	int nab;
   	getline(crystfile,buf); sscanf(buf.c_str(), "%d", &nab);
	gsl_vector * crystalpha = gsl_vector_alloc(nab);
	gsl_vector * crystbeta = gsl_vector_alloc(nab);
	gsl_vector * crystgamma = gsl_vector_alloc(nab);
	gsl_vector * crystweight = gsl_vector_alloc(nab);
	int i = 0;
	wsum = 0;

	double ca, cb, cg, cw;
	getline(crystfile,buf); 
	do {
		sscanf(buf.c_str(), "%le %le %le %le", &ca, &cb, &cg, &cw);
		gvs(crystalpha,i,ca);
		gvs(crystbeta,i,cb);
		gvs(crystgamma,i,cg);
		gvs(crystweight,i,cw);
		wsum += cw;
		i++;
		getline(crystfile,buf); 
	} while ( !crystfile.eof() ) ;
	crystfile.close();
	
	sprintf(str,"%i CRYSTALLITES(a,b,g) AND WEIGHTS READ FROM %s\n",
		nab,crystname);
	lfile << str;
	sprintf(str,"TOTAL INTEGRAL OF CRYSTALLITE WEIGHTS: %15.9g\n",wsum);
	lfile << str;

	// replace gamma, beta, alpha loops with crystal file angles and weights
	for ( int ab=0; ab<nab; ab++ ) {
		alpha = gvg(crystalpha,ab);
		beta = gvg(crystbeta,ab);
		gamma = gvg(crystgamma,ab);
		weight = gvg(crystweight,ab);

		sinbeta = sin(beta);
		sin2beta = sin(2.0*beta);
		cosbeta = cos(beta);
		cos2beta = cos(2.0*beta);
			
		sinalpha = sin(alpha);
		sin2alpha = sin(2.0*alpha);
		cosalpha = cos(alpha);
		cos2alpha = cos(2.0*alpha);

		singamma = sin(gamma);
		sin2gamma = sin(2.0*gamma);
		cosgamma = cos(gamma);
		cos2gamma = cos(2.0*gamma);

		PhiA = cos2alpha*(3.0+cos2beta)*sin2gamma/(-24.0);
		PhiA -= 0.235702260395516 * cos2alpha*sin2beta*singamma;
		PhiA -= 0.471404520791032 * sin2alpha*sinbeta*cosgamma;
		PhiA -= 0.166666666666667 * sin2alpha*cosbeta*cos2gamma;

		PhiB = (1.0 - cos2beta)*sin2gamma/8.0;
		PhiB -= 0.707106781186547*sin2beta*singamma;

		for (int cm=0; cm<distnum; cm++) {
			delta = gmg(dist_params,cm,0);
			eta   = gmg(dist_params,cm,1);

			Phitmp = delta*eta*PhiA + delta*PhiB;
			for (int n=0; n<datanum; n++) {
				mas = gmg(data_params,n,0);
				N   = gmg(dist_params,n,1);

				Phitmp2 = Phitmp/mas;
				Phi_Nprime = N*gamma - Phitmp2;
				csum = gmg(Csum,n,cm);
				ssum = gmg(Ssum,n,cm);
				gms( Csum, n, cm, csum + cos( Phi_Nprime)*weight );
				gms( Ssum, n, cm, ssum + sin( Phi_Nprime)*weight );
			}
		} // end configurational matrix (delta,eta) block
			
	} // End angles block
	double cumsum = 0;
	for (int cm=0; cm<distnum; cm++) {
		for (int n=0; n<datanum; n++) {
			Iab = (pow( gmg( Csum, n, cm), 2.0) 
				+ pow( gmg( Ssum, n, cm), 2.0));
			gsl_matrix_set(dist2data,n,cm, Iab );
			cumsum += Iab;

		} // end sideband block
	} // end config matrix (delta,eta) block
	lfile << "cumulative sum: " << cumsum << endl;

	gsl_vector_free(crystalpha);
	gsl_vector_free(crystbeta);
	gsl_vector_free(crystgamma);
	gsl_vector_free(crystweight);
	gsl_matrix_free(Csum);
	gsl_matrix_free(Ssum);
		
	write_dist2data_matrix();
	return(0);
}

int MASSB::calc_dist2data_crystfile_ab() {
	double pi2 = 6.28318530718;
	double incpi = pi2/double(pp2pi);
	string buf;
	char str[80];

	double alpha,beta,gamma,weight;
	double sinalpha,sinbeta,singamma,cosalpha,cosbeta,cosgamma;
	double sin2alpha,sin2beta,sin2gamma,cos2alpha,cos2beta,cos2gamma;
	double PhiA,PhiB,Phitmp,Phitmp2,Phi_Nprime,Iab,N,csum,ssum,sum, wsum;
	double delta, eta, mas;

	gsl_matrix * Csum = gsl_matrix_calloc(datanum,distnum); 
	gsl_matrix * Ssum = gsl_matrix_calloc(datanum,distnum); 
	dist2data = gsl_matrix_calloc( datanum, distnum ); 

	// Read alpha, beta, and weights from crystal file
	ifstream crystfile;
	crystfile.open( crystname, ios::in );
	if ( !crystfile.is_open() )	{
		cout << "Can't open crystal file " << crystname << endl;
		exit(1);
	}
	int nab;
	double ca, cb, cw;
   	getline(crystfile,buf); sscanf(buf.c_str(), "%d", &nab);
	cout << "number crystallites: " << nab << " X gamma angles: " 
		<< pp2pi << " = " << nab*pp2pi << endl;
	gsl_vector * crystalpha = gsl_vector_alloc(nab);
	gsl_vector * crystbeta = gsl_vector_alloc(nab);
	gsl_vector * crystweight = gsl_vector_alloc(nab);
	int i = 0;
	wsum = 0;

	getline(crystfile,buf); 
	do {
		sscanf(buf.c_str(), "%le %le %le", &ca, &cb, &cw);
		gvs(crystalpha,i,ca*pi2/360.0);
		gvs(crystbeta,i,cb*pi2/360.0);
		gvs(crystweight,i,cw);
		wsum += cw;
		i++;
		getline(crystfile,buf); 
	} while ( !crystfile.eof() ) ;
	crystfile.close();

	sprintf(str,"%i CRYSTALLITES(a,b) AND WEIGHTS READ FROM %s\n",
		nab,crystname);
	lfile << str;
	sprintf(str,"TOTAL INTEGRAL OF CRYSTALLITE WEIGHTS: %15.9g\n",wsum);
	lfile << str;

	for ( int ab=0; ab<nab; ab++ ) {
//		printf("%6d/%6d",ab,nab);
//		cout << flush;

		alpha = gvg(crystalpha,ab);
		beta = gvg(crystbeta,ab);
		weight = gvg(crystweight,ab);

		sinbeta = sin(beta);
		sin2beta = sin(2.0*beta);
		cosbeta = cos(beta);
		cos2beta = cos(2.0*beta);
		
		sinalpha = sin(alpha);
		sin2alpha = sin(2.0*alpha);
		cosalpha = cos(alpha);
		cos2alpha = cos(2.0*alpha);

		gsl_matrix_set_all(Csum,0.0);	
		gsl_matrix_set_all(Ssum,0.0);	
		for (int g=0; g<pp2pi; g++) {
			gamma = double(g)*incpi;
			singamma = sin(gamma);
			sin2gamma = sin(2.0*gamma);
			cosgamma = cos(gamma);
			cos2gamma = cos(2.0*gamma);

			PhiA = cos2alpha*(3.0+cos2beta)*sin2gamma/(-24.0);
			PhiA -= 0.235702260395516 * cos2alpha*sin2beta*singamma;
			PhiA -= 0.471404520791032 * sin2alpha*sinbeta*cosgamma;
			PhiA -= 0.166666666666667 * sin2alpha*cosbeta*cos2gamma;

			PhiB = (1.0 - cos2beta)*sin2gamma/8.0;
			PhiB -= 0.707106781186547*sin2beta*singamma;

			for (int cm=0; cm<distnum; cm++) {
				delta = gmg(dist_params,cm,0);
				eta   = gmg(dist_params,cm,1);
				Phitmp = delta*eta*PhiA + delta*PhiB;

				for (int n=0; n<datanum; n++) {
					mas = gmg(data_params,n,0);
					N   = gmg(data_params,n,1);
					Phitmp2 = Phitmp/mas;
					Phi_Nprime = N*gamma - Phitmp2;

					csum = gmg(Csum,n,cm);
					ssum = gmg(Ssum,n,cm);
					gsl_matrix_set( Csum, n, cm, 
						csum + cos( Phi_Nprime) );
					gsl_matrix_set( Ssum, n, cm, 
						ssum + sin( Phi_Nprime) );
				}
			} // end configurational matrix (delta,eta) block
		} // End gamma block
		for (int cm=0; cm<distnum; cm++) {
			for (int n=0; n<datanum; n++) {
				Iab = (pow( gmg( Csum, n, cm), 2.0) 
					+ pow( gmg( Ssum, n, cm), 2.0));
				sum = gmg(dist2data,n,cm);
				gsl_matrix_set(dist2data,n,cm, sum + Iab*weight );
			} // end sideband block
		} // end config matrix (delta,eta) block
	}
//	double ssn = double(num_norms)/double(datanum*distnum);	
	double ssn = 1.0;
	gsl_matrix_scale(dist2data,ssn/(pow(double(pp2pi),2)));

	double cumsum = 0;	
	for (int cm=0; cm<distnum; cm++) 
		for (int n=0; n<datanum; n++) 
			cumsum += gmg(dist2data,n,cm);
	lfile << "cumulative sum: " << cumsum << endl;
	gsl_vector_free(crystalpha);
	gsl_vector_free(crystbeta);
	gsl_vector_free(crystweight);

	gsl_matrix_free(Csum);
	gsl_matrix_free(Ssum);
		
	get_datadist_md5();
	write_dist2data_matrix();
	return(0);
} 

int MASSB::calc_dist2data_openmp() {
	double incpi = 6.28318530718/double(pp2pi);
	string buf;

	time_t seconds = time(NULL);
	lfile << "OPENMP: begin beta loop "  << ctime(&seconds) << endl;
	cout << "nested loop combinations: " << pow(pp2pi,3)/2 << endl;	
	
	// timestamp
//	T.mark(&lfile,"calc_dist2data_openmp(): calculating trig vectors");

	// pre-compute trig functions, as they're used over and over
	gsl_vector * sinangle = gsl_vector_alloc(pp2pi);	
	gsl_vector * cosangle = gsl_vector_alloc(pp2pi);	
	gsl_vector * sindangle = gsl_vector_alloc(pp2pi);	
	gsl_vector * cosdangle = gsl_vector_alloc(pp2pi);	

	for (int j=0; j<pp2pi; j++) {
		double angle = double(j)*incpi;
		gvs(sinangle,j,sin(angle) );
		gvs(cosangle,j,cos(angle) );
		gvs(sindangle,j,sin(2.0*angle) );
		gvs(cosdangle,j,cos(2.0*angle) );
	}

	// timestamp
//	T.mark(&lfile,"calc_dist2data_openmp(): begin beta loop");

	// start parallel thread monitoring
//	T.zero_coreduty();
	dist2data = gsl_matrix_calloc( datanum, distnum ); 
	//gsl_matrix *Csum, *Ssum, *ldist2data;
	gsl_matrix * ldist2data;// = gsl_matrix_calloc( datanum, distnum ); 
	gsl_matrix * Csum;// = gsl_matrix_calloc(datanum,distnum); 
	gsl_matrix * Ssum;// = gsl_matrix_calloc(datanum,distnum); 

	#pragma omp parallel for num_threads(4) firstprivate(Csum,Ssum,ldist2data) schedule (dynamic,2) 

	// integrate over beta
	for (int b=0; b<pp2pi/2; b++) {
///		T.simpletimestamp(&lfile);
//		T.incr_coreduty();

		double gamma;
		double sinalpha,sinbeta,singamma,cosalpha,cosbeta,cosgamma;
		double sin2alpha,sin2beta,sin2gamma,cos2alpha,cos2beta,cos2gamma;
		double PhiA,PhiB,Phitmp,Phitmp2,Phi_Nprime,Iab,N,csum,ssum,sum;
		int a;
		double delta, eta, mas;
	
		ldist2data = gsl_matrix_calloc( datanum, distnum ); 
		gsl_matrix_set_all(ldist2data,0.0);// = gsl_matrix_calloc( datanum, distnum ); 
//		gsl_matrix_set_all(Csum,0.0);// = gsl_matrix_calloc(datanum,distnum); 
//		gsl_matrix_set_all(Ssum,0.0);// = gsl_matrix_calloc(datanum,distnum); 

		sinbeta = gvg(sinangle,b);
		sin2beta = gvg(sindangle,b);
		cosbeta = gvg(cosangle,b);
		cos2beta = gvg(cosdangle,b);

		// integrate over alpha
		for (a=0; a<pp2pi; a++) {

			sinalpha = gvg(sinangle,a);
			sin2alpha = gvg(sindangle,a);
			cosalpha = gvg(cosangle,a);
			cos2alpha = gvg(cosdangle,a);
			
			//gsl_matrix_set_all(Csum,0.0);	
			//gsl_matrix_set_all(Ssum,0.0);	
			Csum = gsl_matrix_calloc(datanum,distnum); 
			Ssum = gsl_matrix_calloc(datanum,distnum); 

			for (int g=0; g<pp2pi; g++) {
				gamma = g*incpi;

				singamma = gvg(sinangle,g);
				sin2gamma = gvg(sindangle,g);
				cosgamma = gvg(cosangle,g);
				cos2gamma = gvg(cosdangle,g);

				PhiA = cos2alpha*(3.0+cos2beta)*sin2gamma/(-24.0);
				PhiA -= 0.235702260395516 * cos2alpha*sin2beta*singamma;
				PhiA -= 0.471404520791032 * sin2alpha*sinbeta*cosgamma;
				PhiA -= 0.166666666666667 * sin2alpha*cosbeta*cos2gamma;

				PhiB = (1.0 - cos2beta)*sin2gamma/8.0;
				PhiB -= 0.707106781186547*sin2beta*singamma;

				for (int cm=0; cm<distnum; cm++) {
					delta = gmg(dist_params,cm,0);
					eta   = gmg(dist_params,cm,1);

					Phitmp = delta*eta*PhiA + delta*PhiB;
					for (int n=0; n<datanum; n++) {
						mas = gmg(data_params,n,0);
						N   = gmg(data_params,n,1);

						Phitmp2 = Phitmp/mas;
						Phi_Nprime = N*gamma - Phitmp2;
						csum = gmg(Csum,n,cm);
						ssum = gmg(Ssum,n,cm);
						gsl_matrix_set( Csum, n, cm, 
							csum + cos( Phi_Nprime) );
						gsl_matrix_set( Ssum, n, cm, 
							ssum + sin( Phi_Nprime) );
					}
				} // end configurational matrix (delta,eta) block
			} // End gamma block
			for (int cm=0; cm<distnum; cm++) {
				for (int n=0; n<datanum; n++) {
					Iab = (pow( gmg( Csum, n, cm), 2.0) 
						+ pow( gmg( Ssum, n, cm), 2.0));
					sum = gmg(ldist2data,n,cm);
					gsl_matrix_set(ldist2data,n,cm, sum + Iab*sinbeta );
				} // end sideband block
			} // end config matrix (delta,eta) block
			gsl_matrix_free(Csum);
			gsl_matrix_free(Ssum);
		} // End alpha block

		#pragma omp critical
		gsl_matrix_add(dist2data,ldist2data);
		gsl_matrix_free(ldist2data);
	} // End beta block
//	T.report_coreduty( &lfile, pp2pi/2 );
	
	// timestamp
//	T.mark(&lfile,"powderopt = 1 finished beta looping");

//		cout << endl;
	
	seconds = time(NULL);
	lfile << "OPENMP: ended beta loop "  << ctime(&seconds) << endl << flush;
	gsl_matrix_scale(dist2data,3.14159265358979/(pow(double(pp2pi),4.0)));
	
	double cumsum = 0;	
	for (int cm=0; cm<distnum; cm++) 
		for (int n=0; n<datanum; n++) 
			cumsum += gmg(dist2data,n,cm);
	lfile << "cumulative sum: " << cumsum << endl;
	cout << "cumulative sum = " << cumsum << endl;

	gsl_vector_free(sinangle);
	gsl_vector_free(cosangle);
	gsl_vector_free(sindangle);
	gsl_vector_free(cosdangle);
	
	get_datadist_md5();
	write_dist2data_matrix();

	return(0);
}

int MASSB::calc_dist2data() {
	double pi2 = 6.28318530718;
	double incpi = pi2/double(pp2pi);
	string buf;

	cout << "nested loop combinations: " << pow(pp2pi,3)/2 << endl;	
	
	// timestamp
//	T.mark(&lfile,"powderopt 4 (OpenMP): calculating trig vectors");

	// pre-compute trig functions, as they're used over and over
	gsl_vector * sinangle = gsl_vector_alloc(pp2pi);	
	gsl_vector * cosangle = gsl_vector_alloc(pp2pi);	
	gsl_vector * sindangle = gsl_vector_alloc(pp2pi);	
	gsl_vector * cosdangle = gsl_vector_alloc(pp2pi);	

	for (int j=0; j<pp2pi; j++) {
		double angle = double(j)*incpi;
		gvs(sinangle,j,sin(angle) );
		gvs(cosangle,j,cos(angle) );
		gvs(sindangle,j,sin(2.0*angle) );
		gvs(cosdangle,j,cos(2.0*angle) );
	}

	// timestamp
//	T.mark(&lfile,"powderopt 1: begin beta looping of powderopt");

	// start parallel thread monitoring
//	T.zero_coreduty();
	dist2data = gsl_matrix_calloc( datanum, distnum ); 


	// integrate over beta
	for (int b=0; b<pp2pi/2; b++) {
///		T.simpletimestamp(&lfile);
//		T.incr_coreduty();
	
		double gamma;
		double sinalpha,sinbeta,singamma,cosalpha,cosbeta,cosgamma;
		double sin2alpha,sin2beta,sin2gamma,cos2alpha,cos2beta,cos2gamma;
		double PhiA,PhiB,Phitmp,Phitmp2,Phi_Nprime,Iab,N,csum,ssum,sum;
		int a;
		double delta, eta, mas;
		
		gsl_matrix * Csum = gsl_matrix_calloc(datanum,distnum); 
		gsl_matrix * Ssum = gsl_matrix_calloc(datanum,distnum); 

		sinbeta = gvg(sinangle,b);
		sin2beta = gvg(sindangle,b);
		cosbeta = gvg(cosangle,b);
		cos2beta = gvg(cosdangle,b);

		// integrate over alpha
		for (a=0; a<pp2pi; a++) {

			sinalpha = gvg(sinangle,a);
			sin2alpha = gvg(sindangle,a);
			cosalpha = gvg(cosangle,a);
			cos2alpha = gvg(cosdangle,a);
			
			gsl_matrix_set_all(Csum,0.0);	
			gsl_matrix_set_all(Ssum,0.0);	

			for (int g=0; g<pp2pi; g++) {
				gamma = g*incpi;

				singamma = gvg(sinangle,g);
				sin2gamma = gvg(sindangle,g);
				cosgamma = gvg(cosangle,g);
				cos2gamma = gvg(cosdangle,g);

				PhiA = cos2alpha*(3.0+cos2beta)*sin2gamma/(-24.0);
				PhiA -= 0.235702260395516 * cos2alpha*sin2beta*singamma;
				PhiA -= 0.471404520791032 * sin2alpha*sinbeta*cosgamma;
				PhiA -= 0.166666666666667 * sin2alpha*cosbeta*cos2gamma;

				PhiB = (1.0 - cos2beta)*sin2gamma/8.0;
				PhiB -= 0.707106781186547*sin2beta*singamma;

				for (int cm=0; cm<distnum; cm++) {
					delta = gmg(dist_params,cm,0);
					eta   = gmg(dist_params,cm,1);

					Phitmp = delta*eta*PhiA + delta*PhiB;
					for (int n=0; n<datanum; n++) {
						mas = gmg(data_params,n,0);
						N   = gmg(dist_params,n,1);

						Phitmp2 = Phitmp/mas;
						Phi_Nprime = N*gamma - Phitmp2;
						csum = gmg(Csum,n,cm);
						ssum = gmg(Ssum,n,cm);
						gsl_matrix_set( Csum, n, cm, 
							csum + cos( Phi_Nprime) );
						gsl_matrix_set( Ssum, n, cm, 
							ssum + sin( Phi_Nprime) );
					}
				} // end configurational matrix (delta,eta) block
			} // End gamma block
			for (int cm=0; cm<distnum; cm++) {
				for (int n=0; n<datanum; n++) {
					Iab = (pow( gmg( Csum, n, cm), 2.0) 
						+ pow( gmg( Ssum, n, cm), 2.0));
					sum = gmg(dist2data,n,cm);
					gsl_matrix_set(dist2data,n,cm, sum + Iab*sinbeta );
				} // end sideband block
			} // end config matrix (delta,eta) block
		} // End alpha block
//		gsl_matrix_free(Csum);
//		gsl_matrix_free(Ssum);

	} // End beta block
//	T.report_coreduty( &lfile, pp2pi/2 );
	
	// timestamp
//	T.mark(&lfile,"powderopt = 1 finished beta looping");

//		cout << endl;
	
	gsl_matrix_scale(dist2data,3.14159265358979/(pow(double(pp2pi),4)));
	
	double cumsum = 0;
	for (int n=0; n<datanum; n++) cumsum += gmg(dist2data,n,0);
	cout << "sidebandsum = " << cumsum << endl;

	gsl_vector_free(sinangle);
	gsl_vector_free(cosangle);
	gsl_vector_free(sindangle);
	gsl_vector_free(cosdangle);

	write_dist2data_matrix();

	return(0);
}

int MASSB::get_datadist_md5() {
	char out[120];

	ofstream datadist;
	datadist.open("md5.tmp", ios::out);
	datadist << "DATANUM " << datanum << endl;
	datadist << "DATADIM " << datadim << endl;
	for (int r=0; r<datanum; r++) {
		for (int c=0; c<datadim; c++) {
			sprintf(out,"%20g ",gmg(data_params,r,c) );
			datadist << out;
		}
		datadist << endl;
	}
	datadist << "DISTNUM " << distnum << endl;
	datadist << "DISTDIM " << distdim << endl;
	for (int r=0; r<distnum; r++) {
		for (int c=0; c<distdim; c++) {
			sprintf(out,"%20g ",gmg(dist_params,r,c) );
			datadist << out;
		}
		datadist << endl;
	}
	datadist.close();
	

	char com[80];
	char md5_33[81];
	int md;
	string buf;

	strcpy(com,"md5sum ");
	strcat(com,"md5.tmp ");
	strcat(com,"> md5sum.tmp");
	md = system(com);
	strcpy(com,"debug: you're right");

	ifstream md5tmpfile;
	md5tmpfile.open("md5sum.tmp",ios::in);
	getline(md5tmpfile,buf);
//	sscanf(buf.c_str(),"%s",datadist_md5);
	sscanf(buf.c_str(),"%s",md5_33);
	strncpy(datadist_md5,md5_33,33);
	md5tmpfile.close();

	return(0);
}

int MASSB::write_dist2data_matrix_records( ofstream *d2dm ) {
	// Provide for a check of matrix element integrity, i.e. we'll want to
	// catch the error if a matrix file calculated for one set of data and
	// distribution parameters are used with other data and or distributions

	char str[21];
	*d2dm << "DATADIST_MD5CHECK " << datadist_md5 << endl;
	for (int r=0; r<datanum; r++) {
		for (int c=0; c<distnum; c++) {
			sprintf(str,"%20g\n",gmg(dist2data,r,c) );
			*d2dm << str;
		}
	}
	return(0);
}

int MASSB::write_dist2data_matrix() {
	// Provide for a check of matrix element integrity, i.e. we'll want to
	// catch the error if a matrix file calculated for one set of data and
	// distribution parameters are used with other data and or distributions

	ifstream md5tmpfile;
	string buf;
	md5tmpfile.open("md5sum.tmp",ios::in);
	getline(md5tmpfile,buf);
	sscanf(buf.c_str(),"%s",datadist_md5);
	md5tmpfile.close();

	ofstream d2dm;
	const char * ext = ".d2dm";
	open_write_file(ext, &d2dm);

	d2dm << "DETAILS_LOG " << logfilename << endl;
	write_dist2data_matrix_records( &d2dm );

	return(0);
}

int MASSB::read_dist2data_matrix_records( ifstream *d2dm ) {
	
	string buf;
	dist2data = gsl_matrix_calloc( datanum, distnum ); 

	double me;
	for (int r=0; r<datanum; r++) {
		for (int c=0; c<distnum; c++) {
			getline(*d2dm,buf);
			sscanf(buf.c_str(),"%le",&me);
			gms(dist2data,r,c, me);
		}
	}

	return(0);
}

int MASSB::read_dist2data_matrix() {
	ifstream d2dm;
	string buf;
	
	d2dm.open( d2dmfile, ios::in );
	if ( !d2dm.is_open() )	{
		cout << "Can't open d2dmfile " << d2dmfile << endl;
		exit(0);
	}
	getline(d2dm,buf);	// throw this one away
	// need to make this more general if we ever find the case that
	// the ndist file needs to report more than DATANUM and DATADIM
	for (int j=0; j<1; j++) {
		getline(d2dm,buf);
		parseinputline(buf);
	}

	get_datadist_md5();
	if ( strcmp(datadist_matrix_md5,datadist_md5) ) {
		cerr << "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *" << endl;
		cerr << "* matrix file was calculated with a different configuration *" << endl;
		cerr << "* than those given in DISTFILE and DATAFILE                 *" << endl;
		cerr << "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *" << endl;
		cerr << "    matrix file md5: " << datadist_matrix_md5 << endl;
		cerr << "data/dist files md5: " << datadist_md5 << endl;
		exit(0);
	}

	read_dist2data_matrix_records( &d2dm );	

//	presetadjdensity = 0;
//	for (int cm=0; cm<distnum; cm++) 
//		presetadjdensity += gvg(&fixedpopulation.vector,cm);
//	lfile << "total pre-set density: " << presetadjdensity << endl;
	gsl_blas_dgemv(CblasNoTrans, 1.0, dist2data, 
		&fixedpopulation.vector, 0.0, &calcfix.vector );

	write_ndata();
	return(0);
}

double MASSB::calc_state() {
	double ret;
//	if ( !strcmp( statistics, "boltzmann") ) {
		ret = calc_bs_state();
//	} 

	return(ret);
}

double MASSB::calc_bs_state() {
	// Use Lagrange multipliers to calculate the density distribution

	gsl_vector * sumvect = gsl_vector_alloc(distnum);
	gsl_vector * expvect = gsl_vector_alloc(distnum);
	gsl_vector * ones = gsl_vector_alloc(distnum);
	gsl_vector_set_all(ones,1.0);

	// Find partition coefficient (Z),
	double Zz;
	gsl_blas_dgemv(CblasTrans, 1.0, dist2data, lagrange, 0.0, sumvect);
	for (int n=0; n<distnum; n++) gsl_vector_set(expvect,n, exp( gvg(sumvect,n)));
	gsl_blas_ddot(expvect,ones,&Zz);
	Z = Zz;
//if (cpuid == 0) {cerr << "Z =" << Z << endl;}

	// and density at each point in distribution 
	gsl_vector_memcpy( &population.vector,expvect);
	gsl_vector_scale( &population.vector,1.0/Z);

	// as well as the corresponding sideband intensities for the data
	gsl_blas_dgemv(CblasNoTrans, 1.0, dist2data, &population.vector, 0.0, &calcfit.vector );

	// adj data to exclude the components specified in the adjustment file
	gsl_vector_memcpy(&residual.vector, &datanamp.vector );

	gsl_blas_daxpy(-1.0,&calcfix.vector, &residual.vector );
	gsl_vector_scale( &residual.vector, 1.0/(1.0-presetadjdensity) );

	gsl_blas_daxpy(-1.0, &calcfit.vector, &residual.vector );
	gsl_vector_scale( &residual.vector, -1.0);

	gsl_vector * onesdatanum = gsl_vector_alloc(datanum);
	gsl_vector_set_all(onesdatanum,1.0);
	gsl_blas_ddot(&residual.vector,onesdatanum,&avgdev);
	avgdev *= 1.0/datanum;

	gsl_vector * tmp = gsl_vector_alloc(datanum);
	gsl_vector_memcpy(tmp,&residual.vector);
//	gsl_vector_div(tmp,sidebandrelnerr);	
//	gsl_vector_div(sidebandres,sidebandrelnerr);	
	
	chi2 = pow( gsl_blas_dnrm2( tmp ), 2.0);
	chi2 /= double(datanum);

	gsl_vector_free(sumvect);
	gsl_vector_free(expvect);
	gsl_vector_free(ones);
	gsl_vector_free(onesdatanum);
	gsl_vector_free(tmp);

	entropy = 0;
	for (int i=0; i<distnum; i++) {
		if ( gvg(&population.vector,i) > 1E-12 ) {
			entropy -= gvg(&population.vector,i) * log( gvg(&population.vector,i) );
		}
	}
	entropy /= log(distnum);

	return chi2;
//	return (pow( 500.0*pow(chi2,2.0) + pow(entropy,2.0), 0.5 ) );
}

int MASSB::inputpar( char *key, char *val) {
	char str[256];
	if (! strcmp(key, "DATAFILE") ) {
		strcpy(datafile,val); 
		sprintf(str,"INPUT DATAFILE (object.datafile) = %s\n",datafile);
		lfile << str << flush;
	} else if (! strcmp(key, "DISTFILE") ) {
		strcpy(distfile,val); 
		sprintf(str,"INPUT DATAFILE (object.distfile) = %s\n",distfile);
		lfile << str << flush;
	} else if (! strcmp(key, "INFILENAME") ) {
		strcpy(infilename,val); 
		sprintf(str,"INPUT INFILENAME (object.infilename) = %s\n",infilename);
		lfile << str << flush;
	} else if (! strcmp(key, "OUTFILENAME") ) {
		strcpy(outfilename,val); 
		sprintf(str,"INPUT OUTFILENAME (object.outfilename) = %s\n",outfilename);
		lfile << str << flush;
	} else if (! strcmp(key, "D2DMFILE") ) {
		strcpy(d2dmfile,val); 
		sprintf(str,"INPUT D2DMFILE (object.d2dmfile) = %s\n",d2dmfile);
		lfile << str << flush;
	} else if (! strcmp(key, "DATANORMDIM") )	{
		ssf_int( val, &datanormdim);
		sprintf(str,"INPUT DATANORMDIM (object.datanormdim) = %i\n",datanormdim);
		lfile << str << flush;
	} else if (! strcmp(key, "NDATAFORMAT") ) {
		strcpy(ndataformat,val); 
		sprintf(str,"INPUT DATAFILE (object.ndataformat) = %s\n",ndataformat);
		lfile << str << flush;
	} else if (! strcmp(key, "DISTPARAMS") )	{
		ssf_doublearray( val, &distdim, &rawdistparamlimits);
		sprintf(str,"INPUT DISTPARAMS (object.rawdistparams) = %s\n",val);
		lfile << str << flush;
	} else if (! strcmp(key, "PARAMS1") )	{
		ssf_doublearray( val, &datadim1, &params1);
		sprintf(str,"INPUT PARAMS1 (object.params1) = %s:\n",val);
		lfile << str << flush;
		for (int i=0; i<datadim1; i++) {
			sprintf(str,"%f ",params1[i]);
			lfile << str << flush;
		}
		lfile << endl;
	} else if (! strcmp(key, "PARAMS2") )	{
		ssf_doublearray( val, &datadim2, &params2);
		sprintf(str,"INPUT PARAMS2 (object.params2) = %s\n",val);
		lfile << str << flush;
		for (int i=0; i<datadim2; i++) {
			sprintf(str,"%f ",params2[i]);
			lfile << str << flush;
		}
		lfile << endl;
	} else if (! strcmp(key, "DATAPARAMS") )	{
		ssf_doublearray( val, &datadim, &rawdataparamlimits);
		sprintf(str,"INPUT DATAPARAMS (object.rawdataparams) = %s\n",val);
		lfile << str << flush;
	} else if (! strcmp(key, "SIGMA") )	{
		ssf_double( val, &sigma);
		sprintf(str,"INPUT SIGMA (object.sigma) = %f\n",sigma);
		lfile << str << flush;
	} else if (! strcmp(key, "PP2PI") )	{
		ssf_int( val, &pp2pi);
		sprintf(str,"INPUT PP2PI (object.pp2pi) = %i\n",pp2pi);
		lfile << str << flush;
	} else if (! strcmp(key, "ETAT") )	{
		ssf_int( val, &etat);
		sprintf(str,"INPUT ETAT (object.etat) = %i\n",etat);
		lfile << str << flush;
	} else if (! strcmp(key, "DISTNUM") )	{
		ssf_int( val, &distnum);
		sprintf(str,"INPUT DISTNUM (object.distnum) = %i\n",distnum);
		lfile << str << flush;
	} else if (! strcmp(key, "DISTDIM") )	{
		ssf_int( val, &distdim);
		sprintf(str,"INPUT DISTDIM (object.distdim) = %i\n",distdim);
		lfile << str << flush;
	} else if (! strcmp(key, "DATANUM") )	{
		ssf_int( val, &datanum);
		sprintf(str,"INPUT DATANUM (object.datanum) = %i\n",datanum);
		lfile << str << flush;
	} else if (! strcmp(key, "DATADIM") )	{
		ssf_int( val, &datadim);
		sprintf(str,"INPUT DATADIM (object.datadim) = %i\n",datadim);
		lfile << str << flush;
	} else if (! strcmp(key, "NUMNORM") )	{
		ssf_int( val, &num_norms);
		sprintf(str,"INPUT NUMNORMS (object.num_norms) = %i\n",num_norms);
		lfile << str << flush;
	} else if (! strcmp(key, "POWDEROPT") )	{
		ssf_int( val, &powderopt);
		sprintf(str,"INPUT POWDEROPT (object.powderopt) = %i\n",powderopt);
		lfile << str << flush;
	} else if (! strcmp(key, "CRYSTFILE") ) {
		strcpy(crystname,val); 
		sprintf(str,"INPUT CRYSTFILE (object.crystname) = %s\n",crystname);
		lfile << str << flush;
	} else if (! strcmp(key, "DETAILS_LOG") ) {
		strcpy(details_log,val); 
		sprintf(str,"INPUT DETAILS_LOG (object.details_log) = %s\n",details_log);
		lfile << str << flush;
	} else if (! strcmp(key, "DATADIST_MD5CHECK") ) {
		strcpy(datadist_matrix_md5,val); 
		sprintf(str,"INPUT DATADIST_MD5CHECK (object.datadist_matrix_md5) = %s\n",
			datadist_matrix_md5);
		lfile << str << flush;
	} else { 
		cerr << "I Don't Recognize " << key << endl;
		exit(0);
	}
	return(0);
}

int MASSB::parseinputs( int c, char *v[] ) {
	strcpy(code,v[0]);
	lfile << "***********************\n";
	lfile << "CODE: " << code << endl;
	simpletimestamp(&lfile); 
	lfile << "***********************\n";
	for (int i=1; i<c; i+=2) {
		inputpar( v[i], v[i+1] );
	}
	return(0);
}	

int MASSB::parseinputline( string s ) {
	char key[80];
	char val[120];
	sscanf(s.c_str(),"%s %s",key,val);
	inputpar( key, val);
	return(0);
}	

int MASSB::open_read_file (const char *ext, ifstream *infile ) {
    char filename[256];
	outpath = getenv("OUTPATH");
	outbase = getenv("OUTBASE");

    strcpy(filename,outpath);
    strcat(filename,"/" );
    strcat(filename,outbase);
    strcat(filename,"/" );
    strcat(filename,outbase);
    strcat(filename,ext);
    
	infile->open(filename, ios::in);
	if ( !infile->is_open() ) {
		cerr << "Could not open input file " << filename << endl;
		exit(0);
	}
	return(0);
}

int MASSB::open_write_file (const char *ext, ofstream *outfile ) {
    char filename[256];
	outpath = getenv("OUTPATH");
	outbase = getenv("OUTBASE");

    strcpy(filename,outpath);
    strcat(filename,"/" );
    strcat(filename,outbase);
    strcat(filename,"/" );
    strcat(filename,outbase);
    strcat(filename,ext);
    
	outfile->open(filename, ios::out);
	if ( !outfile->is_open() ) {
		cerr << "Could not open input file " << filename << endl;
		exit(0);
	}
	return(0);
}

int MASSB::open_append_file ( const char *ext, ofstream *appfile, char *name ) {
	outpath = getenv("OUTPATH");
	outbase = getenv("OUTBASE");

    strcpy(name,outpath);
    strcat(name,"/" );
    strcat(name,outbase);
    strcat(name,"/" );
    strcat(name,outbase);
    strcat(name,ext);
    
	appfile->open(name, ios::app);
	if ( !appfile->is_open() ) {
		cerr << "Could not open input file " << name << endl;
		exit(0);
	}
	return(0);
}
