#ifndef MASSB_INCLUDED_
#define MASSB_INCLUDED_


#include <sys/stat.h>
#include "inputs.h"
#include "utils.h"
#include "cputhreadtime.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
using namespace std;

class MASSB {
public:
	MASSB();
	~MASSB();
	MASSB( const MASSB& );
	int copy_min( const MASSB& ); 

	int append_logfile();
	int read_data();
	int write_ndata();
	int norm_data();
	int read_ndata();
	int read_full_bsetat();
	int write_full_bsetat();
	int parseinputs( int, char ** );
	int parseinputline( string );
	int set_dist();
	int set_data();
	int set_data_explicit();
	int read_dist();
	int write_dist();
	int calc_dist2data_crystfile_abg();
	int calc_dist2data_crystfile_ab();
	int calc_dist2data_openmp();
//	int calc_dist2data_mpi();
	int calc_dist2data();
	int write_dist2data_matrix();
	int read_dist2data_matrix();
	int get_datadist_md5();
	int open_append_file ( const char *, ofstream *, char * );
	int write_ndata_records( ofstream *);
	int write_dist_records( ofstream *);
	int pre2dat();


	int set_preset();
	double calc_state();
	double calc_bs_state();
	int calc_jacobian();
	int log_jacobian();
	int find_state();
	int write_out();
	int write_sim();
	int write_mix();
	int set_visual();
	int write_gp_eta0();
	int write_gp_2D();
	int update_visual();
	int update_log( char *);

	int pp2pi;				// number angles per 2pi for numerical integration
	int distnum;			// total num distribution parameter permutations
	int distdim;			// number of distribution parameters
	int  datadim;			// data dimension; how many experimental parameters
	int  datadim1;			// data dimension; how many experimental parameters
	int  datadim2;			// data dimension; how many experimental parameters
	int  datanum;			// number of total experimental parameter combinations
	int num_norms;			// number of differnt data subsets (e.g. dif't mas rates)
	int  datanormdim;		// data dimension for which data should be normalized (e.g. mas)
	double sigma;			// error value to use when shifting preset calculation to data
	
	gsl_matrix * dist_params;	// all distribution parameter permutations
	gsl_matrix * distribution;	// density and errors of distribution
	gsl_vector_view population;	// fit density column from distribution matrix
	gsl_vector_view fixedpopulation; // fixed density column from distribution matrix
	gsl_matrix * data_params;

	gsl_matrix * dist2data;		// the Herzfeld-Berger matrix 

	gsl_vector_view residual;
	gsl_vector_view calcfit;
	gsl_vector_view calcfix;
	gsl_vector_view datanamp;
	
	CPUThreadTime T;		// time and thread efficiency tracking
	ofstream lfile;			// logfile 

	char datadist_md5[33];
	// values take iterative specifications from the controlling execuatable
	char searchmethod[20];
	char searchfile[80];	
	int cpuid;
	int etat;

	char bsetatfile[256];
	char datafile[256];		// speed index amp error in a file
	char distfile[256];		// distribution parameters etc file
	char infilename[256];		// allpurpose file name
	char outfilename[256];		// allpurpose file name
	
	// population distribution, and other configuration structures
	gsl_vector * lagrange;		// the configuration of Lagrange multipliers
	double chi2;				// essentially just the rmsd
	double entropy;				// the Shannon entropy of the population distri
	double avgdev;				// the average deviation of the fit

	char * outpath;		// path to output directory (subdir will be created)
	char * outbase;		// basename to be used for output files
	char * safety;		// whether to prompt for overwrite confirmation

	// just a simple utility to check on class object copy modification
	int check;

	// input variables
	char mode[4];			// fit or sim

	// a kludge because gsl_function routines can't otherwise access Class members
	int copy_to_public();	// copy private data structures to public ones
	int copy_from_public(); // copy public data structures back to private ones

	// public versions of the primary private versions below
	// for access from unfriendlies.
	int pcms;
	int pnsb;					// total number of sidebands over all MAS rates
	gsl_vector * psideband;		// sideband index (e.g. -2, -1, 0, 1, 2, etc)
	gsl_vector * psidebandnamp;	// normalize sb amps (norm to 1 for ea MAS rate)
	gsl_vector * psidebandnerr;	// errors in normalized amplitudes
	gsl_vector * psidebandcalc;	// calculated sideband instensities for given
								// population distribution
	gsl_vector * psidebandres;	// residual between calc and normalized exp amps
	gsl_matrix * phb;			// the Herzfeld-Berger matrix 
	gsl_vector * ppopulation;	// the population distribution
	gsl_vector * plagrange;		// the configuration of Lagrange multipliers
	gsl_matrix * pJ;			// Jacobian

private:
	friend class SA_search;

	char logfilename[256];
	char d2dmfile[256];
	
//	int ssf_doublearray( char [], int *, double *);
	int inputpar( char *, char *);
	void setdistparamrecurloop ( int, int *, double * );
	void setdataparamrecurloop ( int, int *, double * );
	int open_write_file ( const char *, ofstream * );
	int open_read_file ( const char *, ifstream * );

	int read_ndata_records( ifstream *);
	int read_dist_records( ifstream *);
	int read_dist2data_matrix_records( ifstream *);
	int write_ndata_header( ofstream *);
	int write_dist_header( ofstream *);
	int write_dist2data_matrix_records( ofstream *);

	void CMI2DE( int, double*, double* );
	void DE2CMI( double, double, int* );
	int write_state();			// write state file
	int read_state();			// read state file

	// input variables
	char code[80];

	char ndataformat[256];
	gsl_matrix * data_notnorm;
	gsl_matrix * data_norm;
	gsl_vector * norm_params;	// vector of different param vals used in data, each one independently normalized

	int * dist_dim_num;		// number of values of each parameter
	int * data_dim_num;		// number of values of each parameter
	double * rawdistparamlimits;	// min max res triplet for each parameter
	double * rawdataparamlimits;	// min max res triplet for each parameter
	double * params1;				// for use with explicit setting of data or dist parameters
	double * params2;
	char datadist_matrix_md5[33];
	char details_log[256];
	
	char statefile[256];		// file which holds info from previous state
	char crystname[80];			// full path to crystal file if powderopt != 0
	char statistics[14];	// statistics to use for analysis


	char presetfile[256];	// name of file with preset density
	int threads;				// number of threads requested
	int powderopt;			// 0 for full calc, 2 or 3 for 2-angle or 3-angle
							// crystal file, respectively
	double fixmas;			// mas rate for simulations
	double fixdelta;		// delta for simulations
	double fixeta;			// eta for simulations
	int fixminus;			// furthest negative sideband index value for sims
	int fixplus;			// furthest positive sideband index value for sims
	double deltamax;		// maximum delta for BS reconstructions
	double deltamin;		// minimum delta for BS reconstructions
	double deltares;		// resolution in delta for BS reconstructions
	double etamax;			// maximum eta for BS reconstructions
	double etamin;			// minimum eta for BS reconstructions
	double etares;			// resolution in eta for BS reconstructions
	

	// MAS manifold data structures
								// population distribution
	gsl_vector * sidebandres;	// residual between calc and normalized exp amps

	gsl_vector * presetpop;		// pre-set population for forced density or sim
	gsl_vector * presetnamp;	// normalized amplitudes corresponding to 

	gsl_matrix * J;				// jacobian -- d(sideband)/d(lambda)
								// pre-set population
	double presetadjdensity;		// total of preset density adjustment
	int * sb_at_speed;			// number of sidebands at each different MAS speed

	// Herzfeld-Berger matrix structures
	int num_mas;				// number of spinning speeds included in data
	int cms;					// number of delta x eta configurations
	int numdelta;				// number of different delta values
	int numeta;					// number of different eta values
								// sideband index varies along rows
								// (delta, eta) configuration varies along cols


	// Boltzmann Statistics specifications
	double janneal;				// attenuation factor for jacobian calculation
	double jscale;				// scaling of whole jacobian
	double Z;					// the partition coefficient

	// general truths
	double pi2;

	// 
	char gpcommand[256];			// command to update gp postscript;

	
};

#endif
