#include "sa_search.h"
#include "utils.h"

SA_search::SA_search( MASSB &P ) {
//	P.T.seconds = time(NULL);
//	P.lfile << "Beginning Boltzmann ME: " << ctime(&P.T.seconds) << endl;
//	cout << "Beginning Boltzmann ME:\n";

//	P.set_visual();
}

int SA_search::inputpar( char *key, char *val, bool report) {
	char str[256];
	if (! strcmp(key, "MASSBFILE") ) {
		strcpy(massbfile,val); 
		sprintf(str,"INPUT MASSBFILE (object.massbfile) = %s\n",massbfile);
	} else if (! strcmp(key, "MINTEMP") )	{
		ssf_double( val, &mintemp);
		sprintf(str,"INPUT MINTEMP (object.mintemp) = %g\n",mintemp);
	} else if (! strcmp(key, "INITTEMP") )	{
		ssf_double( val, &inittemp);
		sprintf(str,"INPUT INITTEMP (object.inittemp) = %g\n",inittemp);
	} else if (! strcmp(key, "MAXSTEP") )	{
		ssf_double( val, &maxstep);
		sprintf(str,"INPUT MAXSTEP (object.maxstep) = %g\n",maxstep);
	} else if (! strcmp(key, "BOLTZMANN") )	{
		ssf_double( val, &boltzmann);
		sprintf(str,"INPUT BOLTZMANN (object.boltzmann) = %g\n",boltzmann);
	} else if (! strcmp(key, "ATTENTEMP") )	{
		ssf_double( val, &attentemp);
		sprintf(str,"INPUT ATTENTEMP (object.attentemp) = %g\n",attentemp);
	} else if (! strcmp(key, "ETAT") )	{
		ssf_int( val, &etat);
		sprintf(str,"INPUT ETAT (object.etat) = %i\n",etat);
	} else if (! strcmp(key, "ITERT") )	{
		ssf_int( val, &iterT);
		sprintf(str,"INPUT ITERT (object.iterT) = %i\n",iterT);
	} else { 
		cerr << "I Don't Recognize " << key << endl;
		exit(0);
	}
	if (report == 0) {
		*lfile << str << flush;
	}
	// or make the above zero, and pass the tid into these functions
	// such that only master writes to lfile. See whether this would 
	// provide economy down the track or whether it's only meaningful
	// here.
	return(0);
}

// may not need this one
int SA_search::parseinputline( string s ) {
	char key[80];
	char val[80];
	sscanf(s.c_str(),"%s %s",key,val);
	inputpar( key, val, 1);
	return(0);
}

	
int SA_search::parseinputs( int c, char *v[] ) {
	char code[48];
	strcpy(code,v[0]);

	if ( cpuid == 0 ) {	
		*lfile << "***********************\n";
		*lfile << "CODE: " << code << endl;
		simpletimestamp(lfile);
		*lfile << "***********************\n";
	}
	for (int i=1; i<c; i+=2) {
		inputpar( v[i], v[i+1], cpuid );
	}
	return(0);
}	

int SA_search::Lagrange_siman( MASSB &P ) {
	char str[80];
//	P.T.seconds = time(NULL);
//	P.lfile << "Beginning Lagrange_siman: " << ctime(&P.T.seconds) << endl;
//	cout << "Beginning Lagrange_siman:\n";

//	Gnuplot gsb = Gnuplot("impulses");

	int ranseed = gen_random_seed( sjid );
	const gsl_rng_type * Tr;
	gsl_rng * r;
	gsl_rng_env_setup();
//	Tr = gsl_rng_default;
	Tr = gsl_rng_taus2;
	r = gsl_rng_alloc(Tr);
	gsl_rng_set(r, ranseed );

	// THIS TO CHANGE AFTER STATE FUNCTIONALITY INCLUDED IN MASSB     
	char si[3], ti[3];
	sprintf(si,".%d",etat);
	sprintf(ti,".%d",cpuid);

	// different files for different threads -- append cpuid too.
	ofstream saout;	
	char saname[179];
	sprintf(saname,"%s/%s/%s.t%02i.sa.%i",
		P.outpath,P.outbase,P.outbase,cpuid,etat);
	saout.open(saname, ios::out);

	char format[79];
	strcpy(format,"%6d %9ld %12g %10g %8.5f %4.1f %4.1f %4.1f ");
	sprintf(str,"#%5s %9s %12s %10s %8s %4s %4s %4s\n",
		"iter","evals","temp","chi2","H ","#lss","#acc","$rej");
	saout << str;

	MASSB x(P), new_x(P), best_x(P);  

	double E, new_E, best_E;
	double H, new_H, best_H;
	int i, done;
	double T = inittemp;
	double u;
	int n_evals = 0, n_iter = 0, n_accepts, n_rejects, n_eless;

	E = P.calc_state();
	H = P.entropy;
	best_E = E;
	best_H = H;
	
//	This needs to be sent back to the master thread for output to P's lfile,
//	or perhaps write to temp files and let the master thread concatenate
// 	into P's lfile in the end.

	if (cpuid == 0) {
		sprintf(str,"Initial Chi2 Energy: %12.6g Inital Entropy: %12.6g\n",E,H); 
		*lfile << str << flush;
	}

//	sprintf(str,"%5d B %8g S %8g Temp: %8g-%8g by %6.4f\n",
//		iterT,boltzmann,maxstep,T,mintemp,attentemp);
//	P.lfile << str;
//	cout << str;
	
	done = 0;

	while (!done) {
		n_accepts = 0;
		n_rejects = 0;
		n_eless = 0;
    	for (i = 0; i < iterT; ++i) {
			new_x.copy_min(x);

			// take a step
			for (int lm=0; lm<P.datanum; lm++) {
				u = gsl_rng_uniform(r);
				gsl_vector_set( new_x.lagrange, lm, u * 2.0 * maxstep - 
					maxstep + gvg( x.lagrange, lm) );
			}
			new_E = new_x.calc_state();
			new_H = new_x.entropy;
/*
			sprintf(str,"E %12.6g L: ",new_E); cout << str;
			for (int lm=0; lm<P.nsb; lm++) {
				sprintf(str,"%12.6g ",gvg(new_x.lagrange,lm));
				cout << str;
			}
			cout << endl;
*/

			// if we fell over the edge, report it
			if ( !isfinite(new_E) ) {
				for (int bs = 0; bs<62; bs++ ) cout << "\b";
//				cout << "BANG. Chi2 is not finite at temp " << T ;
				++n_rejects;
				continue;
			}
		
			// if energy of configuration following step is the lowest yet, 
			// replace best known configuration with it
			if(new_E <= best_E){
				best_x.copy_min(new_x);
	        	best_E=new_E;
				best_H=new_H;
				P.copy_min(new_x);
			}
			// keep track of finite Ef() evaluations 
			++n_evals;

			// now take the crucial step: see if the new point is accepted
			// or not, as determined by the boltzman probability. If step is
			// is taken, write it to the .la file together with the 
			// Lagrange multiplier configuration.
			// if the energy of the step is lower than the current config
			// take the step. Also take a higher energy step, with a random
			// probability.

			if (new_E < E) {
	    	   	// take a step 
				x.copy_min(new_x);
				E = new_E;
				H = new_H;
				++n_eless;
			} else if (gsl_rng_uniform(r) < exp (-(new_E-E)/(boltzmann*T)) ) {
				// also take a step 
				x.copy_min(new_x);
				E = new_E;
				H = new_H;
				++n_accepts;
			} else {
				++n_rejects;
			}
    	}
	
		//print report
		sprintf(str,format,n_iter,n_evals,T,x.chi2,x.entropy, 
			100.0*double(n_eless)/double(iterT), 
			100.0*double(n_accepts)/double(iterT), 
			100.0*double(n_rejects)/double(iterT) );
		saout << str;
		for (int i=0; i<P.datanum; i++) {
			sprintf(str,"%10g ", gvg(x.lagrange,i) );
			saout << str;
		}
		saout << endl;
	
		// Some updates to screen bypass for parallel computing
//		if ( double(n_iter)/5.0 - double(n_iter/5) < 1E-6 ) {
//			for (int bs = 0; bs<96; bs++ ) cout << "\b";
//			printf("Temp %10.8f Current/Best E %14.6g/%14.6g Current/Best H %8.5f/%8.5f",T,E,best_E,H,best_H);
//			P.write_out();
//			P.update_visual();
//		}
	
	    // apply the cooling schedule to the temperature 
		T /= attentemp;
		++n_iter;
		if (T < mintemp) done = 1;
	}

	// This all needs to be structures such that the master thread collects results
	// and decides which is best, and writes this to a new full_massb file, and
	// as well as a single lagrange multiplier file.

	//at the end, copy the result onto the initial point
	P.copy_min(best_x);

	P.calc_state();

//	cout << endl;
//	P.lfile << "CHI2   " <<	P.chi2 << endl;
//	P.lfile << "AVGDEV " <<  P.avgdev << endl;
//	P.lfile << "ENTROPY " << P.entropy << endl;

//	cout << "CHI2   " << 	P.chi2 << endl;
//	cout << "AVGDEV " <<  P.avgdev << endl;
//	cout << "ENTROPY " << P.entropy << endl;

	saout << flush;
	saout.close();
//	P.write_out();

//	P.T.seconds = time(NULL);
//	P.lfile << "Ending  Lagrange_siman: " << ctime(&P.T.seconds) << endl;
//	cout << "Ending  Lagrange_siman: " << ctime(&P.T.seconds) << endl;
	
	gsl_rng_free(r);

//	P.write_state();
	return(0);
}

