#!/bin/bash -l

export BIN="../../Src"
export OUTPATH="`pwd`"
export OWSAFETY="nosafe"

export OUTBASE="sample_data_1200b_fit"
echo $OUTBASE

init=1
noise=1
fit=1

if (( $init ))
	then
echo "initializing ..."

rm -r $OUTBASE

$BIN/init_spectralfit_data\
	SPECTRUMFILE "tmp1-1200b.dat"\
	INTEGRALFILE "tmp2-1200b.dat"\
	DERIVATIVEFILE "tmp3-1200b.dat"\
	APPENDOUTFILE "-init"\
	MAS 1200\
	MINSB -7\
	MAXSB 8\
	FWHM 500\
	FWHMISO 100\
	CENTRE 571\
	CENTREISO 750\
	AMPISO 3000\
	FWHMISO2 100\
	CENTREISO2 620\
	AMPISO2 3000\
	ALPHA 1

$BIN/fit_gnuplot\
	APPENDOUTFILE "-init"
fi

if (( $noise ))
	then
echo "analyzing baseline ..."

$BIN/analyze_baseline\
	DATAFITFILE $OUTBASE/${OUTBASE}-init.lfit\
	NMINFREQ 15000\
	NMAXFREQ 15200
fi

if (( $fit ))
	then
echo "fitting ..."

echo "... spectrum only ..."
$BIN/fit_spectral_data\
	DATAFITFILE $OUTBASE/${OUTBASE}-init.lfit\
	APPENDOUTFILE "-4a"\
	FITCONFIG 4\
	FITAMPS 1\
	FITAMPISO 1\
	FITFWHM 1\
	FITFWHMISO 1\
	FITCENTRE 1\
	FITCENTREISO 1\
	FITAMPISO2 1\
	FITFWHMISO2 1\
	FITCENTREISO2 1\
	FITALPHA 0\
	MAXITER 500\
	SIGMASPE 0.04\
	SIGMAINT 0.0001\
	SIGMADER 0.008

echo " ... spectrum, integral and derivative ..."
$BIN/fit_spectral_data\
	DATAFITFILE $OUTBASE/${OUTBASE}-4a.lfit\
	APPENDOUTFILE "-7a"\
	FITCONFIG 7\
	FITAMPS 1\
	FITAMPISO 1\
	FITFWHM 1\
	FITFWHMISO 1\
	FITCENTRE 1\
	FITCENTREISO 1\
	FITAMPISO2 1\
	FITFWHMISO2 1\
	FITCENTREISO2 1\
	FITALPHA 0\
	MAXITER 500\
	SIGMASPE 0.04\
	SIGMAINT 0.0001\
	SIGMADER 0.008

$BIN/fit_gnuplot\
	APPENDOUTFILE "-7a"
fi
