#include "sbfit.h"
using namespace std;

int main ( int argc, char *argv[]  ) {

	SBFIT S;
	S.append_logfile(); 
	
	int check = 131071;
	for (int i=1; i<argc; i+=2) {
		if ( !strcmp(argv[i],"DATAFITFILE" ) ) check -= 1;
		else if ( !strcmp(argv[i],"APPENDOUTFILE" ) ) check -= 2;
		else if ( !strcmp(argv[i],"FITCONFIG" ) ) check -= 4;
		else if ( !strcmp(argv[i],"FITAMPS" ) ) check -= 8;
		else if ( !strcmp(argv[i],"FITAMPISO" ) ) check -= 16; // 
		else if ( !strcmp(argv[i],"FITFWHM" ) ) check -= 32;
		else if ( !strcmp(argv[i],"FITFWHMISO" ) ) check -= 64; // 
		else if ( !strcmp(argv[i],"FITCENTRE" ) ) check -= 128;
		else if ( !strcmp(argv[i],"FITCENTREISO" ) ) check -= 256; //
		else if ( !strcmp(argv[i],"MAXITER" ) ) check -= 512;
		else if ( !strcmp(argv[i],"SIGMASPE" ) ) check -= 1024;
		else if ( !strcmp(argv[i],"SIGMAINT" ) ) check -= 2048;
		else if ( !strcmp(argv[i],"SIGMADER" ) ) check -= 4096;
		else if ( !strcmp(argv[i],"FITCENTREISO2" ) ) check -= 8192; //
		else if ( !strcmp(argv[i],"FITFWHMISO2" ) ) check -= 16384; //
		else if ( !strcmp(argv[i],"FITAMPISO2" ) ) check -= 32768; //
		else if ( !strcmp(argv[i],"FITALPHA" ) ) check -= 65536; //
	}
	if (check == 122880) {
		cerr << "[fit_spectral_data] check = " << check << endl;
		cerr << "only one isotropic peak to be optimised in fit" << endl;
	} else if (check == 123216) {
		cerr << "[fit_spectral_data] check = " << check << endl;
		cerr << "no isotropic peaks to be optimised in fit" << endl;
	} else if (check != 0) {
		cerr << "[fit_spectral_data] check = " << check 
			<< " Need to specify: " << endl;
		cerr << "[1] DATAFITFILE (raw & fit spectrum, integral, derivative)" 
			<< endl;
		cerr << "[2] APPENDOUTFILE (\"append to unoriented output filename\")"
			<< endl;
		cerr << "[4] FITCONFIG (sum of 4 spectrum, 2 integral, 1 derivative)" 
			<< endl;
		cerr << "[8] FITAMPS (fit MAS amplitudes 0 or 1)" << endl;
		cerr << "[16] FITAMPISO (fit isotropic amplitude 0 or 1)" << endl;
		cerr << "[32] FITFWHM (fit full width half max of mas sidebands 0 or 1)"
			<< endl;
		cerr << "[64] FITFWHMISO (fit full width half max of isotropic 0 or 1)" 
			<< endl;
		cerr << "[128] FITCENTRE (fit centre freq of mas centreband 0 or 1)" 
			<< endl;
		cerr << "[256] FITCENTREISO (fit freq of isotropic peak 0 or 1)" 
			<< endl;
		cerr << "[512] MAXITER (maximum fitting iterations)" 
			<< endl;
		cerr << "[1024] SIGMASPE (rms of noise in spectrum)" 
			<< endl;
		cerr << "[2048] SIGMAINT (rms of noise in integral)" 
			<< endl;
		cerr << "[4096] SIGMADER (rms of noise in derivative)" 
			<< endl;
		cerr << "[8192] FITCENTREISO2 (fit freq of 2nd isotropic peak 0 or 1)" 
			<< endl;
		cerr << "[16384] FITFWHMISO2 (fit full width half max of 2nd isotropic 0 or 1)" 
			<< endl;
		cerr << "[32768] FITAMPISO2 (fit amplitude of 2nd isotropic peak 0 or 1)" 
			<< endl;
		cerr << "[65536] FITALPHA (not currently implemented: set to 0)" 
			<< endl;
		exit(0);
	}
	int suc;
	suc = S.parseinputs( argc, argv );
	if (suc) {
		cerr << "parseinputs failed" << endl;
		exit(2);
	}
	suc = S.read_datafit();
	if (suc) {
		cerr << "read_datafit failed" << endl;
		exit(2);
	}
	suc = S.nonlinfit();
	if (suc) {
		cerr << "nonlinfit failed" << endl;
		exit(2);
	}

	suc = S.calc_spectrum();
	if (suc) {
		cerr << "calc_spectrum failed" << endl;
		exit(2);
	}
	suc = S.calc_derivative();
	if (suc) {
		cerr << "calc_derivative failed" << endl;
		exit(2);
	}
	suc = S.calc_integral();
	if (suc) {
		cerr << "calc_integral failed" << endl;
		exit(2);
	}

	suc = S.write_datafit();
	if (suc) {
		cerr << "write_datafit failed" << endl;
		exit(2);
	}
	suc = S.write_memasdata();
	if (suc) {
		cerr << "write_memasdata failed" << endl;
		exit(2);
	}
	suc = S.write_lrntz();
	if (suc) {
		cerr << "write_lrntz failed" << endl;
		exit(2);
	}
	return(0);
}
