#include "sbfit.h"
using namespace std;

int main ( int argc, char *argv[]  ) {

	SBFIT S;
	S.append_logfile(); 
	
	int check = 65535;
	for (int i=1; i<argc; i+=2) {
		if ( !strcmp(argv[i],"SPECTRUMFILE" ) ) check -= 1;
		else if ( !strcmp(argv[i],"INTEGRALFILE" ) ) check -= 2;
		else if ( !strcmp(argv[i],"DERIVATIVEFILE" ) ) check -= 4;
		else if ( !strcmp(argv[i],"APPENDOUTFILE" ) ) check -= 8;
		else if ( !strcmp(argv[i],"MINSB" ) ) check -= 16;
		else if ( !strcmp(argv[i],"MAXSB" ) ) check -= 32;
		else if ( !strcmp(argv[i],"MAS" ) ) check -= 64;
		else if ( !strcmp(argv[i],"FWHM" ) ) check -= 128;
		else if ( !strcmp(argv[i],"FWHMISO" ) ) check -= 256;
		else if ( !strcmp(argv[i],"CENTRE" ) ) check -= 512;
		else if ( !strcmp(argv[i],"CENTREISO" ) ) check -= 1024;
		else if ( !strcmp(argv[i],"AMPISO" ) ) check -= 2048;
		else if ( !strcmp(argv[i],"CENTREISO2" ) ) check -= 4096;
		else if ( !strcmp(argv[i],"FWHMISO2" ) ) check -= 8192;
		else if ( !strcmp(argv[i],"AMPISO2" ) ) check -= 16384;
		else if ( !strcmp(argv[i],"ALPHA" ) ) check -= 32768;
	}
	if (check == 28672) {
		cerr << "[init_spectralfit_data] check = " << check << endl;
		cerr << "only one isotropic peak stipulated" << endl;
	} else if (check == 64768) {
		cerr << "[fit_spectral_data] check = " << check << endl;
		cerr << "no isotropic peaks stipulated" << endl;
	} else if (check != 0) {
		cerr << "[init_spectralfit_data] check = " << check 
			<< " Need to specify: " << endl;
		cerr << "[1] SPECTRUMFILE (2 cols: frequency intensity)" << endl;
		cerr << "[2] INTEGRALFILE (2 cols: frequency integral)" << endl;
		cerr << "[4] DERIVATIVEFILE (2 cols: frequency derivative)" << endl;
		cerr << "[8] APPENDOUTFILE (\"append to unoriented output filename\")"
			<< endl;
		cerr << "[16] MINSB (most negative sideband index)" << endl;
		cerr << "[32] MAXSB (most positive sideband index)" << endl;
		cerr << "[64] MAS (mas speed)" << endl;
		cerr << "[128] FWHM (full width half max of mas sidebands)" << endl;
		cerr << "[256] FWHMISO (full width half max of isotropic)" << endl;
		cerr << "[512] CENTRE (centre freq of mas centreband)" << endl;
		cerr << "[1024] CENTREISO (freq of isotropic peak)" << endl;
		cerr << "[2048] AMPISO (intensity of isotropic peak)" << endl;
		cerr << "[4096] CENTREISO2 (centre freq of 2nd isotropic peak)" << endl;
		cerr << "[8192] FWHMISO2 (full width half max of 2nd isotropic peak)" << endl;
		cerr << "[16384] AMPISO2 (intensity of 2nd isotropic peak)" << endl;
		cerr << "[32768] ALPHA (fit freq of isotropic peak 0 or 1)" << endl;

		exit(1);
	}

	int suc;
	suc = S.parseinputs( argc, argv );
	if (suc) {
		cerr << "parseinputs failed" << endl;
		exit(2);
	}
	suc = S.count_xydata( S.spectrumfile );
	if (suc) {
		cerr << "count_xydata failed" << endl;
		exit(2);
	}
	suc = S.init_data();
	if (suc) {
		cerr << "init_data failed" << endl;
		exit(2);
	}
	suc = S.read_spectrum();
	if (suc) {
		cerr << "read_spectrum failed" << endl;
		exit(2);
	}
	suc = S.read_integral();
	if (suc) {
		cerr << "read_integral failed" << endl;
		exit(2);
	}
	suc = S.read_derivative();
	if (suc) {
		cerr << "read_derivative failed" << endl;
		exit(2);
	}
	suc = S.init_fit_params();
	if (suc) {
		cerr << "init_fit_params failed" << endl;
		exit(2);
	}
	suc = S.calc_spectrum();
	if (suc) {
		cerr << "calc_spectrum failed" << endl;
		exit(2);
	}
	suc = S.calc_integral();
	if (suc) {
		cerr << "calc_integral failed" << endl;
		exit(2);
	}
	suc = S.calc_derivative();
	if (suc) {
		cerr << "calc_derivative failed" << endl;
		exit(2);
	}
	
	suc = S.write_datafit();
	if (suc) {
		cerr << "write_datafit failed" << endl;
		exit(2);
	}
	suc = S.write_lrntz();
	if (suc) {
		cerr << "write_lrntz failed" << endl;
		exit(2);
	}
	suc = S.write_memasdata();
	if (suc) {
		cerr << "write_memasdata failed" << endl;
		exit(2);
	}
	return(0);
}
