#include "inputs.h"

Inputs::Inputs( const char * infile ) {
	ifstream inparams;
	inparams.open( infile, ios::in );
	if ( !inparams.is_open() ) {
		cout << "Can't open input params from " << infile << endl;
		exit(0);
    }
	
	// read through data file once to count number of lines
	numlines = -1;
	string buf;
	while ( !inparams.eof() ) {
		getline(inparams,buf);
		if ( iscomment( buf ) ) continue;
        numlines++;
    }

	// Declare size of new filelines array
	filelines = new char[numlines][100];
	cid = new int[numlines];
	cid_index = new int[numlines];

	inparams.clear();
	inparams.seekg(0);

	for (int fl=0; fl<numlines; fl++) {
		getline(inparams,buf);
		if ( iscomment( buf ) ) {
			fl--;
			continue;
		}
		strcpy(	filelines[fl],buf.c_str() );
	}
	inparams.close();
}

Inputs::~Inputs () {
	delete intvals;
	delete[] intLABELS;
	delete doublevals;
	delete[] doubleLABELS;
	delete[] charvals;
	delete[] charLABELS;
	delete[] filelines;
	delete cid;
	delete cid_index;

}

int Inputs::alloc_arrays (int charnum, const char (*chararr)[20], 
	int intnum, const char (*intarr)[20], int doublenum, const char (*doublearr)[20]) {

	if (charnum > 0 ) {
		charLABELS = new char[charnum][20];
		charvals = new char[charnum][80];
		for (int i=0; i<charnum; i++) strcpy(charLABELS[i],chararr[i]);
	} else {
		charLABELS = 0;
		charvals = 0;
	}

	if (intnum > 0 ) {
		intLABELS = new char[intnum][20];
		intvals = new int [intnum];
		for (int i=0; i<intnum; i++) strcpy(intLABELS[i],intarr[i]);
	} else {
		intLABELS = 0;
		intvals = 0;
	}

	if (doublenum > 0 ) {
		doubleLABELS = new char [doublenum][20];
		doublevals = new double [doublenum];
		for (int i=0; i<doublenum; i++) strcpy(doubleLABELS[i],doublearr[i]);
	} else {
		doubleLABELS = 0;
		doublevals = 0;
	}

	int c = 0; int i = 0; int d = 0;

	char tmpchar20[20];
	char tmpchar80[80];
	int ptr; 
	bool found;
	int tmpint;	
	double tmpdouble;

	for (int fl=0; fl<numlines; fl++) {
		ptr = 0;
		found = 0;
		strcpy(tmpchar20,"");
		strcpy(tmpchar80,"");
		sscanf(filelines[fl],"%[^:]:%n",tmpchar20,&ptr); // need to insert colon (:)
	
		// copy detected value to the *i^th value of the appropriate array
		// cid indicates whether the fl^th value is a char*, int or double
		// cid_index indicates which member of the appropriate array holds the variable	
		for (int ci=0; ci<charnum; ci++) {
			if ( !strcmp(tmpchar20,charLABELS[ci]) ) {
				sscanf(filelines[fl]+ptr,"%s",tmpchar80);
				strcpy(charvals[ci],tmpchar80);
				cid[fl] = 0;
				cid_index[fl] = ci;
				c++;
				found = 1;
				break;
			} 
		}
		if ( found == 1 ) continue;

		for (int ii=0; ii<intnum; ii++) {
			if ( !strcmp(tmpchar20,intLABELS[ii]) ) {
				sscanf(filelines[fl]+ptr,"%d",&tmpint);
				intvals[ii] = tmpint;
				cid[fl] = 1;
				cid_index[fl] = ii;
				i++;
				found = 1;
				break;
			} 
		}
		if ( found == 1 ) continue;

		for (int di=0; di<doublenum; di++) {
			if ( !strcmp(tmpchar20,doubleLABELS[di]) ) {
				sscanf(filelines[fl]+ptr,"%le",&tmpdouble);
				doublevals[di] = tmpdouble;
				cid[fl] = 2;
				cid_index[fl] = di;
				d++;
				found = 1;
				break;
			} 
		}
		if ( found == 1 ) continue;

		cerr << "could not recognize input entry " << tmpchar20 << endl;
		exit(0);
	}
	bool counterr = 0;	
	if (c != charnum ) {
		cerr << "Did not find expected number of input strings. Need:" << endl;
		for (int ci=0; ci<charnum; ci++) cerr << charLABELS[ci] << endl;
		counterr=1;
	}
	if (i != intnum ) {
		cerr << "did not find expected number of input integers. Need:" << endl;
		for (int ii=0; ii<intnum; ii++) cerr << intLABELS[ii] << endl;
		counterr=1;
	}
	if (d != doublenum ) {
		cerr << "did not find expected number of input doubles. Need:" << endl;
		for (int di=0; di<doublenum; di++) cerr << doubleLABELS[di] << endl;
		counterr=1;
	}
	if (counterr == 1) exit(0);	

	return(1);
}

int Inputs::report_inputs ( ofstream *lfile ) {
	char str[80];
	for (int fl=0; fl<numlines; fl++) {
		*lfile << filelines[fl] << endl; 
		if ( cid[fl] == 0 ) {
			sprintf(str," #string = %s\n",charvals[ cid_index[fl] ]);
			*lfile << str;
		}
		else if ( cid[fl] == 1 ) {
			sprintf(str," #integer = %d\n",intvals[ cid_index[fl] ]);
			*lfile << str;
		}
		else if ( cid[fl] == 2 ) {
			sprintf(str," #double = %le\n",doublevals[ cid_index[fl] ]);
			*lfile << str;
		}
	}
	return(0);
}
