#include "sbfit.h"
#include "cputhreadtime.h"
#include "utils.h"
#include "vector_stats.h"
using namespace std;

SBFIT::SBFIT( ) {
	outpath = getenv("OUTPATH");
	outbase = getenv("OUTBASE");
	safety = getenv("OWSAFETY");
	// default
//	etat=0;

//	cpuid = 0;

	char dir[218];
	strcpy(dir,outpath);
    strcat(dir,"/" );
	strcat(dir,outbase);
	struct stat st;

	// if nosafe active, don't check
	if ( strcmp(safety,"nosafe") ) {
		if ( stat(dir,&st) != 0 ) {	
    		char command[256];
    		strcpy(command,"mkdir ");
    		strcat(command,dir);
    		system(command);
    		
			strcpy(command,"touch ");
    		strcat(command,dir);
    		strcat(command,"/");
    		strcat(command,outbase);
    		strcat(command,".log");
    		system(command);
		} else {
			cerr << "* * * * * * * * * * *  * * * * * * *" << endl;
			cerr << "* " << outpath << " already exists." << endl;
			cerr << "* Potentially over-write files (y|N)? ";
			char yN[2];
			cin >> yN;
			if ( strcmp(yN,"y") ) exit(0);
		}
	} // even if nosafe is active, create if it doesn't exist
	else if ( stat(dir,&st) != 0 ) {
		char command[256];
		strcpy(command,"mkdir ");
		strcat(command,dir);
		system(command);
	
//		sprintf(command, "touch %s/%s.t%i.log.%i",dir,outbase,cpuid,etat);
//		system(command);
	}

	// in case 0 or only 1 isotropic(s) are/is used
	fwhm_iso = 1;
	fwhm_iso2 = 1;
	centre_iso = 0;
	centre_iso2 = 0;
	iso_A = 0;
	iso_A2 = 0;

	fit_fwhm_iso = 0;
	fit_fwhm_iso2 = 0;
	fit_centre_iso = 0;
	fit_centre_iso2 = 0;
	fit_iso_A = 0;
	fit_iso_A2 = 0;

	nvar = 8;
}

int SBFIT::append_logfile() {
	const char * ext = ".log";
	open_append_file(ext, &lfile, logfilename);
	return(0);
}

SBFIT::SBFIT( const SBFIT& srcSBFIT ) {
	// copy everything by hand. 
	//constant = srcSBFIT.constant;

	//J = gsl_matrix_calloc(numdata,datanum);
	//matrix_element = gsl_matrix_calloc(rows,columns);
	//vector_element = gsl_vector_alloc(elements);	

	//gsl_matrix_memcpy( matrix_element, srcSBFIT.matrix_element );
	//gsl_matrix_memcpy( vector_element, srcSBFIT.vector_element );	
}

int SBFIT::copy_min( const SBFIT &srcSBFIT ) {
	// Here it's important to deep-copy only critical structures.
	// everything is already allocated
	// It's OK that everything else refers to the exact same thing

//	distnum = srcSBFIT.distnum;
//	datanum = srcSBFIT.datanum;
	//constant = srcSBFIT.constant;

	//gsl_vector_memcpy( vector_element, srcSBFIT.vector_element );	
	//gsl_matrix_memcpy( matrix_element, srcSBFIT.matrix_element );	
	
	return (0);
}
	
SBFIT::~SBFIT () {
	if ( lfile.is_open() ) {
		lfile << flush;
		lfile.close();
	}

	//matrix_element = gsl_matrix_alloc(1,1);
	//gsl_matrix_free( matrix_element);
	//vector_element = gsl_vector_alloc(1,1);
   // gsl_matrix_free( vector_element );
	
}

int SBFIT::count_xydata( char * datafile ) {
	lfile << "count_xydata()..." << endl;
	char str[80];

	ifstream data;
	data.open( datafile, ios::in );
	if ( !data.is_open() )	{
		cout << "Can't open datafile " << datafile << endl;
		exit(0);
	}
	
	numdata = -1;
	string buf;
	double d1,d2;
	do {
		getline(data,buf);
		if ( iscomment( buf ) ) continue;	
		numdata++;
		if (numdata == 0) {
			sscanf(buf.c_str(), "%le %le", &d1,&d2);
		}
	} while ( !data.eof() );

	sprintf(str,"INFERRED: object.numdata = %i\n",numdata);
	lfile << str << flush;
	data.close();
	lfile << "  numdata = " << numdata << endl << flush;
	return(0);
}

int SBFIT::read_spectrum() {
	lfile << "read_spectrum()..." << endl;
	read_xydata( spectrumfile );
	gsl_vector_memcpy(freq,xdata);
	gsl_vector_memcpy(&spectrum.vector,ydata);

	return(0);
}

int SBFIT::read_integral() {
	lfile << "read_integral()..." << endl;
	read_xydata( integralfile );
//	gsl_vector_memcpy(freq,xdata);
	gsl_vector_memcpy(&integral.vector,ydata);

	return(0);
}

int SBFIT::read_derivative() {
	lfile << "read_derivative()..." << endl;
	read_xydata( derivativefile );
//	gsl_vector_memcpy(freq,xdata);
	gsl_vector_memcpy(&derivative.vector,ydata);

	return(0);
}

int SBFIT::read_xydata( char * datafile ) {
	xdata = gsl_vector_calloc( numdata );
	ydata = gsl_vector_calloc( numdata );

	ifstream data;
	data.open( datafile, ios::in );
	if ( !data.is_open() )	{
		cout << "Can't open datafile " << datafile << endl;
		exit(0);
	}

	string buf;	
	double x,y;
	for (int r=0; r<numdata; r++) {
		getline(data,buf);
		if ( iscomment( buf ) ) {
			r--;
			continue;
		}
		sscanf(buf.c_str(), "%le %le", &x,&y);
		gvs(xdata,r,x);
		gvs(ydata,r,y);
	}
	data.close();
	return(0);
}

int SBFIT::init_data() {
	lfile << "init_data()..." << endl;
	lfile << "  numdata = " << numdata << endl;

	// experimental data
	freq = gsl_vector_alloc(numdata);
	spintder = gsl_vector_calloc(3*numdata);
	spectrum = gsl_vector_subvector(spintder,0,numdata);
	integral = gsl_vector_subvector(spintder,numdata,numdata);
	derivative = gsl_vector_subvector(spintder,2*numdata,numdata);

	spintder_fit = gsl_vector_calloc(3*numdata);
	spectrum_fit= gsl_vector_subvector(spintder_fit,0,numdata);
	integral_fit= gsl_vector_subvector(spintder_fit,numdata,numdata);
	derivative_fit= gsl_vector_subvector(spintder_fit,2*numdata,numdata);

	// data structures for fit information
	numsb = maxsb - minsb + 1;
	lfile << "  numsb = " << numsb << endl;

	lorentz = gsl_matrix_calloc(3*numdata,numsb);
	lorentz_spe_fit = gsl_matrix_submatrix(lorentz,0,0,numdata,numsb);
	lorentz_int_fit 
		= gsl_matrix_submatrix(lorentz,numdata,0,numdata,numsb);
	lorentz_der_fit 
		= gsl_matrix_submatrix(lorentz,2*numdata,0,numdata,numsb);

	// isotropic lorentzian
	lorentz_iso = gsl_vector_calloc(3*numdata);
	lorentz_iso_spe_fit = gsl_vector_subvector(lorentz_iso,0,numdata);
	lorentz_iso_int_fit 
		= gsl_vector_subvector(lorentz_iso,numdata,numdata);
	lorentz_iso_der_fit 
		= gsl_vector_subvector(lorentz_iso,2*numdata,numdata);
	
	L_iso_dfreq = gsl_vector_calloc(numdata);
	QintQnu0_iso_elem = gsl_vector_calloc(numdata);
	Uint_iso = gsl_vector_calloc(numdata);
	Vint_iso = gsl_vector_calloc(numdata);
	Uder_iso = gsl_vector_calloc(numdata);
	Vder_iso = gsl_vector_calloc(numdata);
	Vder2_iso = gsl_vector_calloc(numdata);

	// isotropic2 lorentzian
	lorentz_iso2 = gsl_vector_calloc(3*numdata);
	lorentz_iso2_spe_fit = gsl_vector_subvector(lorentz_iso2,0,numdata);
	lorentz_iso2_int_fit 
		= gsl_vector_subvector(lorentz_iso2,numdata,numdata);
	lorentz_iso2_der_fit 
		= gsl_vector_subvector(lorentz_iso2,2*numdata,numdata);

	L_iso2_dfreq = gsl_vector_calloc(numdata);
	QintQnu0_iso2_elem = gsl_vector_calloc(numdata);
	Uint_iso2 = gsl_vector_calloc(numdata);
	Vint_iso2 = gsl_vector_calloc(numdata);
	Uder_iso2 = gsl_vector_calloc(numdata);
	Vder_iso2 = gsl_vector_calloc(numdata);
	Vder2_iso2 = gsl_vector_calloc(numdata);

	lorentz_An = gsl_vector_calloc(numsb);
	lorentz_An_err = gsl_vector_calloc(numsb);

	residual = gsl_vector_calloc(3.0*numdata);
	residual_spe = gsl_vector_subvector(residual,0,numdata);
	residual_int = gsl_vector_subvector(residual,numdata,numdata);
	residual_der = gsl_vector_subvector(residual,2*numdata,numdata);

	// jacobian is partial of functions at each data point with respect to:
	// 0. center of isotropic peak
	// 1. fwhm of isotropic peak
	// 2. Amplitude of isotropic peak
	// 3. center frequency of MAS manifold
	// 4. fwhm of MAS peaks
	// 5 .. numsb+3 MAS An (except for dependent -1 An)
	// therefore dimension is numdata x numsb+4
	J = gsl_matrix_calloc(3*numdata,numsb+nvar);
	J_spe = gsl_matrix_submatrix(J,0,0,numdata,numsb+nvar-1);
	J_int = gsl_matrix_submatrix(J,numdata,0,numdata,numsb+nvar-1);
	J_der = gsl_matrix_submatrix(J,2*numdata,0,numdata,numsb+nvar-1);

	Ldfreq = gsl_matrix_calloc(numdata,numsb);
	QintQnu0_elem = gsl_matrix_calloc(numdata,numsb);
	Uint = gsl_matrix_calloc(numdata,numsb);
	Vint = gsl_matrix_calloc(numdata,numsb);
	Uder = gsl_matrix_calloc(numdata,numsb);
	Vder = gsl_matrix_calloc(numdata,numsb);
	Vder2 = gsl_matrix_calloc(numdata,numsb);
	
	return(0);
}

int SBFIT::noise_stats () {
	lfile << "noise_stats() ..." << endl;
	ifstream xydata;
	string buf;
	int Nn = 0;
	double x;
	for (int i=0; i<numdata; i++) {
		x = gvg(freq,i);
		if ( abs(x) >= nminfreq && abs(x) <= nmaxfreq ) {
			Nn++;
		}
	}

	gsl_vector * noise_spe;
	gsl_vector * noise_int;
	gsl_vector * noise_der;

	noise_spe = gsl_vector_calloc(Nn);
	noise_int = gsl_vector_calloc(Nn);
	noise_der = gsl_vector_calloc(Nn);

	Nn = 0;	
	for (int i=0; i<numdata; i++){
		x = gvg(freq,i);
		if ( abs(x) >= nminfreq && abs(x) <= nmaxfreq ) {
			gvs(noise_spe,Nn,gvg(&spectrum.vector,Nn));
			gvs(noise_int,Nn,gvg(&integral.vector,Nn));
			gvs(noise_der,Nn,gvg(&derivative.vector,Nn));
			Nn++;
		}
	}
	vector_stats V(noise_spe,0);
	navg_spe = V.mean;
	nstdev_spe = V.stdev;

	vector_stats W(noise_int,0);
	navg_int = W.mean;
	nstdev_int = W.stdev;
	
	vector_stats X(noise_der,0);
	navg_der = X.mean;
	nstdev_der = X.stdev;

	// open file and write noise autocorrelations from V.K, W.K, X.K
	ofstream noise_autocorr;
	const char * ext = ".nac";
	open_write_file(ext, &noise_autocorr);

	char str[81];
	lfile << "Noise:" << endl; 
	sprintf(str,"#lag spectrm integrl dervtiv\n");
	noise_autocorr << str;
	cout << str;
	lfile << str;
	sprintf(str,"#rms %7.4f %7.4f %7.4f\n", nstdev_spe, nstdev_int, nstdev_der);
	noise_autocorr << str;
	cout << str;
	lfile << str;
	sprintf(str,"#avg %7.4f %7.4f %7.4f\n", navg_spe, navg_int, navg_der);
	noise_autocorr << str;
	cout << str;
	lfile << str;
	sprintf(str,"SSPE=%-.4f\n",nstdev_spe); cout << str;
	sprintf(str,"SINT=%-.4f\n",nstdev_int); cout << str;
	sprintf(str,"SDER=%-.4f\n",nstdev_der); cout << str;

	for (unsigned int n=0; n < Nn; n++) {
		sprintf(str,"%4d %7.4f %7.4f %7.4f\n",
			n,gvg(V.K,n),gvg(W.K,n),gvg(X.K,n) );
		noise_autocorr << str;
	}

	gsl_vector_free(noise_spe);
	gsl_vector_free(noise_int);
	gsl_vector_free(noise_der);
	
//	char str[81];
//	sprintf(str,"  average: %12g\n", averagenoise); lfile << str;
//	sprintf(str,"  stdev:   %12g\n", stdevnoise); 	lfile << str;
//	sprintf(str,"  #points: %12d\n",	Nn); 				lfile << str;
	return(0);
}


int SBFIT::write_datafit() {
	lfile << "write_datafit() ..." << endl;

	char str[128];
	ofstream outfile;
	const char * ext = ".lfit";
	char app[260];
	strcpy(app,appendoutfile);
	strcat(app,ext);
	open_write_file(app, &outfile);

	outfile << "#NUMDATA " << numdata << endl;
	outfile << "#MINSB " << minsb << endl;
	outfile << "#MAXSB " << maxsb << endl;
	outfile << "#MAS " << mas << endl;
	outfile << "#AMPS \"";
	for (int s=0; s<numsb; s++) {
		sprintf(str,"%7.5f ",gvg(lorentz_An,s) );
		outfile << str;
	}
	outfile << "\"" << endl;
	outfile << "#FWHM " << fwhm << endl;
	outfile << "#CENTRE " << centre << endl;
	outfile << "#AMPISO " << iso_A << endl;
	outfile << "#FWHMISO " << fwhm_iso << endl;
	outfile << "#CENTREISO " << centre_iso << endl;
	outfile << "#AMPISO2 " << iso_A2 << endl;
	outfile << "#FWHMISO2 " << fwhm_iso2 << endl;
	outfile << "#CENTREISO2 " << centre_iso2 << endl;

	for (int i=0; i<numdata; i++) {
		sprintf(str,"%12.5le %12.5le %12.5le %12.5le %12.5le %12.5le %12.5le\n",
			gsl_vector_get( freq,i),
			gsl_vector_get( &spectrum.vector,i),
			gsl_vector_get( &spectrum_fit.vector,i),
			gsl_vector_get( &integral.vector,i),
			gsl_vector_get( &integral_fit.vector,i),
			gsl_vector_get( &derivative.vector,i),
			gsl_vector_get( &derivative_fit.vector,i) );
		outfile << str;
	}
	outfile.close();
	return(0);
}

int SBFIT::read_datafit() {
	lfile << "read_datafit() ..." << endl;

	string buf;
	ifstream infile;
	infile.open(datafitfile, ios::in);
	if ( !infile.is_open() ) {
		cerr << "Couldn't open datafit file for read" 
			<< datafitfile << " for input" << endl;
		exit(0);
	}
	
	//NUMDATA
	getline(infile,buf);
	parseinputline(buf);
	//MINSB
	getline(infile,buf);
	parseinputline(buf);
	//MAXSB
	getline(infile,buf);
	parseinputline(buf);
	//MAS
	getline(infile,buf);
	parseinputline(buf);
	//AMPS
	getline(infile,buf);
	parseinputline(buf);
	//FWHM
	getline(infile,buf);
	parseinputline(buf);
	//CENTRE
	getline(infile,buf);
	parseinputline(buf);
	//AMPISO
	getline(infile,buf);
	parseinputline(buf);
	//FWHMISO
	getline(infile,buf);
	parseinputline(buf);
	//CENTREISO
	getline(infile,buf);
	parseinputline(buf);
	//AMPISO2
	getline(infile,buf);
	parseinputline(buf);
	//FWHMISO2
	getline(infile,buf);
	parseinputline(buf);
	//CENTREISO2
	getline(infile,buf);
	parseinputline(buf);

	// allocate data structures
	init_data();
	for (int s=0; s<numsb; s++) {
		gvs(lorentz_An,s,rawamps[s]);
	}

	double f, s, fs, i, fi, d, fd;
	for (int ii=0; ii<numdata; ii++){
		getline(infile,buf);
		sscanf(buf.c_str(), "%le %le %le %le %le %le %le", 
			&f,&s,&fs,&i,&fi,&d,&fd);
		gvs( freq,ii,f);
		gvs( &spectrum.vector,ii,s);
		gvs( &spectrum_fit.vector,ii,fs);
		gvs( &integral.vector,ii,i);
		gvs( &integral_fit.vector,ii,fi);
		gvs( &derivative.vector,ii,d);
		gvs( &derivative_fit.vector,ii,fd);
	}
	infile.close();
	return(0);
}

int SBFIT::write_lrntz() {
	lfile << "write_lrntz() ..." << endl;

	char str[81];
	ofstream outfile;
	const char * ext = ".lrntz";
	char app[260];
	strcpy(app,appendoutfile);
	strcat(app,ext);
	open_write_file(app, &outfile);

	outfile << "#FREQ        ISO          ";
	for (int s=0; s<numsb; s++) {
		sprintf(str,"     %4i    ",s);
		outfile << str;
	}
	outfile << endl;

	for (int i=0; i<numdata; i++) {
		sprintf(str,"%12.5le %12.5le %12.5le ",
			gvg( freq,i),
			gvg( &lorentz_iso_spe_fit.vector,i)*iso_A*0.3183098862,
			gvg( &lorentz_iso2_spe_fit.vector,i)*iso_A2*0.3183098862
		);
		outfile << str;
		for (int s=0; s<numsb; s++) {
			sprintf(str,"%12.5le ",
				gmg(&lorentz_spe_fit.matrix,i,s)*gvg(lorentz_An,s)*0.3183098862);
			outfile << str;
		}
		outfile << endl;
	}
	outfile.close();
	return(0);
}

int SBFIT::write_memasdata() {
	lfile << "write_memasdata() ..." << endl;

	char str[81];
	ofstream outfile;
	const char * ext = ".sbdat";
	char app[260];
	strcpy(app,appendoutfile);
	strcat(app,ext);
	open_write_file(app, &outfile);

	sprintf(str,"#%10.1f %3i %10.3g %10.5g\n",mas,0,iso_A*0.31830988,
		iso_A_err*0.31830988 );
	sprintf(str,"#%10.1f %3i %10.3g %10.5g\n",mas,0,iso_A2*0.31830988,
		iso_A2_err*0.31830988 );
	outfile << str;
	for (int s=0; s<numsb; s++) {
		sprintf(str,"%10.1f %3i %10.5g %10.5g\n",
			mas,maxsb-s,gvg(lorentz_An,s)*0.31830988,
			gvg(lorentz_An_err,s)*0.31830988 );
		outfile << str;
		if (maxsb - s == 0 ) {
			double perr = pow(gvg(lorentz_An_err,s),2.0);
			perr += pow(iso_A_err,2.0);
			perr += pow(iso_A2_err,2.0);
			perr = pow(perr,0.5);
			sprintf(str,"#%10.1f %3i %10.5g %10.5g\n",
				mas,0,(iso_A + iso_A2 + gvg(lorentz_An,s))*0.31830988, perr*0.31830988);
			outfile << str;
//QQQQ
		}
	}

	outfile.close();
	return(0);
}

int SBFIT::write_gnuplotscript() {
	lfile << "write_gnuplotscript() ..." << endl;

//	char str[81];
	ofstream outfile;
	const char * ext = ".gp";
	char app[260];
	strcpy(app,appendoutfile);
	strcat(app,ext);
	open_write_file(app, &outfile);

	char bname[256];
	strcpy(bname,outpath);
	strcat(bname,"/");
	strcat(bname,outbase);
	strcat(bname,"/");
	strcat(bname,outbase);
	strcat(bname,appendoutfile);

	outfile << "set term postscript eps enhanced color size 19cm,27cm" 
		<< endl;
	outfile << "set out '" << bname << ".eps'" << endl;
	outfile << "set style line 1 lw 3 lt 1 lc 1" << endl;
	outfile << "set style line 2 lw 1 lt 1 lc 3" << endl;
	outfile << "set style line 3 lw 1 lt 1 lc 2" << endl;
	outfile << "set label 1 'centre\\_iso: " << centre_iso 
		<< "' at screen 0.85,0.67" << endl;
	outfile << "set label 2 'fwhm\\_iso: " << fwhm_iso 
		<< "' at screen 0.85,0.66" << endl;
	outfile << "set label 3 'amp\\_iso: " << iso_A 
		<< "' at screen 0.85,0.65" << endl;
	outfile << "set label 4 'centre\\_iso2: " << centre_iso2 
		<< "' at screen 0.85,0.635" << endl;
	outfile << "set label 5 'fwhm\\_iso2: " << fwhm_iso2 
		<< "' at screen 0.85,0.625" << endl;
	outfile << "set label 6 'amp\\_iso2: " << iso_A2 
		<< "' at screen 0.85,0.615" << endl;
	outfile << "set label 7 'centre: " << centre 
		<< "' at screen 0.85,0.60" << endl;
	outfile << "set label 8 'fwhm: " << fwhm 
		<< "' at screen 0.85,0.59" << endl;
	for (int s=0; s<numsb; s++) {
		outfile << "set label " << s+9 << " 'sb " << maxsb-s << ": "
			<< gvg(lorentz_An,s)*0.3183098862 << "' at screen 0.85," 
			<< 0.58 - double(s)*0.01 << endl;
	}

	outfile << "set xrange [" << gvg(freq,0) << ":" << gvg(freq,numdata-1) 
		<< "]" << endl;
	outfile << "set mxtics 5" << endl;
	outfile << "set xtics out" << endl;
//	outfile << "set mxtics out" << endl;
	outfile << "set tmargin 3" << endl;
	outfile << "set multiplot" << endl;// layout 3,1 scale 0.7,1" << endl;
//	outfile << "set style data lines" << endl;
	outfile << "set size 0.7,0.33" << endl;

	outfile << "set title '" << outbase << "' noenhanced" << endl;

	outfile << "set origin 0.1,0.67" << endl;
	outfile << "plot '" << bname 
		<< ".lfit' u 1:2 ls 1 w l not, '' u 1:3 ls 2 w l not,\\" 
		<< endl;
	outfile << "  0 lt 0 not,\\" << endl;
	outfile << "  '" << bname << ".lrntz' u 1:2 lt 1 w filledcurves not,\\" << endl;
	outfile << "  '' u 1:3 lt 1 w filledcurves not,\\" << endl;
	outfile << "  '' u 1:" << maxsb+4 << " w l ls 3 not,\\" << endl;
	outfile << "  '' u 1:" << maxsb+5 << " w l ls 3 not,\\" << endl;
	outfile << "  '' u 1:" << maxsb+3 << " w l ls 3 not#,\\" << endl;
	for (int s=4; s<numsb+3; s++) {
		if ( abs(maxsb+3 - s) < 0.5) continue;
		if ( abs(maxsb+4 - s) < 0.5) continue;
		if ( abs(maxsb+5 - s) < 0.5) continue;
		outfile << "#  '' u 1:" << s << " w l ls 3 not\\" << endl;
	}
	outfile << "#  '' u 1:" << numsb+4 << " w l ls 3 not" << endl;

	outfile << "unset title" << endl;
	outfile << "set origin 0.1,0.33" << endl;
	outfile << "plot '" << bname 
		<< ".lfit' u 1:4 w l ls 1 not, '' u 1:5 w l ls 2 not,\\" << endl;
	outfile << "  0 lt 0 not" << endl;

	outfile << "set origin 0.1,0.0" << endl;
	outfile << "plot '" << bname 
		<< ".lfit' u 1:6 w l ls 1 not, '' u 1:7 w l ls 2 not,\\" << endl;
	outfile << "  0 lt 0 not" << endl;



	// inset
	outfile << "set origin 0.75,0.67" << endl;
	outfile << "set size 0.25,0.33" << endl;
	outfile << "set xtics 500" << endl;
	outfile << "set xrange [" << centre+0.6*mas << ":" << centre - 0.6*mas 
		<< "]" << endl;
	outfile << "plot '" << bname 
		<< ".lfit' u 1:2 ls 1 w l not, '' u 1:3 ls 2 w l not,\\" 
		<< endl;
	outfile << "  0 lt 0 not,\\" << endl;
	outfile << "  '" << bname << ".lrntz' u 1:2 lt 1 w filledcurves not,\\" << endl;
	outfile << "  '' u 1:3 lt 1 w filledcurves not,\\" << endl;
	outfile << "  '' u 1:" << maxsb+4 << " w l ls 3 not" << endl;
	// end inset
	
	// inset deriv
	outfile << "set origin 0.75,0.0" << endl;
	outfile << "plot '" << bname 
		<< ".lfit' u 1:6 w l ls 1 not, '' u 1:7 w l ls 2 not,\\" << endl;
	outfile << "  0 lt 0 not" << endl;
	// end inset deriv

	// inset deriv2
	outfile << "set origin 0.15,0.225" << endl;
	outfile << "set size 0.6,0.1" << endl;
	outfile << "set xtics 2000" << endl;
	outfile << "set xrange [" << centre+(maxsb+1)*mas << ":" 
		<< centre - (-minsb-1)*mas << "]" << endl;
	outfile << "set yrange [-100:100]" << endl;
	outfile << "plot '" << bname 
		<< ".lfit' u 1:6 w l ls 1 not, '' u 1:7 w l ls 2 not,\\" << endl;
	outfile << "  0 lt 0 not" << endl;
	// end inset deriv2

	outfile << "unset multiplot" << endl;
	outfile.close();
	return(0);
}

int SBFIT::init_fit_params() {
	lfile << "init_fit_params()..." << endl;

	char str[81];
	for (int s=0; s<numsb; s++) {
		double ss = double(maxsb-s)*mas + centre; 
		int i = index_of_closest_value( freq, ss );
		gvs(lorentz_An, s, 1.57*fwhm*gvg( &spectrum.vector, i) );
		sprintf(str,"  sideband %3d freq %10.3f init %8.6f\n",
			maxsb-s,ss,gvg(lorentz_An,s) );
		lfile << str << flush;
		if ( gvg(lorentz_An,s) <= 0 ) {
			cerr << "reconsider using sideband " << maxsb-s 
				<< ", as initial estimate is negative" << endl;
		}
	}

	return(0);
}

int SBFIT::calc_spectrum() {
	lfile << "calc_spectrum() ..." << flush << endl;

	char str[80];
	lfile << "  using:" << endl;

	sprintf(str,"    iso_A       %10.3g\n",iso_A); 
		lfile << str << flush;
	sprintf(str,"    centre_iso  %10.3g\n",centre_iso); 
		lfile << str << flush;
	sprintf(str,"    fwhm_iso    %10.3g\n",fwhm_iso); 
		lfile << str << flush;
	sprintf(str,"    iso_A2       %10.3g\n",iso_A2); 
		lfile << str << flush;
	sprintf(str,"    centre_iso2  %10.3g\n",centre_iso2); 
		lfile << str << flush;
	sprintf(str,"    fwhm_iso2    %10.3g\n",fwhm_iso2); 
		lfile << str << flush;
	sprintf(str,"    centre      %10.3g\n",centre); 
		lfile << str << flush;
	sprintf(str,"    fwhm        %10.3g\n",fwhm); 
		lfile << str << flush;

	for (int sb=0; sb<numsb; sb++) {
		sprintf(str,"        sb %3d      %10.3g\n",maxsb-sb,gvg(lorentz_An,sb) ); 
		lfile << str << flush;
	}
		
	
	// calculate the matrix of individual, unscaled, normalized lorenztians
	double fwhm2 = pow(fwhm,2.0);
	double fwhm_d = 2.0*fwhm;
	for (int sb=0; sb<numsb; sb++) {
		int n = maxsb - sb;
		double u = centre + double(n)*mas;
		for (int nu=0; nu<numdata; nu++ ) {
			double dfreq = u-gvg(freq,nu);
			gms(Ldfreq,nu,sb,dfreq);
			gms(&lorentz_spe_fit.matrix,nu,sb, fwhm_d/(fwhm2+4.0*pow(dfreq,2.0)));
		}
	}
	
	// calculate the spectrum (all at once multiplying by An/pi)
	gsl_blas_dgemv(CblasNoTrans,0.3183098862,&lorentz_spe_fit.matrix,
		lorentz_An, 0.0, &spectrum_fit.vector);

	// isotropic
	double fwhm_iso_d = 2.0 * fwhm_iso;	
	double fwhm_isosq = pow(fwhm_iso,2.0);

	for (int nu=0; nu<numdata; nu++ ) {
		// L_iso_dfreq is the same as the maxsb^th column of Ldfreq, which could
		// be used with vector_view, but leave it for now
		gvs(L_iso_dfreq, nu, centre_iso - gvg(freq,nu) );
		double dfreqsq = pow(gvg(L_iso_dfreq,nu), 2.0);
		gvs(&lorentz_iso_spe_fit.vector, nu, fwhm_iso_d/(fwhm_isosq + 4.0*dfreqsq));
	}

	gsl_blas_daxpy(iso_A*0.3183098862, &lorentz_iso_spe_fit.vector, 
		&spectrum_fit.vector );

	// isotropic2
	fwhm_iso_d = 2.0 * fwhm_iso2;	
	fwhm_isosq = pow(fwhm_iso2,2.0);

	for (int nu=0; nu<numdata; nu++ ) {
		// L_iso_dfreq is the same as the maxsb^th column of Ldfreq, which could
		// be used with vector_view, but leave it for now
		gvs(L_iso2_dfreq, nu, centre_iso2 - gvg(freq,nu) );
		double dfreqsq = pow(gvg(L_iso2_dfreq,nu), 2.0);
		gvs(&lorentz_iso2_spe_fit.vector, nu, fwhm_iso_d/(fwhm_isosq + 4.0*dfreqsq));
	}

	gsl_blas_daxpy(iso_A2*0.3183098862, &lorentz_iso2_spe_fit.vector, 
		&spectrum_fit.vector );

	// calculate residual (is fit-data)
	gsl_vector_memcpy(&residual_spe.vector, &spectrum_fit.vector );
	gsl_blas_daxpy(-1.0, &spectrum.vector, &residual_spe.vector );
	gsl_vector_scale(&residual_spe.vector,1.0/sigmaspe);

	return(0);
}

int SBFIT::calc_integral() {
	lfile << "calc_integral() ..." << flush << endl;
	
	// calculate the matrix of individual, unscaled, normalized lorenztians
	// definite integrals from left edge of spectrum to given frequency nu
	double fwhm2 = pow(fwhm,2.0);
	double fwhm_d = 2.0*fwhm;
	for (int nu=0; nu<numdata; nu++ ) {
		double U = fwhm_d * (gvg(freq,0) - gvg(freq,nu) );
		for (int sb=0; sb<numsb; sb++) {
			int n = maxsb - sb;
			double u = centre + double(n)*mas;
			double V = fwhm2 + 4.0 * (u-gvg(freq,0)) * (u-gvg(freq,nu) );

			gms(QintQnu0_elem,nu,sb, 2.0*(centre+n*mas) - gvg(freq,0)-nu);
			gms(Uint,nu,sb,U);
			gms(Vint,nu,sb,V);
			gms(&lorentz_int_fit.matrix,nu,sb, atan2(U,V) );
		}
	}
	
	// calculate the spectrum (all at once multiplying by An/pi)
	gsl_blas_dgemv(CblasNoTrans,0.3183098862,&lorentz_int_fit.matrix,
		lorentz_An,0.0,&integral_fit.vector);

	// isotropic
	double fwhm_iso_d = 2.0 * fwhm_iso;	
	double fwhm_isosq = pow(fwhm_iso,2.0);
	for (int nu=0; nu<numdata; nu++ ) {
		double U = fwhm_iso_d * (gvg(freq,0) - gvg(freq,nu) );
		double V = fwhm_isosq + 4.0 * (centre_iso-gvg(freq,0)) 
			* (centre_iso-gvg(freq,nu) );
		gvs(QintQnu0_iso_elem,nu, 2.0*centre_iso - gvg(freq,0)-nu);
		gvs(Uint_iso,nu,U);
		gvs(Vint_iso,nu,V);
		gvs(&lorentz_iso_int_fit.vector,nu,atan2(U,V) );
	}
	gsl_blas_daxpy(iso_A/3.14159, &lorentz_iso_int_fit.vector, &integral_fit.vector );

	// isotropic2
	fwhm_iso_d = 2.0 * fwhm_iso2;	
	fwhm_isosq = pow(fwhm_iso2,2.0);
	for (int nu=0; nu<numdata; nu++ ) {
		double U = fwhm_iso_d * (gvg(freq,0) - gvg(freq,nu) );
		double V = fwhm_isosq + 4.0 * (centre_iso2-gvg(freq,0)) 
			* (centre_iso2-gvg(freq,nu) );
		gvs(QintQnu0_iso2_elem,nu, 2.0*centre_iso2 - gvg(freq,0)-nu);
		gvs(Uint_iso2,nu,U);
		gvs(Vint_iso2,nu,V);
		gvs(&lorentz_iso2_int_fit.vector,nu,atan2(U,V) );
	}
	gsl_blas_daxpy(iso_A2/3.14159, &lorentz_iso2_int_fit.vector, &integral_fit.vector );


	// scale and shift
	gsl_vector_scale(&integral_fit.vector,0.0001);
	gsl_vector_scale(&integral_fit.vector,abs(1.0/(gvg(freq,0)-gvg(freq,1))) );
	
	int i = index_of_closest_value( freq, centre );

//	gsl_vector_add_constant(&integral_fit.vector, 
//		gvg(&integral.vector,i) - gvg(&integral_fit.vector,i) );
	
	gsl_vector_add_constant(&integral_fit.vector, -gvg(&integral_fit.vector,i) );
	gsl_vector_add_constant(&integral.vector, -gvg(&integral.vector,i) );
	
	// calculate residual (is fit-data)
	gsl_vector_memcpy(&residual_int.vector, &integral_fit.vector );
	gsl_blas_daxpy(-1.0, &integral.vector, &residual_int.vector );

	gsl_vector_scale(&residual_int.vector,1.0/sigmaint);
	
	return(0);
}

int SBFIT::calc_derivative() {
	lfile << "calc_derivative() ..." << flush << endl;
	
	// calculate the matrix of individual, unscaled, normalized lorenztians
	// derivatives. We presume this will only be run if calc_spectrum was
	// run first, in which case Ldfreq already contains the calculations
	// which we would have to loop here to calculate again.
	gsl_matrix_memcpy(Uder,Ldfreq);
	gsl_matrix_scale(Uder,fwhm*16.0);

	gsl_matrix_memcpy(Vder,Ldfreq);
	gsl_matrix_mul_elements(Vder,Ldfreq);
	gsl_matrix_scale(Vder,4.0);
	gsl_matrix_add_constant( Vder, pow(fwhm,2.0) );

	gsl_matrix_memcpy(Vder2,Vder);
	gsl_matrix_mul_elements(Vder2,Vder);

	gsl_matrix_memcpy(&lorentz_der_fit.matrix,Uder);
	gsl_matrix_div_elements(&lorentz_der_fit.matrix,Vder2);

	// calculate the spectrum (all at once multiplying by An/pi)
	gsl_blas_dgemv(CblasNoTrans,-0.3183098862*abs(gvg(freq,0)-gvg(freq,1)),
		&lorentz_der_fit.matrix, lorentz_An,0.0,&derivative_fit.vector);

	// isotropic
	gsl_vector_memcpy(Uder_iso,L_iso_dfreq);
	gsl_vector_scale(Uder_iso,fwhm_iso*8.0);

	gsl_vector_memcpy(Vder_iso,L_iso_dfreq);
	gsl_vector_mul(Vder_iso,L_iso_dfreq);
	gsl_vector_scale(Vder_iso,4.0);
	gsl_vector_add_constant( Vder_iso, pow(fwhm_iso,2.0) );

	gsl_vector_memcpy(Vder2_iso,Vder_iso);
	gsl_vector_mul(Vder2_iso,Vder_iso);
	
	gsl_vector_memcpy(&lorentz_iso_der_fit.vector,Uder_iso);
	gsl_vector_div(&lorentz_iso_der_fit.vector,Vder2_iso);

	gsl_blas_daxpy(-abs(gvg(freq,0)-gvg(freq,1))*iso_A/3.14159,&lorentz_iso_der_fit.vector,&derivative_fit.vector);

	// isotropic2
	gsl_vector_memcpy(Uder_iso2,L_iso2_dfreq);
	gsl_vector_scale(Uder_iso2,fwhm_iso2*8.0);

	gsl_vector_memcpy(Vder_iso2,L_iso2_dfreq);
	gsl_vector_mul(Vder_iso2,L_iso2_dfreq);
	gsl_vector_scale(Vder_iso2,4.0);
	gsl_vector_add_constant( Vder_iso2, pow(fwhm_iso2,2.0) );

	gsl_vector_memcpy(Vder2_iso2,Vder_iso2);
	gsl_vector_mul(Vder2_iso2,Vder_iso2);
	
	gsl_vector_memcpy(&lorentz_iso2_der_fit.vector,Uder_iso2);
	gsl_vector_div(&lorentz_iso2_der_fit.vector,Vder2_iso2);

	gsl_blas_daxpy(-abs(gvg(freq,0)-gvg(freq,1))*iso_A2/3.14159,&lorentz_iso2_der_fit.vector,&derivative_fit.vector);


	// calculate residual (is fit-data)
	gsl_vector_memcpy(&residual_der.vector, &derivative_fit.vector );
	gsl_blas_daxpy(-1.0, &derivative.vector, &residual_der.vector );

	gsl_vector_scale(&residual_der.vector,1.0/sigmader);
	return(0);
}


// jacobian is partial of functions at each data point with respect to:
// 0. center of isotropic peak
// 1. fwhm of isotropic peak
// 2. Amplitude of isotropic peak
// 3. center of isotropic2 peak
// 4. fwhm of isotropic2 peak
// 5. Amplitude of isotropic2 peak
// 6. center frequency of MAS manifold
// 7. fwhm of MAS peaks
// 8 .. numsb+3 MAS An (except for dependent -1 An)

int SBFIT::calc_spectrum_partials() {
	lfile << "calc_spectrum_partials() ..." << flush << endl;
	gsl_matrix_set_all(&J_spe.matrix,0.0);

	gsl_matrix * work = gsl_matrix_calloc(numdata,numsb);

	char str[80];
	lfile << "  using:" << endl;
	sprintf(str,"    iso_A       %8.2f\n",iso_A); lfile << str << flush;
	sprintf(str,"    centre_iso  %8.2f\n",centre_iso); lfile << str << flush;
	sprintf(str,"    fwhm_iso    %8.2f\n",fwhm_iso); lfile << str << flush;
	sprintf(str,"    iso_A2      %8.2f\n",iso_A2); lfile << str << flush;
	sprintf(str,"    centre_iso2 %8.2f\n",centre_iso2); lfile << str << flush;
	sprintf(str,"    fwhm_iso2   %8.2f\n",fwhm_iso2); lfile << str << flush;
	sprintf(str,"    centre      %8.2f\n",centre); lfile << str << flush;
	sprintf(str,"    fwhm        %8.2f\n",fwhm); lfile << str << flush;

	// partial with respect to centre frequency of isotropic2 peak
	if (fit_centre_iso2 != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_spe.matrix,0);
		gsl_vector_memcpy(&Jcv.vector,&lorentz_iso2_spe_fit.vector);
		gsl_vector_mul(&Jcv.vector,&lorentz_iso2_spe_fit.vector);
		gsl_vector_mul(&Jcv.vector,L_iso2_dfreq);
		gsl_vector_scale(&Jcv.vector,-4.0*iso_A2/(fwhm_iso2*3.141592654) );
	}

	// partial with respect to fwhm of isotropic2 peak
	if (fit_fwhm_iso2 != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_spe.matrix,1);
		gsl_vector_memcpy(&Jcv.vector,&lorentz_iso2_spe_fit.vector);
		gsl_vector_scale(&Jcv.vector,-1.0);
		gsl_vector_add_constant(&Jcv.vector,1.0/fwhm_iso2);
		gsl_vector_mul(&Jcv.vector,&lorentz_iso2_spe_fit.vector);
		gsl_vector_scale(&Jcv.vector,iso_A2 * 0.3183098862);
		gsl_vector_scale(&Jcv.vector,fwhm_iso2);
	}

	// partial with respect to amplitude of isotropic2 peak iso_A2
	if (fit_iso_A2 != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_spe.matrix,2);
		gsl_vector_memcpy(&Jcv.vector, &lorentz_iso2_spe_fit.vector);
		gsl_vector_scale(&Jcv.vector,0.3183098862);
		gsl_vector_scale(&Jcv.vector,iso_A2);
	}

	// partial with respect to centre frequency of isotropic peak
	if (fit_centre_iso != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_spe.matrix,3);
		gsl_vector_memcpy(&Jcv.vector,&lorentz_iso_spe_fit.vector);
		gsl_vector_mul(&Jcv.vector,&lorentz_iso_spe_fit.vector);
		gsl_vector_mul(&Jcv.vector,L_iso_dfreq);
		gsl_vector_scale(&Jcv.vector,-4.0*iso_A/(fwhm_iso*3.141592654) );
	}

	// partial with respect to fwhm of isotropic peak
	if (fit_fwhm_iso != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_spe.matrix,4);
		gsl_vector_memcpy(&Jcv.vector,&lorentz_iso_spe_fit.vector);
		gsl_vector_scale(&Jcv.vector,-1.0);
		gsl_vector_add_constant(&Jcv.vector,1.0/fwhm_iso);
		gsl_vector_mul(&Jcv.vector,&lorentz_iso_spe_fit.vector);
		gsl_vector_scale(&Jcv.vector,iso_A * 0.3183098862);
		gsl_vector_scale(&Jcv.vector,fwhm_iso);
	}

	// partial with respect to amplitude of isotropic peak iso_A
	if (fit_iso_A != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_spe.matrix,5);
		gsl_vector_memcpy(&Jcv.vector, &lorentz_iso_spe_fit.vector);
		gsl_vector_scale(&Jcv.vector,0.3183098862);
		gsl_vector_scale(&Jcv.vector,iso_A);
	}
	// partial with respect to centre frequency of MAS manifold (nu_0)
	if (fit_centre != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_spe.matrix,6);
		gsl_matrix_memcpy(work,&lorentz_spe_fit.matrix);
		gsl_matrix_mul_elements(work,&lorentz_spe_fit.matrix);
		gsl_matrix_mul_elements(work,Ldfreq);
		gsl_blas_dgemv(CblasNoTrans,-4.0/(fwhm*3.141592654),work, lorentz_An,
			0.0,&Jcv.vector);
	}

	// partial with respect to fwhm of MAS manifold peaks (fwhm)
	if (fit_fwhm != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_spe.matrix,7);
		gsl_matrix_memcpy(work,&lorentz_spe_fit.matrix);
		gsl_matrix_scale(work,-1.0);
		gsl_matrix_add_constant(work,1.0/fwhm);
      	gsl_matrix_mul_elements(work,&lorentz_spe_fit.matrix);
		gsl_blas_dgemv(CblasNoTrans,0.3183098862,work,lorentz_An,0.0,&Jcv.vector);
		gsl_vector_scale(&Jcv.vector,fwhm);
	}

	// partials with respect to MAS manifold An
	if (fit_An != 0 ) {
		gsl_vector_view m1sb = gsl_matrix_column(&lorentz_spe_fit.matrix,maxsb+1);
		
		for (int sb=0; sb<numsb; sb++) {
			if (sb == maxsb+1) continue;
			int sbi = sb+nvar;
			if (sb > maxsb+1) sbi = sb+nvar-1;
			gsl_vector_view Jcv = gsl_matrix_column(&J_spe.matrix,sbi);
			gsl_vector_view Ln = gsl_matrix_column(&lorentz_spe_fit.matrix,sb);

			gsl_vector_memcpy(&Jcv.vector,&Ln.vector);
			gsl_blas_daxpy(maxsb-sb, &m1sb.vector, &Jcv.vector);
			gsl_vector_scale(&Jcv.vector,gvg(lorentz_An,sb) );
		}
	}
	gsl_matrix_scale(&J_spe.matrix,1.0/sigmaspe);
	gsl_matrix_free(work);
	return(0);
}

int SBFIT::calc_integral_partials() {
	lfile << "calc_integral_partials() ..." << flush << endl;
	gsl_matrix_set_all(&J_int.matrix,0.0);

	gsl_matrix * work = gsl_matrix_calloc(numdata,numsb);
	gsl_matrix * xork = gsl_matrix_calloc(numdata,numsb);

	gsl_vector * vork = gsl_vector_calloc(numdata);
	
	// partial with respect to centre frequency of isotropic2 peak
	if (fit_centre_iso2 != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_int.matrix,0);
		gsl_vector_memcpy(&Jcv.vector,Uint_iso2);
		gsl_vector_memcpy(vork,Vint_iso2);
		gvf(&Jcv.vector,&vsquare);
		gvf(vork,&vsquare);
		gsl_vector_add(&Jcv.vector,vork);
		gvf(&Jcv.vector,&vinvert);
		gsl_vector_mul(&Jcv.vector,QintQnu0_iso2_elem);
		gsl_vector_mul(&Jcv.vector,Uint_iso2);
		gsl_vector_scale(&Jcv.vector,-1.273239545*iso_A2);
	}

	// partial with respect to fwhm of isotropic2 peak 
	if (fit_fwhm_iso2 != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_int.matrix,1);
		gsl_vector_memcpy(&Jcv.vector,Uint_iso2);
		gsl_vector_memcpy(vork,Vint_iso2);
		gvf(&Jcv.vector,&vsquare);
		gvf(vork,&vsquare);
		gsl_vector_add(&Jcv.vector,vork);
      gvf(&Jcv.vector,&vinvert);
		gsl_vector_memcpy(vork,Vint_iso2);
		gsl_vector_scale(vork,1.0/fwhm_iso2);
		gsl_vector_add_constant(vork,-2.0*fwhm_iso2);
		gsl_vector_mul(&Jcv.vector,vork);
		gsl_vector_mul(&Jcv.vector,Uint_iso2);
		gsl_vector_mul(&Jcv.vector,&lorentz_iso2_spe_fit.vector);
		gsl_vector_scale(&Jcv.vector,iso_A2*0.3183098862);
		gsl_vector_scale(&Jcv.vector,fwhm_iso2);
	}

	// partial with respect to amplitude of isotropic2 peak iso_A
	if (fit_iso_A2 != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_int.matrix,2);
		gsl_vector_memcpy(&Jcv.vector, &lorentz_iso2_int_fit.vector);
		gsl_vector_scale(&Jcv.vector,iso_A2);
	}

	// partial with respect to centre frequency of isotropic peak
	if (fit_centre_iso != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_int.matrix,3);
		gsl_vector_memcpy(&Jcv.vector,Uint_iso);
		gsl_vector_memcpy(vork,Vint_iso);
		gvf(&Jcv.vector,&vsquare);
		gvf(vork,&vsquare);
		gsl_vector_add(&Jcv.vector,vork);
		gvf(&Jcv.vector,&vinvert);
		gsl_vector_mul(&Jcv.vector,QintQnu0_iso_elem);
		gsl_vector_mul(&Jcv.vector,Uint_iso);
		gsl_vector_scale(&Jcv.vector,-1.273239545*iso_A);
	}

	// partial with respect to fwhm of isotropic peak 
	if (fit_fwhm_iso != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_int.matrix,4);
		gsl_vector_memcpy(&Jcv.vector,Uint_iso);
		gsl_vector_memcpy(vork,Vint_iso);
		gvf(&Jcv.vector,&vsquare);
		gvf(vork,&vsquare);
		gsl_vector_add(&Jcv.vector,vork);
      gvf(&Jcv.vector,&vinvert);
		gsl_vector_memcpy(vork,Vint_iso);
		gsl_vector_scale(vork,1.0/fwhm_iso);
		gsl_vector_add_constant(vork,-2.0*fwhm_iso);
		gsl_vector_mul(&Jcv.vector,vork);
		gsl_vector_mul(&Jcv.vector,Uint_iso);
		gsl_vector_mul(&Jcv.vector,&lorentz_iso_spe_fit.vector);
		gsl_vector_scale(&Jcv.vector,iso_A*0.3183098862);
		gsl_vector_scale(&Jcv.vector,fwhm_iso);
	}

	// partial with respect to amplitude of isotropic peak iso_A
	if (fit_iso_A != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_int.matrix,5);
		gsl_vector_memcpy(&Jcv.vector, &lorentz_iso_int_fit.vector);
		gsl_vector_scale(&Jcv.vector,iso_A);
	}

	// partial with respect to centre frequency of MAS manifold (nu_0)
	if (fit_centre != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_int.matrix,6);
		gsl_matrix_memcpy(work,Uint);
		gsl_matrix_memcpy(xork,Vint);
		gmf(work,&vsquare);
		gmf(xork,&vsquare);
		gsl_matrix_add(work,xork);
		gmf(work,&vinvert);
		gsl_matrix_mul_elements(work,QintQnu0_elem);
		gsl_matrix_mul_elements(work,Uint);
		gsl_blas_dgemv(CblasNoTrans,-1.273239545,work, lorentz_An,
			0.0,&Jcv.vector);
	}

	// partial with respect to fwhm of MAS manifold peaks (fwhm)
	if (fit_fwhm != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_int.matrix,7);
		gsl_matrix_memcpy(work,Uint);
		gsl_matrix_memcpy(xork,Vint);
		gmf(work,&vsquare);
		gmf(xork,&vsquare);
		gsl_matrix_add(work,xork);
		gmf(work,&vinvert);
		gsl_matrix_memcpy(xork,Vint);
		gsl_matrix_scale(xork,1.0/fwhm);
		gsl_matrix_add_constant(xork,-2.0*fwhm);
		gsl_matrix_mul_elements(work,xork);
		gsl_matrix_mul_elements(work,Uint);
		gsl_blas_dgemv(CblasNoTrans,0.3183098862,work,lorentz_An,0.0,&Jcv.vector);
		gsl_vector_scale(&Jcv.vector,fwhm);
	}

	// partials with respect to MAS manifold An
	if (fit_An != 0 ) {
		gsl_vector_view m1sb = gsl_matrix_column(&lorentz_int_fit.matrix,maxsb+1);
		for (int sb=0; sb<numsb; sb++) {
			if (sb == maxsb+1) continue;
			int sbi = sb+nvar;
			if (sb > maxsb+1) sbi = sb+nvar-1;
			gsl_vector_view Jcv = gsl_matrix_column(&J_int.matrix,sbi);
			gsl_vector_view Ln = gsl_matrix_column(&lorentz_int_fit.matrix,sb);

			gsl_vector_memcpy(&Jcv.vector,&Ln.vector);
			gsl_blas_daxpy(maxsb-sb, &m1sb.vector, &Jcv.vector);
			gsl_vector_scale(&Jcv.vector, gvg(lorentz_An,sb) );
		}
	}
	gsl_matrix_scale(&J_int.matrix,0.0001/abs(gvg(freq,0)-gvg(freq,1)));
	
	gsl_matrix_scale(&J_int.matrix,1.0/sigmaint);

	gsl_matrix_free(work);
	gsl_matrix_free(xork);
	gsl_vector_free(vork);

	return(0);
}

int SBFIT::calc_derivative_partials() {
	lfile << "calc_derivative_partials() ..." << flush << endl;
	gsl_matrix_set_all(&J_der.matrix,0.0);

	gsl_matrix * work = gsl_matrix_calloc(numdata,numsb);
//	gsl_matrix * xork = gsl_matrix_calloc(numdata,numsb);

//	gsl_vector * vork = gsl_vector_calloc(numdata);
	
	// partial with respect to centre frequency of isotropic peak
	if (fit_centre_iso2 != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_der.matrix,0);
		gsl_vector_memcpy(&Jcv.vector,Uder_iso2);
		gsl_vector_mul(&Jcv.vector,Uder_iso2);
		gsl_vector_div(&Jcv.vector,Vder_iso2);
		gsl_vector_scale(&Jcv.vector,-1.0/fwhm_iso2);
		gsl_vector_add_constant(&Jcv.vector,4.0*fwhm_iso2);
		gsl_vector_div(&Jcv.vector,Vder2_iso2);
//		gsl_vector_div(&Jcv.vector,Vder_iso2);
		gsl_vector_scale(&Jcv.vector,0.6366197724*iso_A2);
	}

	// partial with respect to fwhm of isotropic peak
	if (fit_fwhm_iso2 != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_der.matrix,1);
		gsl_vector_memcpy(&Jcv.vector,Vder_iso2);
		gvf(&Jcv.vector,&vinvert);
		gsl_vector_scale(&Jcv.vector,-4.0*fwhm_iso2);
		gsl_vector_add_constant(&Jcv.vector,1.0/fwhm_iso2);
		gsl_vector_mul(&Jcv.vector,Uder_iso2);
		gsl_vector_div(&Jcv.vector,Vder2_iso2);
//		gsl_vector_div(&Jcv.vector,Vder_iso2);
		gsl_vector_scale(&Jcv.vector,iso_A2*0.3183098862);
		gsl_vector_scale(&Jcv.vector,fwhm_iso2);
	}

	// partial with respect to amplitude of isotropic peak iso_A
	if (fit_iso_A2 != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_der.matrix,2);
		gsl_vector_memcpy(&Jcv.vector, &lorentz_iso2_der_fit.vector);
		gsl_vector_scale(&Jcv.vector,iso_A2);
	}
	
	// partial with respect to centre frequency of isotropic peak
	if (fit_centre_iso != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_der.matrix,3);
		gsl_vector_memcpy(&Jcv.vector,Uder_iso);
		gsl_vector_mul(&Jcv.vector,Uder_iso);
		gsl_vector_div(&Jcv.vector,Vder_iso);
		gsl_vector_scale(&Jcv.vector,-1.0/fwhm_iso);
		gsl_vector_add_constant(&Jcv.vector,4.0*fwhm_iso);
		gsl_vector_div(&Jcv.vector,Vder2_iso);
//		gsl_vector_div(&Jcv.vector,Vder_iso);
		gsl_vector_scale(&Jcv.vector,0.6366197724*iso_A);
	}

	// partial with respect to fwhm of isotropic peak
	if (fit_fwhm_iso != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_der.matrix,4);
		gsl_vector_memcpy(&Jcv.vector,Vder_iso);
		gvf(&Jcv.vector,&vinvert);
		gsl_vector_scale(&Jcv.vector,-4.0*fwhm_iso);
		gsl_vector_add_constant(&Jcv.vector,1.0/fwhm_iso);
		gsl_vector_mul(&Jcv.vector,Uder_iso);
		gsl_vector_div(&Jcv.vector,Vder2_iso);
//		gsl_vector_div(&Jcv.vector,Vder_iso);
		gsl_vector_scale(&Jcv.vector,iso_A*0.3183098862);
		gsl_vector_scale(&Jcv.vector,fwhm_iso);
	}

	// partial with respect to amplitude of isotropic peak iso_A
	if (fit_iso_A != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_der.matrix,5);
		gsl_vector_memcpy(&Jcv.vector, &lorentz_iso_der_fit.vector);
		gsl_vector_scale(&Jcv.vector,iso_A);
	}

	// partial with respect to centre frequency of MAS manifold (nu_0)
	if (fit_centre != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_der.matrix,6);
		gsl_matrix_memcpy(work,Uder);
		gsl_matrix_mul_elements(work,Uder);
		gsl_matrix_div_elements(work,Vder);
		gsl_matrix_scale(work,-1.0/fwhm);
		gsl_matrix_add_constant(work,4.0*fwhm);
		gsl_matrix_div_elements(work,Vder2);
//		gsl_matrix_div_elements(work,Vder);
		gsl_blas_dgemv(CblasNoTrans,0.6366197724,work, lorentz_An,
			0.0,&Jcv.vector);
	}

	// partial with respect to fwhm of MAS manifold peaks (fwhm)
	if (fit_fwhm != 0 ) {
		gsl_vector_view Jcv = gsl_matrix_column(&J_der.matrix,7);
		gsl_matrix_memcpy(work,Vder);
		gmf(work,&vinvert);
		gsl_matrix_scale(work,-4.0*fwhm);
		gsl_matrix_add_constant(work,1.0/fwhm);
		gsl_matrix_mul_elements(work,Uder);
		gsl_matrix_div_elements(work,Vder2);
//		gsl_matrix_div_elements(work,Vder);
		gsl_blas_dgemv(CblasNoTrans,0.3183098862,work,lorentz_An,0.0,&Jcv.vector);
		gsl_vector_scale(&Jcv.vector,fwhm);
	}

	// partials with respect to MAS manifold An
	if (fit_An != 0 ) {
		gsl_vector_view m1sb = gsl_matrix_column(&lorentz_der_fit.matrix,maxsb+1);
		for (int sb=0; sb<numsb; sb++) {
			if (sb == maxsb+1) continue;
			gsl_vector_view Ln = gsl_matrix_column(&lorentz_der_fit.matrix,sb);
			int sbi = sb+nvar;
			if (sb > maxsb+1) sbi = sb+nvar-1;
			gsl_vector_view Jcv = gsl_matrix_column(&J_der.matrix,sbi);

			gsl_vector_memcpy(&Jcv.vector,&Ln.vector);
			gsl_blas_daxpy(maxsb-sb, &m1sb.vector, &Jcv.vector);
			gsl_vector_scale(&Jcv.vector, gvg(lorentz_An,sb) );
		}
	}
	gsl_matrix_scale(&J_der.matrix,1.0/sigmader);
	gsl_matrix_free(work);
	
	return(0);
}
/*
int _lorentzian( double A, double u, double fwhm, double x ) {
	return( A*0.6366197724*fwhm/(pow(fwhm,2.0)+4.0*pow(u-x,2.0)) );
}
int _lorentz_integral( double A, double u, double fwhm, double xl, double x ) {
	return( A*0.6366197724*fwhm/(pow(fwhm,2.0)+4.0*pow(u-x,2.0)) );
}
*/

int SBFIT::nonlinfit() {
	lfile << "nonlinfit()..." << endl << flush;

	// keep number of floaters and size of jacobian to a minimum,
	// but we need to translate back and forth between fit parameters index
	// and the full numsb+4 dimensionality. Do this with binary code
	binarysum = 0;
	int numfloatpars = 0;
	if (fit_centre_iso2 != 0) {
		binarysum += 256;
		numfloatpars++;
	}
	if (fit_fwhm_iso2 != 0) {
		binarysum += 128;
		numfloatpars++;
	}
	if (fit_iso_A2 != 0) {
		binarysum += 64;
		numfloatpars++;
	}
	if (fit_centre_iso != 0) {
		binarysum += 32;
		numfloatpars++;
	}
	if (fit_fwhm_iso != 0) {
		binarysum += 16;
		numfloatpars++;
	}
	if (fit_iso_A != 0) {
		binarysum += 8;
		numfloatpars++;
	}
	if (fit_centre != 0) {
		binarysum += 4;
		numfloatpars++;
	}
	if (fit_fwhm != 0) {
		binarysum += 2;
		numfloatpars++;
	}
	if (fit_An != 0) {
		binarysum += 1;
		numfloatpars += numsb-1;
	}

	// interpret fitconfig
	int multibranch = 0;
	int fc_tmp=fitconfig;
	if ( fc_tmp > 3 ) {
		multibranch++;
		fc_tmp -= 4;
	} 
	if ( fc_tmp > 1 ) {
		multibranch++;
		fc_tmp -= 2;
	}
	if (fc_tmp > 0 ) {
		multibranch++;
		fc_tmp -= 1;
	}
	if (fc_tmp != 0 ) {
		cerr << "fitconfig not interpreted properly" << endl;
		exit(0);
	}

	lfile << "  binarysum = " << binarysum << endl;
	lfile << "  multibranch = " << multibranch << endl;	
	lfile << "  numfloatpars = " << numfloatpars << endl;	

	// gaussian centers, gshift, lw, and overall amplitude control
	gsl_vector * floaters = gsl_vector_calloc( numfloatpars );
	gsl_vector * nlfit_residual = gsl_vector_calloc(multibranch*numdata);
	gsl_matrix * nlfit_jacobian = 
		gsl_matrix_calloc(numfloatpars,multibranch*numdata);

	// set initial values
	int next = 0;
	int binsum = binarysum;
	if ( binsum > 255 ) {
		gvs(floaters,next,centre_iso2);
		binsum -= 256;
		next++;
	}
	if ( binsum > 127 ) {
		gvs(floaters,next,log( fwhm_iso2 ) );
		binsum -= 128;
		next++;
	}
	if ( binsum > 63 ) {
		gvs(floaters,next,log( iso_A2 ) );
		binsum -= 64;
		next++;
	}
	if ( binsum > 31 ) {
		gvs(floaters,next,centre_iso);
		binsum -= 32;
		next++;
	}
	if ( binsum > 15 ) {
		gvs(floaters,next,log( fwhm_iso ) );
		binsum -= 16;				
		next++;
	}
	if ( binsum > 7 ) {
		gvs(floaters,next,log( iso_A ) );
		binsum -= 8;
		next++;
	}
	if ( binsum > 3 ) {
		gvs(floaters,next, centre );
		binsum -= 4;
		next++;
	}
	if ( binsum > 1 ) {
		gvs(floaters,next,log( fwhm ) );
		binsum -= 2;
		next++;
	}
	if ( binsum > 0 ) {
		binsum -= 1;
		for (int s=0; s<numsb; s++) {
			if (s==maxsb+1) continue;
			if (gvg( lorentz_An, s ) <= 0) {
				gvs(floaters,next,0.001);
			} else {
				gvs(floaters,next,log( gvg( lorentz_An, s ) ) );
			}
			next++;
		}
	}
	if (binsum != 0 || next != numfloatpars) {
		cerr << "error in initial assignment of float parameters" << endl;
		exit(0);
	}

	// Declare pointer of type ...
	const gsl_multifit_fdfsolver_type *Ts;
	// declare pointer to a new solver
	gsl_multifit_fdfsolver *s;
	int status;

	// covariance matrix 
	gsl_matrix *covar = gsl_matrix_calloc (numfloatpars,numfloatpars);
	gsl_multifit_function_fdf fitf;
		fitf.f = _nlfit_f;
		fitf.df = _nlfit_df;
		fitf.fdf = _nlfit_fdf;
		fitf.n = multibranch*numdata;
		fitf.p = numfloatpars;
		fitf.params = this;	// i.e. a pointer to this object

	// Solver details
	Ts = gsl_multifit_fdfsolver_lmsder;
	s = gsl_multifit_fdfsolver_alloc (Ts, multibranch*numdata, numfloatpars);

	gsl_multifit_fdfsolver_set (s, &fitf, floaters);

	// A LA /Users/jgehman/BoltzmannStatistics/BS-MAS/Src/BoltmannME.cc
	// we need an object function to both calculate residual and Jacobian
	// and just copy them back and forth as necessary. residual *could* be
	// alternatively repaked residual?
	int iter = 0;
	print_state_fit (iter, s);
	lfile << flush;
	char str[81];
	do {
		iter++;
		status = gsl_multifit_fdfsolver_iterate (s);
		sprintf (str,"status numeric = %d\t", status);
		lfile << str; 
		sprintf (str,"status = %s\n", gsl_strerror (status));
		lfile << str;
		print_state_fit (iter, s);
		lfile << flush;
		if (status)
			break;
		status = gsl_multifit_test_delta (s->dx, s->x, 0.00001, 0.00001);
	} while (status == GSL_CONTINUE && iter < maxit);
	gsl_multifit_covar (s->J, 0.0, covar);

	lfile << "Final Gaussian Fit parameters:\n";
	#define FIT(i) gsl_vector_get(s->x,i)
	#define ERR(i) sqrt(gsl_matrix_get(covar,i,i))
	#define COV(i,j) gsl_matrix_get(covar,i,j)/(ERR(i)*ERR(j))

	// write results to logfile, assign parameter errors, etc
	next = 0;
	binsum = binarysum;

	if ( binsum > 255 ) {
//		centre_iso = FIT(next);
		centre_iso2_err = ERR(next);
		sprintf(str,"  fitpar %2d",next);
		lfile << str;
		sprintf(str,"  centre_iso2: %10.5g +/- %10.5g\n",
			centre_iso2,centre_iso2_err);
		lfile << str << flush;
		binsum -= 256;
		next++;
	} else {
		sprintf(str,"  fitpar --");
		lfile << str;
		centre_iso2_err = 0;
		sprintf(str,"  centre_iso2: %10.5g +/- %10.5g\n",
			centre_iso2,centre_iso2_err);
		lfile << str;
	}

	if ( binsum > 127 ) {
//		fwhm_iso = FIT(next);
		fwhm_iso2_err = ERR(next)*fwhm_iso2;
		sprintf(str,"  fitpar %2d",next);
		lfile << str;
		sprintf(str,"    fwhm_iso2: %10.5g +/- %10.5g\n",fwhm_iso2,fwhm_iso2_err);
		lfile << str << flush;
		sprintf(str,"       (raw): %10.5g +/- %10.5g\n",FIT(next),ERR(next));
		lfile << str << flush;
		binsum -= 128;
		next++;
	} else {
		sprintf(str,"  fitpar --");
		lfile << str;
		fwhm_iso2_err = 0;
		sprintf(str,"    fwhm_iso2: %10.5g +/- %10.5g\n",fwhm_iso2,fwhm_iso2_err);
		lfile << str << flush;
	}

	if ( binsum > 63 ) {
//		iso_A = FIT(next);
		iso_A2_err = iso_A2 * ERR(next);
		sprintf(str,"  fitpar %2d",next);
		lfile << str;
		sprintf(str,"       iso_A2: %10.5g +/- %10.5g\n",iso_A2,iso_A2_err);
		lfile << str << flush;
		sprintf(str,"       (raw): %10.5g +/- %10.5g\n",FIT(next),ERR(next));
		lfile << str << flush;
		binsum -= 64;
		next++;
	} else {
		sprintf(str,"  fitpar --");
		lfile << str;
		iso_A2_err = 0.0;
		sprintf(str,"    iso_A2: %10.5g +/- %10.5g\n",iso_A2,iso_A2);
		lfile << str << flush;
	}

	if ( binsum > 31 ) {
//		centre_iso = FIT(next);
		centre_iso_err = ERR(next);
		sprintf(str,"  fitpar %2d",next);
		lfile << str;
		sprintf(str,"  centre_iso: %10.5g +/- %10.5g\n",
			centre_iso,centre_iso_err);
		lfile << str << flush;
		binsum -= 32;
		next++;
	} else {
		sprintf(str,"  fitpar --");
		lfile << str;
		centre_iso_err = 0;
		sprintf(str,"  centre_iso: %10.5g +/- %10.5g\n",
			centre_iso,centre_iso_err);
		lfile << str;
	}

	if ( binsum > 15 ) {
//		fwhm_iso = FIT(next);
		fwhm_iso_err = ERR(next)*fwhm_iso;
		sprintf(str,"  fitpar %2d",next);
		lfile << str;
		sprintf(str,"    fwhm_iso: %10.5g +/- %10.5g\n",fwhm_iso,fwhm_iso_err);
		lfile << str << flush;
		sprintf(str,"       (raw): %10.5g +/- %10.5g\n",FIT(next),ERR(next));
		lfile << str << flush;
		binsum -= 16;
		next++;
	} else {
		sprintf(str,"  fitpar --");
		lfile << str;
		fwhm_iso_err = 0;
		sprintf(str,"    fwhm_iso: %10.5g +/- %10.5g\n",fwhm_iso,fwhm_iso_err);
		lfile << str << flush;
	}

	if ( binsum > 7 ) {
//		iso_A = FIT(next);
		iso_A_err = iso_A * ERR(next);
		sprintf(str,"  fitpar %2d",next);
		lfile << str;
		sprintf(str,"       iso_A: %10.5g +/- %10.5g\n",iso_A,iso_A_err);
		lfile << str << flush;
		sprintf(str,"       (raw): %10.5g +/- %10.5g\n",FIT(next),ERR(next));
		lfile << str << flush;
		binsum -= 8;
		next++;
	} else {
		sprintf(str,"  fitpar --");
		lfile << str;
		iso_A_err = 0.0;
		sprintf(str,"    iso_A2: %10.5g +/- %10.5g\n",iso_A2,iso_A2);
		lfile << str << flush;
	}

	if ( binsum > 3 ) {
//		centre = FIT(next);
		centre_err = ERR(next);
		sprintf(str,"  fitpar %2d",next);
		lfile << str;
		binsum -= 4;
		next++;
	} else {
		sprintf(str,"  fitpar --");
		lfile << str;
		centre_err = 0.0;
	}
	sprintf(str,"      centre: %10.5g +/- %10.5g\n",centre,centre_err);
	lfile << str << flush;

	if ( binsum > 1 ) {
//		fwhm = FIT(next);
		fwhm_err = fwhm * ERR(next);
		sprintf(str,"  fitpar %2d",next);
		lfile << str;
		sprintf(str,"        fwhm: %10.5g +/- %10.5g\n",fwhm,fwhm_err);
		lfile << str << flush;
		sprintf(str,"       (raw): %10.5g +/- %10.5g\n",FIT(next),ERR(next));
		lfile << str << flush;
		binsum -= 2;
		next++;
	} else {
		sprintf(str,"  fitpar --");
		lfile << str;
		fwhm_err = 0.0;
	}

	if ( binsum > 0 ) {
		binsum -= 1;
		for (int ss=0; ss<numsb; ss++) {
			if (ss==maxsb+1) {
				sprintf(str,"  fitpar --");
				lfile << str;
				sprintf(str,"       sb-01: %10.5g +/- %10.5g\n",
					gvg(lorentz_An,ss),0.0);
				lfile << str << flush;
			} else {
				sprintf(str,"  fitpar %2d",next);
				lfile << str;
				gvs(lorentz_An_err, ss, gvg(lorentz_An,ss)*ERR(next) );
				sprintf(str,"       sb%03d: %10.5g +/- %10.5g\n",
					maxsb-ss,gvg(lorentz_An,ss)*0.31830988,
					gvg(lorentz_An_err,ss)*0.31830988 );
				lfile << str;
				sprintf(str,"       (raw): %10.5g +/- %10.5g\n",FIT(next),ERR(next));
				lfile << str << flush;
				next++;
			}
		}
	}

	lfile << " Final Fit Parameter Covariances:\n";
	for (int f=0; f < numfloatpars; f++ ) {
		for (int g=0; g <= f; g++ ) {
			sprintf(str,"%6.3f ",COV(f,g));
			lfile << str;
		}
		lfile << endl << flush;
	}

	gsl_vector_free(floaters);
	gsl_vector_free(nlfit_residual);
	gsl_matrix_free(nlfit_jacobian);
	gsl_matrix_free(covar);
	return(0);


}

int _nlfit_f(const gsl_vector *f, void *P, gsl_vector *resid) {

	// Make a proper Pake object from the refernce passed in under void type
	SBFIT * G = (class SBFIT *)P;

	G->interpret_floaters( f );

	// interpret fitconfig
//	int multibranch = 0;
	int fc_tmp=G->fitconfig;
	int bracket = 0;
	gsl_vector_view rvv;
	if ( fc_tmp > 3 ) {
		G->calc_spectrum();
		rvv = gsl_vector_subvector(resid,bracket*G->numdata,G->numdata);
		gsl_vector_memcpy(&rvv.vector,&(G->residual_spe.vector) );
		bracket++;
		fc_tmp -= 4;
	} 
	if ( fc_tmp > 1 ) {
		G->calc_integral();
		rvv = gsl_vector_subvector(resid,bracket*G->numdata,G->numdata);
		gsl_vector_memcpy(&rvv.vector,&(G->residual_int.vector) );
		bracket++;
		fc_tmp -= 2;
	}
	if (fc_tmp > 0 ) {
		G->calc_derivative();
		rvv = gsl_vector_subvector(resid,bracket*G->numdata,G->numdata);
		gsl_vector_memcpy(&rvv.vector,&(G->residual_der.vector) );
		bracket++;
		fc_tmp -= 1;
	}
	if (fc_tmp != 0 ) {
		cerr << "fitting/residual writing not working properly" << endl;
		exit(0);
	}

	return GSL_SUCCESS;
}

int _nlfit_df (const gsl_vector *f, void *P, gsl_matrix *J) {

	// Make a proper Pake object from the refernce passed in under void type
	SBFIT * G = (class SBFIT *)P;

	G->interpret_floaters( f );

	// interpret fitconfig
//	int multibranch = 0; 
	int fc_tmp=G->fitconfig;
	int bracket = 0;
	gsl_matrix_view jvv;
	gsl_matrix_view pvv;
	int bin[9] = { 256, 128, 64, 32, 16, 8, 4, 2, 1 };

	if ( fc_tmp > 3 ) {
		G->calc_spectrum_partials();
		int next = 0;
		int binsum = G->binarysum;
		int testbinsum = 511;
		for (int b=0; b < G->nvar; b++) {
			testbinsum -= bin[b];
			if ( binsum > testbinsum ) {
				jvv = gsl_matrix_submatrix(J,bracket*G->numdata,next,G->numdata,1);
				pvv = gsl_matrix_submatrix(G->J,bracket*G->numdata,b,G->numdata,1);
				gsl_matrix_memcpy(&jvv.matrix,&pvv.matrix);
				binsum -= bin[b];
				next++;
			}
		}
		if ( binsum > 0 ) {
			binsum -= 1;
			for (int s=0; s < G->numsb; s++) {
				if (s == G->maxsb+1) continue;
				int si = G->nvar+s;	
				if (s > G->maxsb+1) si = G->nvar+s-1;
				jvv = gsl_matrix_submatrix(J,bracket*G->numdata,next,G->numdata,1);
				pvv = gsl_matrix_submatrix(G->J,bracket*G->numdata,si,G->numdata,1);
				gsl_matrix_memcpy(&jvv.matrix,&pvv.matrix);
				next++;
			}
		}
		if (binsum != 0 ) {
			cerr << "error in copying J_spe parameters" << endl;
			exit(0);
		}
		fc_tmp -= 4;
		bracket++;
	} 
	if ( fc_tmp > 1 ) {
		G->calc_integral_partials();
		int next = 0;
		int binsum = G->binarysum;
		int testbinsum = 511;
		for (int b=0; b < G->nvar; b++) {
			testbinsum -= bin[b];
			if ( binsum > testbinsum ) {
				jvv = gsl_matrix_submatrix(J,bracket*G->numdata,next,G->numdata,1);
				pvv = gsl_matrix_submatrix(G->J,bracket*G->numdata,b,G->numdata,1);
				gsl_matrix_memcpy(&jvv.matrix,&pvv.matrix);
				binsum -= bin[b];
				next++;
			}
		}
		if ( binsum > 0 ) {
			binsum -= 1;
			for (int s=0; s < G->numsb; s++) {
				if (s == G->maxsb+1) continue;
				int si = G->nvar+s;	
				if (s > G->maxsb+1) si = G->nvar+s-1;
				jvv = gsl_matrix_submatrix(J,bracket*G->numdata,next,G->numdata,1);
				pvv = gsl_matrix_submatrix(G->J,bracket*G->numdata,si,G->numdata,1);
				gsl_matrix_memcpy(&jvv.matrix,&pvv.matrix);
				next++;
			}
		}
		if (binsum != 0 ) {
			cerr << "error in copying J_int parameters" << endl;
			exit(0);
		}
		bracket++;
		fc_tmp -= 2;
	}
	if (fc_tmp > 0 ) {
		G->calc_derivative_partials();
		int next = 0;
		int binsum = G->binarysum;
		int testbinsum = 511;
		for (int b=0; b < G->nvar; b++) {
			testbinsum -= bin[b];
			if ( binsum > testbinsum ) {
				jvv = gsl_matrix_submatrix(J,bracket*G->numdata,next,G->numdata,1);
				pvv = gsl_matrix_submatrix(G->J,bracket*G->numdata,b,G->numdata,1);
				gsl_matrix_memcpy(&jvv.matrix,&pvv.matrix);
				binsum -= bin[b];
				next++;
			}
		}
		if ( binsum > 0 ) {
			binsum -= 1;
			for (int s=0; s<G->numsb; s++) {
				if (s == G->maxsb+1) continue;
				int si = G->nvar+s;	
				if (s > G->maxsb+1) si = G->nvar+s-1;
				jvv = gsl_matrix_submatrix(J,bracket*G->numdata,next,G->numdata,1);
				pvv = gsl_matrix_submatrix(G->J,bracket*G->numdata,si,G->numdata,1);
				gsl_matrix_memcpy(&jvv.matrix,&pvv.matrix);
				next++;
			}
		}
		if (binsum != 0 ) {
			cerr << "error in copying J_der parameters" << endl;
			exit(0);
		}
		bracket++;
		fc_tmp -= 1;
	}
	if (fc_tmp != 0 ) {
		cerr << "fitting/residual writing not working properly" << endl;
		exit(0);
	}

	return GSL_SUCCESS;
}

int _nlfit_fdf (const gsl_vector *f, void *P, gsl_vector *resid, 
	gsl_matrix *J) {

	_nlfit_f (f,P,resid);
	_nlfit_df (f,P,J);

	return GSL_SUCCESS;
}

int SBFIT::interpret_floaters( const gsl_vector *floaters ) {
	// step down through the possibilities and interpret as we go along
	int next = 0;
	int binsum = binarysum;
	if ( binsum > 255 ) {
		centre_iso2 = gvg(floaters,next);
		binsum -= 256;
		next++;
	}
	if ( binsum > 127 ) {
		fwhm_iso2 = exp( gvg(floaters,next) );
		binsum -= 128;
		next++;
	}
	if ( binsum > 63 ) {
		iso_A2 = exp( gvg(floaters,next) );
		binsum -= 64;
		next++;
	}
	if ( binsum > 31 ) {
		centre_iso = gvg(floaters,next);
		binsum -= 32;
		next++;
	}
	if ( binsum > 15 ) {
		fwhm_iso = exp( gvg(floaters,next) );
		binsum -= 16;
		next++;
	}
	if ( binsum > 7 ) {
		iso_A = exp( gvg(floaters,next) );
		binsum -= 8;
		next++;
	}
	if ( binsum > 3 ) {
		centre = gvg(floaters,next );
		binsum -= 4;
		next++;
	}
	if ( binsum > 1 ) {
		fwhm = exp( gvg(floaters,next) );
		binsum -= 2;
		next++;
	}
	if ( binsum > 0 ) {
		double sum1 = 0.0;
		binsum -= 1;
		for (int s=0; s<numsb; s++) {
			if (s==maxsb+1) continue;
			gvs( lorentz_An, s, exp( gvg(floaters,next) ) );
			sum1 += double(maxsb-s)*gvg(lorentz_An,s);
			next++;
		}
		gvs(lorentz_An, maxsb+1, sum1);
	}
	if (binsum != 0 ) {
		cerr << "error in interpretation of float parameters" << endl;
		exit(0);
	}
	return(0);
}

int SBFIT::print_state_fit (size_t iter, gsl_multifit_fdfsolver * s) {
	char str[81];
	sprintf(str,"iter %3u   |f(x)| = %15.12e\n",
		iter,gsl_blas_dnrm2 (s->f) );
	lfile << str; //cout << str;
//	#define FIT(i) gsl_vector_get(s->x,i)
//	for (int sp=0; sp < numCDn+3; sp++) {
//		sprintf(str,"  %6.3g\n",gsl_vector_get(s->x,sp));
//		lfile << str;
//	}
//	cout << endl;
	lfile << endl;
	return(0);
}

int SBFIT::open_write_file (const char *ext, ofstream *outfile ) {
    char filename[256];
	outpath = getenv("OUTPATH");
	outbase = getenv("OUTBASE");

    strcpy(filename,outpath);
    strcat(filename,"/" );
    strcat(filename,outbase);
    strcat(filename,"/" );
    strcat(filename,outbase);
    strcat(filename,ext);
    
	outfile->open(filename, ios::out);
	if ( !outfile->is_open() ) {
		cerr << "Could not open input file " << filename << endl;
		exit(0);
	}
	return(0);
}

int SBFIT::open_append_file ( const char *ext, ofstream *appfile, char *name ) {
	outpath = getenv("OUTPATH");
	outbase = getenv("OUTBASE");

    strcpy(name,outpath);
    strcat(name,"/" );
    strcat(name,outbase);
    strcat(name,"/" );
    strcat(name,outbase);
    strcat(name,ext);
    
	appfile->open(name, ios::app);
	if ( !appfile->is_open() ) {
		cerr << "Could not open input file " << name << endl;
		exit(0);
	}
	return(0);
}

int SBFIT::inputpar( char *key, char *val) {
	char str[2048];
	if (! strcmp(key, "SPECTRUMFILE") ) {
		strcpy(spectrumfile,val); 
		sprintf(str,"INPUT SPECTRUMFILE (object.spectrumfile) = %s\n",spectrumfile);
		lfile << str << flush;
	} else if (! strcmp(key, "INTEGRALFILE") ) {
		strcpy(integralfile,val); 
		sprintf(str,"INPUT INTEGRALFILE (object.integralfile) = %s\n",integralfile);
		lfile << str << flush;
	} else if (! strcmp(key, "DERIVATIVEFILE") ) {
		strcpy(derivativefile,val); 
		sprintf(str,"INPUT DERIVATIVEFILE (object.derivativefile) = %s\n",derivativefile);
		lfile << str << flush;
	} else if (! strcmp(key, "DATAFITFILE") ) {
		strcpy(datafitfile,val); 
		sprintf(str,"INPUT DATAFITFILE (object.datafitfile) = %s\n",datafitfile);
		lfile << str << flush;
	} else if (! strcmp(key, "OUTFILESET") ) {
		strcpy(outfileset,val); 
		sprintf(str,"INPUT OUTFILESET (object.outfileset) = %s\n",outfileset);
		lfile << str << flush;
	} else if (! strcmp(key, "APPENDOUTFILE") ) {
		strcpy(appendoutfile,val); 
		sprintf(str,"INPUT APPENDOUTFILE (object.appendoutfile) = %s\n",
			appendoutfile);
		lfile << str << flush;
	} else if (! strcmp(key, "NUMDATA") )	{
		ssf_int( val, &numdata);
		sprintf(str,"INPUT NUMDATA (object.numdata) = %i\n",numdata);
		lfile << str << flush;
	} else if (! strcmp(key, "MINSB") )	{
		ssf_int( val, &minsb);
		sprintf(str,"INPUT MINSB (object.minsb) = %i\n",minsb);
		lfile << str << flush;
	} else if (! strcmp(key, "MAXSB") )	{
		ssf_int( val, &maxsb);
		sprintf(str,"INPUT MAXSB (object.maxsb) = %i\n",maxsb);
		lfile << str << flush;
	} else if (! strcmp(key, "MAXITER") )	{
		ssf_int( val, &maxit);
		sprintf(str,"INPUT MAXITER (object.maxit) = %i\n",maxit);
		lfile << str << flush;
	} else if (! strcmp(key, "FITCONFIG") )	{
		ssf_int( val, &fitconfig);
		sprintf(str,"INPUT FITCONFIG (object.fitconfig) = %i\n",fitconfig);
		lfile << str << flush;
	} else if (! strcmp(key, "MAS") )	{
		ssf_double( val, &mas);
		sprintf(str,"INPUT MAS (object.mas) = %f\n",mas);
		lfile << str << flush;
	} else if (! strcmp(key, "AMPS") )	{
		ssf_doublearray( val, &numsb, &rawamps);
		sprintf(str,"INPUT AMPS (object.rawamps) = %s\n",val);
		lfile << str << flush;
		for (int i=0; i<numsb; i++) {
			sprintf(str,"%f ",rawamps[i]);
			lfile << str << endl;
		}
		lfile << endl;
	} else if (! strcmp(key, "AMPISO") )	{
		ssf_double( val, &iso_A);
		sprintf(str,"INPUT AMPISO (object.iso_A) = %f\n",iso_A);
		lfile << str << flush;
	} else if (! strcmp(key, "FWHM") )	{
		ssf_double( val, &fwhm);
		sprintf(str,"INPUT FWHM (object.fwhm) = %f\n",fwhm);
		lfile << str << flush;
	} else if (! strcmp(key, "FWHMISO") )	{
		ssf_double( val, &fwhm_iso);
		sprintf(str,"INPUT FWHMISO (object.fwhm_iso) = %f\n",fwhm_iso);
		lfile << str << flush;
	} else if (! strcmp(key, "ALPHA") )	{
		ssf_double( val, &alpha);
		sprintf(str,"INPUT ALPHA (object.alpha) = %f\n",alpha);
		lfile << str << flush;
	} else if (! strcmp(key, "CENTRE") )	{
		ssf_double( val, &centre);
		sprintf(str,"INPUT CENTRE (object.centre) = %f\n",centre);
		lfile << str << flush;
	} else if (! strcmp(key, "CENTREISO") )	{
		ssf_double( val, &centre_iso);
		sprintf(str,"INPUT CENTREISO (object.centre_iso) = %f\n",centre_iso);
		lfile << str << flush;
	} else if (! strcmp(key, "CENTREISO2") )	{
		ssf_double( val, &centre_iso2);
		sprintf(str,"INPUT CENTREISO2 (object.centre_iso2) = %f\n",centre_iso2);
		lfile << str << flush;
	} else if (! strcmp(key, "FWHMISO2") )	{
		ssf_double( val, &fwhm_iso2);
		sprintf(str,"INPUT FWHMISO2 (object.fwhm_iso2) = %f\n",fwhm_iso2);
		lfile << str << flush;
	} else if (! strcmp(key, "AMPISO2") )	{
		ssf_double( val, &iso_A2);
		sprintf(str,"INPUT AMPISO2 (object.iso_A2) = %f\n",iso_A2);
		lfile << str << flush;
	} else if (! strcmp(key, "FITAMPS") )	{
		ssf_int( val, &fit_An);
		sprintf(str,"INPUT FITAMPS (object.fit_An) = %i\n",fit_An);
		lfile << str << flush;
	} else if (! strcmp(key, "FITAMPISO") )	{
		ssf_int( val, &fit_iso_A);
		sprintf(str,"INPUT FITAMPISO (object.fit_iso_A) = %i\n",fit_iso_A);
		lfile << str << flush;
	} else if (! strcmp(key, "FITAMPISO2") )	{
		ssf_int( val, &fit_iso_A2);
		sprintf(str,"INPUT FITAMPISO2 (object.fit_iso_A2) = %i\n",fit_iso_A2);
		lfile << str << flush;
	} else if (! strcmp(key, "FITFWHM") )	{
		ssf_int( val, &fit_fwhm);
		sprintf(str,"INPUT FITFWHM (object.fit_fwhm) = %i\n",fit_fwhm);
		lfile << str << flush;
	} else if (! strcmp(key, "FITALPHA") )	{
		ssf_int( val, &fit_alpha);
		sprintf(str,"INPUT FITALPHA (object.fit_alpha) = %i\n",fit_alpha);
		lfile << str << flush;
	} else if (! strcmp(key, "FITFWHMISO") )	{
		ssf_int( val, &fit_fwhm_iso);
		sprintf(str,"INPUT FITFWHMISO (object.fit_fwhm_iso) = %i\n",fit_fwhm_iso);
		lfile << str << flush;
	} else if (! strcmp(key, "FITFWHMISO2") )	{
		ssf_int( val, &fit_fwhm_iso2);
		sprintf(str,"INPUT FITFWHMISO2 (object.fit_fwhm_iso2) = %i\n",
			fit_fwhm_iso2);
		lfile << str << flush;
	} else if (! strcmp(key, "FITCENTRE") )	{
		ssf_int( val, &fit_centre);
		sprintf(str,"INPUT FITCENTRE (object.fit_centre) = %i\n",fit_centre);
		lfile << str << flush;
	} else if (! strcmp(key, "FITCENTREISO") )	{
		ssf_int( val, &fit_centre_iso);
		sprintf(str,"INPUT FITCENTREISO (object.fit_centre_iso) = %i\n",
			fit_centre_iso);
		lfile << str << flush;
	} else if (! strcmp(key, "FITCENTREISO2") )	{
		ssf_int( val, &fit_centre_iso2);
		sprintf(str,"INPUT FITCENTREISO2 (object.fit_centre_iso2) = %i\n",
			fit_centre_iso2);
		lfile << str << flush;
	} else if (! strcmp(key, "SIGMASPE") )	{
		ssf_double( val, &sigmaspe);
		sprintf(str,"INPUT SIGMASPE (object.sigmaspe) = %f\n",sigmaspe);
		lfile << str << flush;
	} else if (! strcmp(key, "SIGMAINT") )	{
		ssf_double( val, &sigmaint);
		sprintf(str,"INPUT SIGMAINT (object.sigmaint) = %f\n",sigmaint);
		lfile << str << flush;
	} else if (! strcmp(key, "SIGMADER") )	{
		ssf_double( val, &sigmader);
		sprintf(str,"INPUT SIGMADER (object.sigmader) = %f\n",sigmader);
		lfile << str << flush;
	} else if (! strcmp(key, "NMINFREQ") )	{
		ssf_double( val, &nminfreq);
		sprintf(str,"INPUT NMINFREQ (object.nminfreq) = %f\n",nminfreq);
		lfile << str << flush;
	} else if (! strcmp(key, "NMAXFREQ") )	{
		ssf_double( val, &nmaxfreq);
		sprintf(str,"INPUT NMAXFREQ (object.nmaxfreq) = %f\n",nmaxfreq);
		lfile << str << flush;
	} else { 
		cerr << "I Don't Recognize " << key << endl;
		exit(0);
	}
	return(0);
}

int SBFIT::parseinputs( int c, char *v[] ) {
	strcpy(code,v[0]);
	lfile << endl;
	lfile << "**************************************\n";
	lfile << "** CODE: " << code << endl;
	lfile << "** ";
	simpletimestamp(&lfile); 
	lfile << "**\n";
	for (int i=1; i<c; i+=2) {
		inputpar( v[i], v[i+1] );
	}
	return(0);
}	

int SBFIT::parseinputline( string s ) {
	char key[80];
	char val[2048];

	// strip leading # off if present
 	string s1 (s,0,1);
	size_t hashtest = s1.compare("#");
	string s2(s);
	if (hashtest==0) {
		s2 = s.substr(1);
	} 
	s.swap(s2);	

	// some compilers can't handle the quoted strings. 
	size_t qqtest = s.find("\"");
	if (qqtest!=string::npos) {
		sscanf(s.c_str(),"%s \"%[^\"]\"",key,val);
	} else {
		// Read in key and value
		sscanf(s.c_str(),"%s %s",key,val);
	}
	inputpar( key, val);
	return(0);
}	
