#ifndef SBFIT_INCLUDED_
#define SBFIT_INCLUDED_


#include <sys/stat.h>
#include "inputs.h"
#include "utils.h"
#include "cputhreadtime.h"
#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
using namespace std;

int _nlfit_f(const gsl_vector *, void *, gsl_vector *);
int _nlfit_df(const gsl_vector *, void *, gsl_matrix *);
int _nlfit_fdf(const gsl_vector *, void *, gsl_vector *, gsl_matrix *);

class SBFIT {
public:
	SBFIT();
	~SBFIT();
	SBFIT( const SBFIT& );
	int copy_min( const SBFIT& ); 

	int append_logfile();
	int read_data();
	int parseinputs( int, char ** );
	int parseinputline( string );
	int inputpar( char *, char *);
	int open_append_file ( const char *, ofstream *, char * );
	int open_write_file ( const char *, ofstream * );
	int count_xydata( char * );
	int read_xydata( char * );
	int read_spectrum();
	int read_integral();
	int read_derivative();
	int init_data( );
	int write_datafit();
	int read_datafit();
	int write_lrntz();
	int write_memasdata();
	int write_gnuplotscript();
	int init_fit_params();
	int calc_spectrum();
	int calc_integral();
	int calc_derivative();
	int calc_spectrum_partials();
	int calc_integral_partials();
	int calc_derivative_partials();
	int nonlinfit();
	int print_state_fit (size_t, gsl_multifit_fdfsolver *);
	int interpret_floaters( const gsl_vector * );
	int noise_stats();

//	int calc_jacobian();
	int update_log( char *);
	
	char * outpath;		// path to output directory (subdir will be created)
	char * outbase;		// basename to be used for output files
	char * safety;		// whether to prompt for overwrite confirmation

	int constant;				
	gsl_matrix * matrix_element;
	gsl_vector * vector_element;
	gsl_vector * xdata;
	gsl_vector * ydata;

	// frequency values (the x data axis)
	gsl_vector * freq;
	// spectrum, integrals, and derivative y axis values,
	// concatentated into a single vector but with vector views
	gsl_vector * spintder;
	gsl_vector_view spectrum;
	gsl_vector_view integral;
	gsl_vector_view derivative;

	// MAS rate
	double mas;
	double alpha;
	int fit_alpha;

	// minimum (maximally negative) and maximum (positive) sideband index
	int minsb;
	int maxsb;
	// number of sidebands;
	int numsb;
	int numdata;

	gsl_vector * spintder_fit;
	gsl_vector_view spectrum_fit;
	gsl_vector_view integral_fit;
	gsl_vector_view derivative_fit;

	// matrix of individual lorentzians
	gsl_matrix * lorentz;
	gsl_matrix_view lorentz_spe_fit;
	gsl_matrix_view lorentz_int_fit;
	gsl_matrix_view lorentz_der_fit;

	// vector for isotropic lorentzian
	gsl_vector * lorentz_iso;
	gsl_vector_view lorentz_iso_spe_fit;
	gsl_vector_view lorentz_iso_int_fit;
	gsl_vector_view lorentz_iso_der_fit;

	// vector for isotropic lorentzian
	gsl_vector * lorentz_iso2;
	gsl_vector_view lorentz_iso2_spe_fit;
	gsl_vector_view lorentz_iso2_int_fit;
	gsl_vector_view lorentz_iso2_der_fit;

	// Utilities for calculating jacobians etc
	gsl_matrix * Ldfreq;
	gsl_matrix * Uint;
	gsl_matrix * Uint2;
	gsl_matrix * Vint;
	gsl_matrix * Vint2;
	gsl_matrix * QintQnu0_elem;
	gsl_matrix * Uder;
	gsl_matrix * Vder;
	gsl_matrix * Vder2;

	gsl_vector * L_iso_dfreq;
	gsl_vector * QintQnu0_iso_elem;
	gsl_vector * Uint_iso;
	gsl_vector * Vint_iso;
	gsl_vector * Uder_iso;
	gsl_vector * Vder_iso;
	gsl_vector * Vder2_iso;

	gsl_vector * L_iso2_dfreq;
	gsl_vector * QintQnu0_iso2_elem;
	gsl_vector * Uint_iso2;
	gsl_vector * Vint_iso2;
	gsl_vector * Uder_iso2;
	gsl_vector * Vder_iso2;
	gsl_vector * Vder2_iso2;

	// 1 for spectrum, 2 for integral, 4 for derivative
	int fitconfig;
	// code for which parameters to float
	int binarysum;

	// lorentzian lineshape parameters
	int nvar;
	gsl_vector * lorentz_An;
	double iso_A;
	double fwhm;
	double fwhm_iso;
	double centre;
	double centre_iso;
	double fwhm_iso2;
	double centre_iso2;
	double iso_A2;

	gsl_vector * lorentz_An_err;	
	double iso_A_err;
	double fwhm_err;
	double fwhm_iso_err;
	double centre_err;
	double centre_iso_err;
	double iso_A2_err;
	double fwhm_iso2_err;
	double centre_iso2_err;

	// flag to indicate whether parameter should be optimised or not
	int fit_An;
	int fit_iso_A;
	int fit_fwhm;
	int fit_fwhm_iso;
	int fit_centre;
	int fit_centre_iso;
	int fit_iso_A2;
	int fit_fwhm_iso2;
	int fit_centre_iso2;

	// residuals and jacobian object components
	// for interchange with gsl nonlin fit functions
	gsl_vector * residual;
	gsl_vector_view residual_spe;
	gsl_vector_view residual_int;
	gsl_vector_view residual_der;

	gsl_matrix * J;				// jacobian -- d(sideband)/d(lambda)
	gsl_matrix_view J_spe;
	gsl_matrix_view J_int;
	gsl_matrix_view J_der;

//	nlfit_residual;
//	nlfit_jacobian;

	ofstream lfile;			// logfile 
	char logfilename[256];
	char datafile[256];		// speed index amp error in a file
	char datafitfile[256];		// speed index amp error in a file
	char appendoutfile[260];
	char spectrumfile[256];
	char integralfile[256];
	char derivativefile[256];
	char outfileset[256];
	

	// input variables
	char code[80];

	// for input of amplitudes from file
	double * rawamps;

	int maxit;
	double sigmaspe,sigmaint,sigmader;
		
	// noise stats variables
	double nminfreq, nmaxfreq;
	double nstdev_spe, nstdev_int, nstdev_der;
	double navg_spe, navg_int, navg_der;

};

#endif
