#include "utils.h"

int iscomment ( string str ) {
	string ignore( "#" );
	if ( !str.compare(0,1,ignore,0,1) ) return(1);
	else return(0);
}

void DEBUG (int i ) { cout << "DEBUG: " << i << endl; }

double gvg(const gsl_vector *v, int i) {
	return gsl_vector_get(v,i);
}

void gvs(gsl_vector *v, int i, double d) {
	return gsl_vector_set(v,i,d);
}

double gmg(const gsl_matrix *m, int c, int r) {
	return gsl_matrix_get(m,c,r);
}

void gms(gsl_matrix *v, int i, int j, double d) {
	return gsl_matrix_set(v,i,j,d);
}

int report_array(const gsl_vector *v, int n) {
	char str[179];
	for (int w=0; w<n; w++) {
		sprintf(str," %9.2le", gvg(v,w));
		cout << str;
	}
	cout << endl;
	return(0);
}

int ssf_int( char str[], int *i) {
	sscanf(str,"%i",i);
	return(0);
}
int ssf_double( char str[], double *d) {
	sscanf(str,"%le",d);
	return(0);
}
int ssf_doublearray( char str[], int *i, double **d) {
	// count how many doubles in the string
	int ptr = 0; int ptr_i = 0;
	double tmp;
	(*i) = -1;
	do {
		(*i)++;
		ptr_i = 0;
		sscanf(str+ptr,"%le%n",&tmp,&ptr_i);
		ptr += ptr_i;
	} while (ptr_i != 0);

	(*d) = new double[(*i)];
	// go back and read doubles into array 
	ptr = ptr_i = 0;
	for (int ii=0; ii<(*i); ii++) {
		ptr_i = 0;
		sscanf(str+ptr,"%le%n",&tmp,&ptr_i);
		(*d)[ii] = tmp;
		ptr += ptr_i;
	}
	return(0);
}

int ssf_intarray( char str[], int *i, int **d) {
	// count how many ints in the string
	int ptr = 0; int ptr_i = 0;
	int tmp;
	(*i) = -1;
	do {
		(*i)++;
		ptr_i = 0;
		sscanf(str+ptr,"%d%n",&tmp,&ptr_i);
		ptr += ptr_i;
	} while (ptr_i != 0);

	(*d) = new int[(*i)];
	// go back and read ints into array 
	ptr = ptr_i = 0;
	for (int ii=0; ii<(*i); ii++) {
		ptr_i = 0;
		sscanf(str+ptr,"%d%n",&tmp,&ptr_i);
		(*d)[ii] = tmp;
		ptr += ptr_i;
	}
	return(0);
}

int ssf_count_stringarray( char str[], int *i ) {
	// count how many labels in the string
	int ptr = 0; int ptr_i = 0;
	char tmp[128];
	(*i) = -1;
	do {
		(*i)++;
		ptr_i = 0;
		sscanf(str+ptr,"%[^;]%n",tmp,&ptr_i);
		tmp[0]='\0';
		ptr += ptr_i + 1;
	} while (ptr_i != 0);
	return(0);
}

int ssf_stringarray( char str[], int i, char (*c)[30] ) {
	// go back and read doubles into array 
	int ptr = 0; int ptr_i = 0;
	char tmp[128];
	ptr = ptr_i = 0;
	for (int ii=0; ii<i; ii++) {
		ptr_i = 0;
		sscanf(str+ptr,"%[^;]%n",tmp,&ptr_i);
		sprintf(c[ii],"%s",tmp);
		tmp[0]='\0';
		ptr += ptr_i + 1;
	}
	return(0);
}

/*
REFERNCE
const char line[] = "2004/12/03 12:01:59;info1;info2;info3";
const char *ptr = line;
char field [ 32 ];
int n;

while (*ptr != '\0') {
	int items_read = sscanf(ptr, "%31[^;]%n", field, &n);
	printf("field = \"%s\"\n", field);
	field[0]='\0';

	// advance the pointer by the number of characters read
	if (items_read == 1) ptr += n;
	// didn't find an expected delimiter, done?
	if ( *ptr != ';' ) break; 
	// skip the delimiter
	++ptr; 
}
*/

int gen_random_seed( int rin ) {
	//	reverse the current process ID and 
	//  scramble with an input (e.g. $SLURM_JOB_ED)
	pid_t pid;
	if ((pid = getpid()) < 0) {
		perror("unable to get pid\n");
	} 

	int num = int(pid);
	int rev = 0;
	int digits = log10(num)+1;
	for (int i=0; i< digits; i++) {
		rev *= 10;
		rev += (num % 10);
		num /= 10;
	}
	
	digits = log10(rin)+1;
	int nir = 0;
	for (int i=0; i< digits; i++) {
        nir *= 10;
        nir += (rin % 10);
        rin /= 10;
    }

	int rseed;
	if (rev > nir) {
		rseed = rev - nir;
	} else {
		rseed = rev + nir;
	}
	return(rseed);
}

int simpletimestamp ( ofstream * log ) {
	time_t seconds = time(NULL);
	*log << "TIME: " << ctime(&seconds) << endl;
	return(0);
}

int gvf ( gsl_vector * a, double (*functptr)(double)   ) {
  const size_t N = a->size;
  const size_t stride_a = a->stride;
  size_t i;

  for (i = 0; i < N; i++) {
     a->data[i * stride_a] = functptr( a->data[i * stride_a] );
  }

   return GSL_SUCCESS;
}

int gmf ( gsl_matrix * a, double (*functptr)(double)   ) {
  const size_t N1 = a->size1;
  const size_t N2 = a->size2;
  const size_t tda_a = a->tda;
  size_t i,j;

  for (i = 0; i < N1; i++) {
	for (j = 0; j < N2; j++ ) {
     a->data[i * tda_a + j] = functptr( a->data[i * tda_a + j] );
	}
  }

   return GSL_SUCCESS;
}

int gvfd ( gsl_vector * a, double (*functptr)(double,double), double p) {
  const size_t N = a->size;
  const size_t stride_a = a->stride;
  size_t i;

  for (i = 0; i < N; i++) {
     a->data[i * stride_a] = functptr( a->data[i * stride_a], p );
  }

   return GSL_SUCCESS;
}

int gmfd ( gsl_matrix * a, double (*functptr)(double,double), double p) {
  const size_t N1 = a->size1;
  const size_t N2 = a->size2;
  const size_t tda_a = a->tda;
  size_t i,j;

  for (i = 0; i < N1; i++) {
	for (j = 0; j < N2; j++ ) {
     a->data[i * tda_a + j] = functptr( a->data[i * tda_a + j], p );
	}
  }

   return GSL_SUCCESS;
}

double vpower (double d, double p) { return pow( d,p ); }
double vln (double d) { return log(abs(d)); }
double vsqrt (double d) { return pow( d,0.5); }
double vatan (double d) { return atan( d ); }
double vinvert (double d) { return 1.0/d; }
double vsquare (double d) { return pow( d,2); }

int index_of_closest_value( const gsl_vector *v, double n) {
	int closest;
	double mindist = 1E12;
	for (int vi=0; vi<v->size; vi++) {
		if ( abs( gvg(v,vi) - n ) < mindist ) {
			mindist = abs( gvg(v,vi) - n );
			closest = vi;
		}
	}
	return(closest);
}

double var ( const double sqsum, const double sum, const int n ) {
	double s;
	if (n>1) s = ( n*sqsum - pow(sum,2) ) / (n*(n-1));
	if (n==1) s = 0;
	return s;
}

double covar ( const double sumij, const double sumi, const double sumj, const int n ) {
	double c;
	if (n>1) c =  ( n*sumij - sumi*sumj ) / (n*(n-1));
	if (n==1) c = 0;
	return c;
}

void gvpe(gsl_vector *v, int i, double d) {
    gsl_vector_set(v,i, gvg(v,i)+d);
	return;
}

void gmpe(gsl_matrix *m, int c, int r, double d) {
	gsl_matrix_set(m,c,r,gmg(m,c,r)+d);
	return;
}
