#include "vector_stats.h"
#include "utils.h"
using namespace std;

vector_stats::vector_stats() {
	// do nothing
}

vector_stats::vector_stats ( const gsl_vector * A, 
					const gsl_vector * B, int ignorezeros ) {
	// if ignorezeros != 0, do not include zeros in statistics, and
	// adjust the number of entries accordingly;

	n = A->size;
	if ( ignorezeros != 0 ) {
		for	(int i=0; i < A->size; i++) 
			if ( abs( gvg(A,i) ) < 1E-15 || abs( gvg(B,i) ) < 1E-15 ) n--;
	}

	gsl_vector * sum1; sum1 = gsl_vector_calloc ( n );
	gsl_vector * sumk; sumk = gsl_vector_calloc ( n );
	gsl_vector * sum1k; sum1k = gsl_vector_calloc ( n );
	K = gsl_vector_calloc ( n );	
	gsl_vector * sum1_2; sum1_2 = gsl_vector_calloc ( n );
	gsl_vector * sumk_2; sumk_2 = gsl_vector_calloc ( n );
	gsl_vector * sum1k_2; sum1k_2 = gsl_vector_calloc ( n );
	K2 = gsl_vector_calloc ( n );	

	double crosssum = 0;
	sum=sumsq=sum2=sumsq2=0;
	
	for	(int i=0; i < n; i++) {
		if ( ignorezeros != 0 && abs( gvg(A,i) ) < 1E-15  ) continue;
		if ( ignorezeros != 0 && abs( gvg(B,i) ) < 1E-15  ) continue;

		sum += gvg(A,i);
		sumsq += pow( gvg(A,i) , 2);
		sum2 += gvg(B,i);
		sumsq2 += pow( gvg(B,i) , 2);

		crosssum += gvg(A,i) * gvg(B,i);
		for (int k=0; k< n-i; k++ ) {
			gvpe( sum1, k, gvg(A,i) );
			gvpe( sumk, k, gvg(A,i+k) );
			gvpe( sum1k, k, gvg(A,i)*gvg(A,i+k) );

			gvpe( sum1_2, k, gvg(B,i) );
			gvpe( sumk_2, k, gvg(B,i+k) );
			gvpe( sum1k_2, k, gvg(B,i)*gvg(B,i+k) );
		} 
	}
	
	mean = sum / n;
	if (n>1) stdev = pow( ( n*sumsq - pow(sum,2) ) / (n*(n-1)) , 0.5);
	if (n==1) stdev = 0;

	mean2 = sum2 / n;
	if (n>1) stdev2 = pow( ( n*sumsq2 - pow(sum2,2) ) / (n*(n-1)) , 0.5);
	if (n==1) stdev2 = 0;

	crosscorrcoeff = ( (n*crosssum - sum*sum2) / (n*(n-1)) ) / (stdev * stdev2);

	double d;
	for (int k=0; k<n-1; k++) {
		d = gvg(sum1k,k) - mean * ( gvg(sum1,k) + gvg(sumk,k) + (k-n)* mean );
//		d /= (sumsq - 2.0*sum*mean + pow( mean, 2 ) );
		d /= (n-k-1.0);
		d /= pow(stdev,2);
		gvs(K,k,d);
		
		d = gvg(sum1k_2,k) - mean2 * ( gvg(sum1_2,k) + gvg(sumk_2,k) + (k-n)*mean2 );
//		d /= (sumsq2 - 2.0*sum2*mean2 + pow( mean2, 2 ) );
		d /= (n-k-1.0);
		d /= pow(stdev2,2);
		gvs(K2,k,d);
	}
	gsl_vector_free(sum1k);
	gsl_vector_free(sum1);
	gsl_vector_free(sumk);
	gsl_vector_free(sum1k_2);
	gsl_vector_free(sum1_2);
	gsl_vector_free(sumk_2);
}
	
vector_stats::vector_stats ( const gsl_vector * V, int ignorezeros ) {

	// if ignorezeros != 0, do not include zeros in statistics, and
	// adjust the number of entries accordingly;

	n = V->size;
	if ( ignorezeros != 0 ) {
		for	(int i=0; i < V->size; i++) if ( abs( gvg(V,i) ) < 1E-15 ) n--;
	}

	gsl_vector * sum1; sum1 = gsl_vector_calloc ( n );
	gsl_vector * sumk; sumk = gsl_vector_calloc ( n );
	gsl_vector * sum1k; sum1k = gsl_vector_calloc ( n );
	K = gsl_vector_calloc ( n );	
	K2 = gsl_vector_calloc ( n );	
	
	sum=sumsq=0;
	for	(int i=0; i < n; i++) {
		if ( ignorezeros != 0 && abs( gvg(V,i) ) < 1E-15  ) continue;
		sum += gvg(V,i);
		sumsq += pow( gvg(V,i) , 2);
		for (int k=0; k< n-i; k++ ) {
			gvpe( sum1, k, gvg(V,i) );
			gvpe( sumk, k, gvg(V,i+k) );
			gvpe( sum1k, k, gvg(V,i)*gvg(V,i+k) );
		} 
	}
	
	mean = sum / n;
	if (n>1) stdev = pow( ( n*sumsq - pow(sum,2) ) / (n*(n-1)) , 0.5);
	if (n==1) stdev = 0;

	double d;
	for (int k=0; k<n-1; k++) {
		d = gvg(sum1k,k) - mean * ( gvg(sum1,k) + gvg(sumk,k) + (k-n)*mean );
//		d /= (sumsq - 2.0*sum*mean + pow( mean, 2 ) );
		d /= (n-k-1.0);
		d /= pow(stdev,2);
		gvs(K,k,d);
	}
	gsl_vector_free(sum1k);
	gsl_vector_free(sum1);
	gsl_vector_free(sumk);
}

vector_stats::~vector_stats () {
	K = gsl_vector_alloc(1);
	K2 = gsl_vector_alloc(1);
	gsl_vector_free(K);
	gsl_vector_free(K2);
}


void vector_stats::first_moment_analysis(const gsl_vector * a, const gsl_vector * b) {
	double Isum=0;
	double wIsum=0;
	gsl_vector * ones;
	ones = gsl_vector_alloc(a->size);
	gsl_vector_set_all(ones, 1.0);
	gsl_blas_ddot(b,ones,&Isum);
	gsl_blas_ddot(a,b,&wIsum);

	first_moment = wIsum/Isum;

	// find index of point closest to first_moment, index 
	double closest=1E8;
	for (int i=0; i<a->size; i++) {
		double x = gsl_vector_get(a,i);
		if ( abs(x-first_moment ) < closest ) {
			closest = abs(x - first_moment);
			first_moment_i = i;
		}
	}
	gsl_vector_free(ones);
}

